﻿using System;
using System.Diagnostics;

namespace CSharpRecipes
{
    class MainTester
    {
        static void TestLINQ()
        {
            PrintHeader("1章 LINQ");
            LINQ.TestLinqMessageQueue();
            LINQ.TestSetSemantics();
            LINQ.TestCompiledQuery();
            LINQ.TestLinqForCulture();
            LINQ.TestWeightedMovingAverage();
        }

        static void TestGeneric()
        {
            PrintHeader("2章 Generic");
            Generics.TestGenericClassInstanceCounter();
            Generics.TestGetGenTypeOf();
            Generics.UseNonGenericArrayList();
            Generics.UseGenericList();
            Generics.UseNonGenericQueue();
            Generics.UseGenericQueue();
            Generics.UseNonGenericStack();
            Generics.UseGenericStack();
            Generics.CloneQueue();
            Generics.CloneStack();
            Generics.UseLinkedList();
            Generics.TestNullableStruct();
            Generics.TestReversibleSortedList();
            Generics.MakeCollectionReadOnly();
            Generics.UseNonGenericHashtable();
            Generics.UseGenericDictionary();
            Generics.CloneGenericDictionary();
            Generics.CopyToGenericDictionary();
            Generics.ShowForEachWithDictionary();
            Generics.TestDisposableListCls();
            Generics.ShowSettingFieldsToDefaults();
        }

        static void TestIteratorsAndPartialTypesAndPartialMethods()
        {
            PrintHeader("3章 反復子、部分型、部分メソッド");
            IteratorsAndPartialTypesAndPartialMethods.TestShoppingCart();
            IteratorsAndPartialTypesAndPartialMethods.TestStampCollection();
            IteratorsAndPartialTypesAndPartialMethods.TestIterators();
            IteratorsAndPartialTypesAndPartialMethods.TestIteratorsAndLinq();
            IteratorsAndPartialTypesAndPartialMethods.TestYieldBreak();
            IteratorsAndPartialTypesAndPartialMethods.TestFinallyAndIterators();
            IteratorsAndPartialTypesAndPartialMethods.CreateNestedObjects();
            IteratorsAndPartialTypesAndPartialMethods.TestPartialMethods();
        }

        static void TestExceptionHandling()
        {
            PrintHeader("4章 例外処理");
            ExceptionHandling.TestExceptionCode();
            ExceptionHandling.PreventLossOfException();
            ExceptionHandling.NotPreventLossOfException();
            ExceptionHandling.ReflectionException();
            
            // 未処理例外に対する処理を確認する場合、以下の行のコメントアウトを解除してください。
            // これによりアプリケーションが終了する前に何らかの処理を実行できることが確認できます。
            //ExceptionHandling.SetupLastChanceSEH();
            try
            {
                ExceptionHandling.TestToFullDisplayString();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
            ExceptionHandling.TestSpecializedException();
            Console.WriteLine(ExceptionHandling.GetStackTraceInfo());
            try
            {
                ExceptionHandling.TestPollingAsyncDelegate();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
            ExceptionHandling.TestPollingAsyncDelegate();
            ExceptionHandling.TestExceptionData();
            ExceptionHandling.TestGetNestedSetOfExceptions();
        }

        static void TestDiagnostics()
        {
            PrintHeader("5章 診断");
            Diagnostics.EnableTracing();
            Diagnostics.ProcessRespondingState state;
            foreach (Process P in Process.GetProcesses())
            {
                state = Diagnostics.IsProcessResponding(P);
                if (state == Diagnostics.ProcessRespondingState.NotResponding)
                {
                    Console.WriteLine("{0} は応答がありません。", P.ProcessName);
                }
            }
            Diagnostics.TestEventLogClass();
            Diagnostics.FindAnEntryInEventLog();
            Diagnostics.FindAnEntryInEventLog2();
            Diagnostics.WatchForAppEvent((EventLog.GetEventLogs())[0]);
            Diagnostics.TestCreateSimpleCounter();
            Diagnostics.TestTraceFactoryClass();
            Diagnostics.TestCaptureOutput();
            Diagnostics.TestCustomDebuggerDisplay();
        }
        
        static void TestDelegatesEventsFunctionalProgramming()
        {
            PrintHeader("6章 デリゲート、イベント、ラムダ式");
            DelegatesEventsLambdaExpressions.InvokeInReverse();
            DelegatesEventsLambdaExpressions.InvokeEveryOtherOperation();
            DelegatesEventsLambdaExpressions.InvokeWithTest();
            DelegatesEventsLambdaExpressions.TestIndividualInvokesReturnValue();
            DelegatesEventsLambdaExpressions.TestIndividualInvokesExceptions();
            DelegatesEventsLambdaExpressions.TestSimpleSyncDelegate();
            DelegatesEventsLambdaExpressions.TestSimpleAsyncDelegate();
            DelegatesEventsLambdaExpressions.TestComplexSyncDelegate();
            DelegatesEventsLambdaExpressions.TestCallbackAsyncDelegate();
            DelegatesEventsLambdaExpressions.FindSpecificInterfaces();
            DelegatesEventsLambdaExpressions.TestObserverPattern();
            DelegatesEventsLambdaExpressions.TestUsingLambdaExpressions();
            DelegatesEventsLambdaExpressions.TestParameterModifiers();
            DelegatesEventsLambdaExpressions.TestClosure();
            DelegatesEventsLambdaExpressions.TestFunctors();
        }

        static void TestDataStructsAndAlgorithms()
        {
            PrintHeader("7章 データ構造とアルゴリズム");
            DataStructsAndAlgorithms.CreateHashCodeDataType();
            DataStructsAndAlgorithms.CreatePriorityQueue();
            DataStructsAndAlgorithms.TestMultiMap();
            DataStructsAndAlgorithms.TestBinaryTree();
            DataStructsAndAlgorithms.TestManagedTreeWithNoBinaryTreeClass();
            DataStructsAndAlgorithms.TestNTree();
            DataStructsAndAlgorithms.TestSet();
        }

        static void TestFileSystemIO()
        {
            PrintHeader("8章 ファイルシステムI/O");

            FileSystemIO.ManipulateFileAttributes();
            FileSystemIO.RenameFile();
            FileSystemIO.PlatformIndependentEol();
            FileSystemIO.ManipulateDirectoryAttributes();
            FileSystemIO.RenameDirectory();
            FileSystemIO.SearchDirFileWildcards();
            FileSystemIO.ObtainDirTree();
            FileSystemIO.ParsePath();
            FileSystemIO.ParsePathsEnvironmentVariables();
            FileSystemIO.LaunchInteractConsoleUtilities();
            FileSystemIO.LockSubsectionsOfAFile();
            FileSystemIO.WaitFileSystemAction();
            FileSystemIO.CompareVersionInfo();
            FileSystemIO.TestAllDriveInfo();
            FileSystemIO.TestCompressNewFile();
        }

        static void TestReflection()
        {
            PrintHeader("9章 リフレクション");

            Reflection.ListImportedAssemblies();
            Reflection.ListExportedTypes();
            Reflection.FindOverriddenMethods();
            Reflection.FindMembersInAssembly();
            Reflection.ObtainNestedTypes();
            Reflection.DisplayInheritanceHierarchy();
            Reflection.FindSubclasses();
            Reflection.FindSerializableTypes();
            Reflection.DynamicInvocation();
            Reflection.TestIsGeneric();
            Reflection.TestGetLocalVars();
            Reflection.TestCreateMultiMap();
            Reflection.TestGetIL();
        }

        static void TestSecurity()
        {
            PrintHeader("10章 セキュリティ");
            Security.ControlAccess();
            Security.EncDecString();
            Security.EncDecFile();
            Security.CleanUpCrypto();
            Security.VerifyNonStringCorruption();
            Security.SecurelyStoringData();
            Security.SafeAssert();
            Security.VerifyAssemblyPerms();
            Security.MinimizeAttackSurface();
            Security.TestViewFileRegRights();
            Security.TestGrantRevokeFileRights();
            Security.TestSecureString();
            Security.TestUnicodeEncodingWithSecurity();
        }

        static void TestThreadingSync()
        {
            PrintHeader("11章 スレッドと同期");

            ThreadingSync.PerThreadStatic();
            ThreadingSync.ThreadSafeAccess();
            ThreadingSync.CompletionAsyncDelegate();
            ThreadingSync.PreventSilentTermination();
            ThreadingSync.StoreThreadDataPrivately();
            ThreadingSync.Halo3Session.Play();
            ThreadingSync.TestResetEvent();
            ThreadingSync.TestManualNamedEvent();
            ThreadingSync.TestAutoNamedEvent();
            ThreadingSync.TestInterlocked();
            ThreadingSync.TestReaderWriterLockSlim();
        }

        static void TestNumbersAndEnums()
        {
            PrintHeader("12章 数値、列挙型");
            Console.WriteLine(((float)1 / 3));                      // .3333333
            Console.WriteLine(((float)1 / 3 == (float)0.33333));    // False
            Console.WriteLine(((float)1 / 3 == (float).3333333));   // True
            Console.WriteLine("double.Epsilon: " + double.Epsilon);
            Console.WriteLine("float.Epsilon: " + float.Epsilon);
            Console.WriteLine();

            Console.WriteLine(NumbersEnums.ConvertDegreesToRadians(0));
            Console.WriteLine(NumbersEnums.ConvertDegreesToRadians(90));
            Console.WriteLine(NumbersEnums.ConvertDegreesToRadians(180));
            Console.WriteLine(NumbersEnums.ConvertDegreesToRadians(270));
            Console.WriteLine(NumbersEnums.ConvertDegreesToRadians(360));
            Console.WriteLine(NumbersEnums.ConvertDegreesToRadians(12));

            Console.WriteLine(NumbersEnums.ConvertRadiansToDegrees(1.5707963267949));
            Console.WriteLine(NumbersEnums.ConvertRadiansToDegrees(3.14159265358979));
            Console.WriteLine(NumbersEnums.ConvertRadiansToDegrees(4.71238898038469));
            Console.WriteLine(NumbersEnums.ConvertRadiansToDegrees(6.28318530717959));
            Console.WriteLine(NumbersEnums.ConvertRadiansToDegrees(5));
            Console.WriteLine(NumbersEnums.ConvertRadiansToDegrees(6));
            Console.WriteLine();


            NumbersEnums.TestBitwiseOperators();
            Console.WriteLine();

            NumbersEnums.TestBase10();
            Console.WriteLine();

            NumbersEnums.TestDetermineIfStringIsNumber();
            Console.WriteLine();


            NumbersEnums.TestRound();
            Console.WriteLine();


            Console.WriteLine(NumbersEnums.CelsiusToFahrenheit(0));
            Console.WriteLine(NumbersEnums.CelsiusToFahrenheit(40));
            Console.WriteLine(NumbersEnums.CelsiusToFahrenheit(50));
            Console.WriteLine(NumbersEnums.CelsiusToFahrenheit(-50));
            Console.WriteLine();
            Console.WriteLine(NumbersEnums.CelsiusToKelvin(0));
            Console.WriteLine(NumbersEnums.CelsiusToKelvin(40));
            Console.WriteLine(NumbersEnums.CelsiusToKelvin(50));
            Console.WriteLine(NumbersEnums.CelsiusToKelvin(-50));
            Console.WriteLine();
            Console.WriteLine(NumbersEnums.FahrenheitToCelsius(0));
            Console.WriteLine(NumbersEnums.FahrenheitToCelsius(40));
            Console.WriteLine(NumbersEnums.FahrenheitToCelsius(50));
            Console.WriteLine(NumbersEnums.FahrenheitToCelsius(-50));
            Console.WriteLine();
            Console.WriteLine(NumbersEnums.FahrenheitToKelvin(0));
            Console.WriteLine(NumbersEnums.FahrenheitToKelvin(40));
            Console.WriteLine(NumbersEnums.FahrenheitToKelvin(50));
            Console.WriteLine(NumbersEnums.FahrenheitToKelvin(-50));
            Console.WriteLine();

            NumbersEnums.TestNarrowing();
            Console.WriteLine();


            PrintHeader("Enumerations Chapter Tests");
            NumbersEnums.TestDisplayEnumValue();
            NumbersEnums.TestConvertingEnums();
            NumbersEnums.HandleEnum(Language.All);
            NumbersEnums.HandleEnum(Language.CSharp);
            NumbersEnums.HandleEnum(Language.Other);
            NumbersEnums.HandleEnum(Language.VB6);
            NumbersEnums.HandleEnum(Language.VBNET);
            NumbersEnums.HandleEnum((Language)100);
            NumbersEnums.HandleEnum((Language)1);
            NumbersEnums.HandleEnum((Language)100);
            NumbersEnums.HandleEnum((Language)8);
            NumbersEnums.HandleEnum((Language)3);

            NumbersEnums.TestEnumFlags();

            NumbersEnums.TestTruncate();
        }
        
        static void Main(string[] args)
        {
            var tests = new Action[]
            {
                TestLINQ,
                TestGeneric,
                TestIteratorsAndPartialTypesAndPartialMethods,
                TestExceptionHandling,
                TestDiagnostics,
                TestDelegatesEventsFunctionalProgramming,
                TestDataStructsAndAlgorithms,
                TestFileSystemIO,
                TestReflection,
                TestSecurity,
                TestThreadingSync,
                TestNumbersAndEnums
            };

            foreach (var test in tests)
            {
                try
                {
                    test();
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.ToString());
                }
            }

            // Enterキーが押されるまで待機します
            Console.WriteLine("終了するにはEnterキーを押してください...");
            Console.ReadLine();
        }
        
        #region UtilityCode
        static void PrintHeader(string testtype)
        {
            Console.WriteLine("************************************************************");
            Console.WriteLine("** " + testtype + " のテストコードを実行中...");
            Console.WriteLine("************************************************************");
        }
        #endregion
    }
}
