﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Data.SqlClient;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.Security.Permissions;
using System.Text;
using System.Threading;

// 4.7でオーバーロードした==演算子をテストするため、意図的に同じ変数同士を比較しています
#pragma warning disable 1718 // 同じ変数と比較されました。他の変数と比較しますか？

namespace CSharpRecipes
{
    public static class ExceptionHandling
    {
        #region "4.0 概要"

        public static void CallCOMMethod()
        {
            try
            {
                // COMオブジェクトのメソッドを呼び出します
                //myCOMObj.Method1();
            }
            catch (System.Runtime.InteropServices.ExternalException exte)
            {
                // メソッド呼び出しによって発生する潜在的なCOM例外を処理します
                Console.WriteLine(exte);
            }
            catch (InvalidOperationException ioe)
            {
                // 現在のステータスが不適切なCOMオブジェクトに対するメソッド呼び出しの際に
                // 発生する潜在的なCOM例外を処理します
                Console.WriteLine(ioe);
            }
        }

        public static void RunCommand(string connection, string command)
        {
            SqlConnection sqlConn = null;
            SqlCommand sqlComm = null;

            try
            {
                sqlConn = new SqlConnection(connection);
                sqlComm = new SqlCommand(command, sqlConn);

                sqlConn.Open();
                sqlComm.ExecuteNonQuery();
            }
            finally
            {
                if (sqlComm != null)
                    sqlComm.Dispose();

                if (sqlConn != null)
                    sqlConn.Dispose();
            }
        }

        #endregion 4.0 概要

        #region "4.1 キャッチした例外を再スローするタイミングを把握する"
        public static void TestExceptionCode()
        {
            try
            {
                // *********サンプルコードの開始
                try
                {
                    Console.WriteLine("tryの中");
                    int z2 = 9999999;
                    checked { z2 *= 999999999; }
                }
                catch (OverflowException oe)
                {
                    // オーバーフロー例外が発生したことを記録する
                    EventLog.WriteEntry("MyApplication", oe.Message, EventLogEntryType.Error);
                    throw;
                }
                // *********サンプルコードの終了
            }
            catch (Exception ex)
            {
                // サンプルコードが正しく終了するようこの位置で例外をキャッチしています
                Console.WriteLine(ex);
            }
        }


        #endregion

        #region "4.2 Finallyブロック中においても例外オブジェクトが失われないようにしたい"
        private static void PreventLossOfExceptionFormat()
        {
            try
            {
                //...
            }
            catch (Exception e)
            {
                Console.WriteLine("エラーメッセージ == " + e.Message);
                throw;
            }
            finally
            {
                try
                {
                    //...
                }
                catch (Exception e)
                {
                    Console.WriteLine("finallyブロック中で例外が発生しました。" +
                        "エラーメッセージ：" + e.Message);
                }
            }
        }


        private static void ThrowReplacementException()
        {
            try
            {
                Console.WriteLine("内側のtryブロック");
                int z2 = 9999999;
                checked { z2 *= 999999999; }
            }
            catch (OverflowException ofe)
            {
                Console.WriteLine("オーバーフローが発生しました。エラーメッセージ：" +
                                   ofe.Message);
                throw;
            }
            catch (Exception e)
            {
                Console.WriteLine("別の種類の例外が発生しました。" +
                                  "エラーメッセージ：" + e.Message);
                throw;
            }
            finally
            {
                Console.WriteLine("内側のfinallyブロック");
                throw (new Exception("おおっと"));
            }
        }

        public static void NotPreventLossOfException()
        {
            try
            {
                Console.WriteLine("外側のtryブロック");
                ThrowReplacementException();
            }
            catch (Exception e)
            {
                Console.WriteLine("外側のcatchブロック。例外をキャッチ：" + e.Message);
            }
            finally
            {
                Console.WriteLine("外側のfinallyブロック");
            }
        }

        private static void ThrowException()
        {
            try
            {
                Console.WriteLine("内側のtryブロック");
                int z2 = 9999999;
                checked { z2 *= 999999999; }
            }
            catch (OverflowException ofe)
            {
                Console.WriteLine("オーバーフローが発生しました。エラーメッセージ：" +
                                   ofe.Message);
                throw;
            }
            catch (Exception e)
            {
                Console.WriteLine("別の種類の例外が発生しました。" +
                                  "エラーメッセージ：" + e.Message);
                throw;
            }
            finally
            {
                try
                {
                    Console.WriteLine("内側のfinallyブロック");
                    throw (new Exception("おおっと"));
                }
                catch (Exception e)
                {
                    Console.WriteLine("finallyブロック中でエラーが発生しました。" +
                             "エラーメッセージ：" + e.Message);
                }
            }
        }



        public static void PreventLossOfException()
        {
            try
            {
                Console.WriteLine("外側のtryブロック");
                ThrowException();
            }
            catch (Exception e)
            {
                Console.WriteLine("外側のcatchブロック。例外をキャッチ：" + e.Message);
            }
            finally
            {
                Console.WriteLine("外側のfinallyブロック");
            }
        }
        #endregion

        #region "4.3 リフレクション経由で呼び出すメソッドから発生する例外を処理する"
        public static void ReflectionException()
        {
            Type reflectedClass = typeof(ExceptionHandling);

            try
            {
                MethodInfo methodToInvoke = reflectedClass.GetMethod("TestInvoke");

                if (methodToInvoke != null)
                {
                    methodToInvoke.Invoke(null, null);
                }
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToShortDisplayString());
            }
        }

        public static void TestInvoke()
        {
            throw (new Exception("呼び出し中のメソッド内での例外スロー"));
        }

        public static string ToShortDisplayString(this Exception ex)
        {
            StringBuilder displayText = new StringBuilder();
            WriteExceptionShortDetail(displayText, ex);
            foreach (Exception inner in ex.GetNestedExceptionList()) // from 7.14
            {
                displayText.AppendFormat("**** 内部例外 始点 ****{0}", Environment.NewLine);
                WriteExceptionShortDetail(displayText, inner);
                displayText.AppendFormat("**** 内部例外 終点 ****{0}", Environment.NewLine);
            }
            return displayText.ToString();
        }

        public static void WriteExceptionShortDetail(StringBuilder builder, Exception ex)
        {
            builder.AppendFormat("Message: {0}{1}", ex.Message, Environment.NewLine);
            builder.AppendFormat("Type: {0}{1}", ex.GetType(), Environment.NewLine);
            builder.AppendFormat("Source: {0}{1}", ex.Source, Environment.NewLine);
            builder.AppendFormat("TargetSite: {0}{1}", ex.TargetSite, Environment.NewLine);
        }


        #endregion

        #region "4.4 ハンドルされていない例外を予防する"
        public static void LastChanceHandler(object sender, UnhandledExceptionEventArgs args)
        {
            Exception e = (Exception)args.ExceptionObject;

            Console.WriteLine("未処理の例外 == " + e.ToString());
            if (args.IsTerminating)
            {
                Console.WriteLine("アプリケーションを終了しています");
            }
            else
            {
                Console.WriteLine("アプリケーションは終了されません");
            }
        }

        public static void SetupLastChanceSeh()
        {
            AppDomain currentDomain = AppDomain.CurrentDomain;
            currentDomain.UnhandledException += new UnhandledExceptionEventHandler(LastChanceHandler);

            System.Threading.Thread.Sleep(2000);

            // デバッガを使用してこのコードを実行する場合、未処理の例外ハンドラを設定してあるにもかかわらず
            // デバッガはこのスコープを抜けて、例外が発生したことを通知します。
            // デバッグモードでこのコードを実行する場合には、デバッガにより例外がキャッチされたとしても
            // 実行を継続することによって未処理の例外ハンドラへ処理が進むことを確認できます。
            throw (new Exception("MyUnhandledException"));
        }
        #endregion

        #region "4.5 例外情報を取得する"
        public static string ToFullDisplayString(this Exception ex)
        {
            StringBuilder displayText = new StringBuilder();
            WriteExceptionDetail(displayText, ex);
            foreach (Exception inner in ex.GetNestedExceptionList()) // from 7.14
            {
                displayText.AppendFormat("**** 内部例外 始点 ****{0}", Environment.NewLine);
                WriteExceptionDetail(displayText, inner);
                displayText.AppendFormat("**** 内部例外 終点 ****{0}", Environment.NewLine);
            }
            return displayText.ToString();
        }

        public static void WriteExceptionDetail(StringBuilder builder, Exception ex)
        {
            builder.AppendFormat("Message: {0}{1}", ex.Message, Environment.NewLine);
            builder.AppendFormat("Type: {0}{1}", ex.GetType(), Environment.NewLine);
            builder.AppendFormat("HelpLink: {0}{1}", ex.HelpLink, Environment.NewLine);
            builder.AppendFormat("Source: {0}{1}", ex.Source, Environment.NewLine);
            builder.AppendFormat("TargetSite: {0}{1}", ex.TargetSite, Environment.NewLine);
            builder.AppendFormat("Data:{0}", Environment.NewLine);
            foreach (DictionaryEntry de in ex.Data)
            {
                builder.AppendFormat("\t{0} : {1}{2}",
                    de.Key, de.Value, Environment.NewLine);
            }
            builder.AppendFormat("StackTrace: {0}{1}", ex.StackTrace, Environment.NewLine);
        }

        public static void TestToFullDisplayString()
        {
            Exception innerInner = new Exception("内部内部例外");
            innerInner.Data.Add("内部内部例外のKey1", "内部内部例外のValue1");
            ArgumentException inner = new ArgumentException("内部例外", innerInner);
            inner.Data.Add("内部例外のKey1", "内部例外のValue1");
            NullReferenceException se = new NullReferenceException("テストメッセージ。", inner);
            se.HelpLink = "MyComponent.hlp";
            se.Source = "MyComponent";
            se.Data.Add("外部例外のKey1", "外部例外のValue1");
            se.Data.Add("外部例外のKey2", "外部例外のValue2");
            se.Data.Add("外部例外のKey3", "外部例外のValue3");

            try
            {
                throw (se);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToFullDisplayString());
                //DisplayInnerException(e);
                //Console.WriteLine(e.GetBaseException().ToString());
            }
        }


        public static void DisplayInnerException(Exception exception)
        {
            Console.WriteLine(exception.GetBaseException().ToString());
        }
        #endregion

        #region "4.6 諸悪の根源を素早く見つけ出す方法"
        public static void DisplayException2(Exception exception)
        {
            Console.WriteLine(exception.GetBaseException().ToString());
        }
        #endregion

        #region "4.7 新しい例外型を作成する"
        public static void TestSpecializedException()
        {
            // RemoteComponentExceptionクラスの内部例外をテストするための
            // 汎用の例外オブジェクト
            Exception inner = new Exception("内部例外");

            // 各コンストラクタのテスト
            Console.WriteLine(Environment.NewLine + Environment.NewLine + "各コンストラクタのテスト");
            RemoteComponentException se1 = new RemoteComponentException();
            RemoteComponentException se2 = new RemoteComponentException("se2用のテストメッセージ");
            RemoteComponentException se3 = new RemoteComponentException("se3用のテストメッセージ", inner);
            RemoteComponentException se4 = new RemoteComponentException("se4用のテストメッセージ",
                "MyServer");
            RemoteComponentException se5 = new RemoteComponentException("se5用のテストメッセージ", inner,
                "MyServer");

            // ServerNameプロパティのテスト
            Console.WriteLine(Environment.NewLine + "新規作成したServerNameプロパティのテスト");
            Console.WriteLine("se1.ServerName == " + se1.ServerName);
            Console.WriteLine("se2.ServerName == " + se2.ServerName);
            Console.WriteLine("se3.ServerName == " + se3.ServerName);
            Console.WriteLine("se4.ServerName == " + se4.ServerName);
            Console.WriteLine("se5.ServerName == " + se5.ServerName);

            // オーバーライドされたMessageプロパティのテスト
            Console.WriteLine(Environment.NewLine + "オーバーライドされたMessageプロパティのテスト");
            Console.WriteLine("se1.Message == " + se1.Message);
            Console.WriteLine("se2.Message == " + se2.Message);
            Console.WriteLine("se3.Message == " + se3.Message);
            Console.WriteLine("se4.Message == " + se4.Message);
            Console.WriteLine("se5.Message == " + se5.Message);

            // オーバーライドされたToStringメソッドのテスト
            Console.WriteLine(Environment.NewLine + "オーバーライドされたToStringメソッドのテスト");
            Console.WriteLine("se1.ToString() == " + se1.ToString());
            Console.WriteLine("se2.ToString() == " + se2.ToString());
            Console.WriteLine("se3.ToString() == " + se3.ToString());
            Console.WriteLine("se4.ToString() == " + se4.ToString());
            Console.WriteLine("se5.ToString() == " + se5.ToString());

            // ToBaseStringメソッドのテスト
            Console.WriteLine(Environment.NewLine + "ToBaseStringメソッドのテスト");
            Console.WriteLine("se1.ToBaseString() == " + se1.ToBaseString());
            Console.WriteLine("se2.ToBaseString() == " + se2.ToBaseString());
            Console.WriteLine("se3.ToBaseString() == " + se3.ToBaseString());
            Console.WriteLine("se4.ToBaseString() == " + se4.ToBaseString());
            Console.WriteLine("se5.ToBaseString() == " + se5.ToBaseString());

            // オーバーライドされた==演算子のテスト
            Console.WriteLine(Environment.NewLine + "オーバーライドされた==演算子のテスト");
            Console.WriteLine("se1 == se1 == " + (se1 == se1));
            Console.WriteLine("se2 == se1 == " + (se2 == se1));
            Console.WriteLine("se3 == se1 == " + (se3 == se1));
            Console.WriteLine("se4 == se1 == " + (se4 == se1));
            Console.WriteLine("se5 == se1 == " + (se5 == se1));
            Console.WriteLine("se5 == se4 == " + (se5 == se4));

            // オーバーライドされた!=演算子のテスト
            Console.WriteLine(Environment.NewLine + "オーバーライドされた!=演算子のテスト");
            Console.WriteLine("se1 != se1 == " + (se1 != se1));
            Console.WriteLine("se2 != se1 == " + (se2 != se1));
            Console.WriteLine("se3 != se1 == " + (se3 != se1));
            Console.WriteLine("se4 != se1 == " + (se4 != se1));
            Console.WriteLine("se5 != se1 == " + (se5 != se1));
            Console.WriteLine("se5 != se4 == " + (se5 != se4));

            // オーバーライドされたGetBaseExceptionメソッドのテスト
            Console.WriteLine(Environment.NewLine + "オーバーライドされたGetBaseExceptionメソッドのテスト");
            Console.WriteLine("se1.GetBaseException() == " + se1.GetBaseException());
            Console.WriteLine("se2.GetBaseException() == " + se2.GetBaseException());
            Console.WriteLine("se3.GetBaseException() == " + se3.GetBaseException());
            Console.WriteLine("se4.GetBaseException() == " + se4.GetBaseException());
            Console.WriteLine("se5.GetBaseException() == " + se5.GetBaseException());

            // オーバーライドされたGetHashCodeメソッドのテスト
            Console.WriteLine(Environment.NewLine + "オーバーライドされたGetHashCodeメソッドのテスト");
            Console.WriteLine("se1.GetHashCode() == " + se1.GetHashCode());
            Console.WriteLine("se2.GetHashCode() == " + se2.GetHashCode());
            Console.WriteLine("se3.GetHashCode() == " + se3.GetHashCode());
            Console.WriteLine("se4.GetHashCode() == " + se4.GetHashCode());
            Console.WriteLine("se5.GetHashCode() == " + se5.GetHashCode());

            // シリアライズのテスト
            Console.WriteLine(Environment.NewLine + "シリアライズ/デシリアライズのテスト");
            BinaryFormatter binaryWrite = new BinaryFormatter();
            Stream ObjectFile = File.Create("se1.object");
            binaryWrite.Serialize(ObjectFile, se1);
            ObjectFile.Close();
            ObjectFile = File.Create("se2.object");
            binaryWrite.Serialize(ObjectFile, se2);
            ObjectFile.Close();
            ObjectFile = File.Create("se3.object");
            binaryWrite.Serialize(ObjectFile, se3);
            ObjectFile.Close();
            ObjectFile = File.Create("se4.object");
            binaryWrite.Serialize(ObjectFile, se4);
            ObjectFile.Close();
            ObjectFile = File.Create("se5.object");
            binaryWrite.Serialize(ObjectFile, se5);
            ObjectFile.Close();

            BinaryFormatter binaryRead = new BinaryFormatter();
            ObjectFile = File.OpenRead("se1.object");
            object Data = binaryRead.Deserialize(ObjectFile);
            Console.WriteLine("----------" + Environment.NewLine + Data);
            ObjectFile.Close();
            ObjectFile = File.OpenRead("se2.object");
            Data = binaryRead.Deserialize(ObjectFile);
            Console.WriteLine("----------" + Environment.NewLine + Data);
            ObjectFile.Close();
            ObjectFile = File.OpenRead("se3.object");
            Data = binaryRead.Deserialize(ObjectFile);
            Console.WriteLine("----------" + Environment.NewLine + Data);
            ObjectFile.Close();
            ObjectFile = File.OpenRead("se4.object");
            Data = binaryRead.Deserialize(ObjectFile);
            Console.WriteLine("----------" + Environment.NewLine + Data);
            ObjectFile.Close();
            ObjectFile = File.OpenRead("se5.object");
            Data = binaryRead.Deserialize(ObjectFile);
            Console.WriteLine("----------" + Environment.NewLine + Data + Environment.NewLine
                + "----------");
            ObjectFile.Close();

            Console.WriteLine(Environment.NewLine + "テスト終了" + Environment.NewLine);
        }


        [Serializable]
        public class RemoteComponentException : Exception, ISerializable
        {
            #region コンストラクタ
            // 一般的な例外クラスのコンストラクタ
            public RemoteComponentException()
                : base()
            {
            }

            public RemoteComponentException(string message)
                : base(message)
            {
            }

            public RemoteComponentException(string message, Exception innerException)
                : base(message, innerException)
            {
            }

            // 新規作成したServerNameパラメータを受け付けるコンストラクタ
            public RemoteComponentException(string message, string serverName)
                : base(message)
            {
                this.ServerName = serverName;
            }

            public RemoteComponentException(string message, Exception innerException, string serverName)
                : base(message, innerException)
            {
                this.ServerName = serverName;
            }

            // シリアライズ用コンストラクタ
            protected RemoteComponentException(SerializationInfo exceptionInfo,
                StreamingContext exceptionContext)
                : base(exceptionInfo, exceptionContext)
            {
                this.ServerName = exceptionInfo.GetString("ServerName");
            }
            #endregion // コンストラクタ

            #region プロパティ
            // サーバ名を保持する読み取り専用プロパティ
            public string ServerName { get; private set; }

            public override string Message
            {
                get
                {
                    if (string.IsNullOrEmpty(this.ServerName))
                        return string.Format(Thread.CurrentThread.CurrentCulture,
                            "{0}{1}名称不明のサーバでエラーが発生しました。",
                            base.Message, Environment.NewLine);
                    else
                        return
                            string.Format(Thread.CurrentThread.CurrentCulture,
                                "{0}{1}サーバ ({2}) でエラーが発生しました。",
                                base.Message, Environment.NewLine, this.ServerName);
                }
            }
            #endregion // プロ派thい

            #region オーバーライドメソッド
            // ToStringメソッド
            public override string ToString()
            {
                string errorString =
                    string.Format(Thread.CurrentThread.CurrentCulture,
                    "現在のクライアントに対してサーバ側でエラーが発生しました。" +
                    "{0}サーバ名：{1}{0}{2}",
                    Environment.NewLine, this.ServerName, this.ToFullDisplayString());
                return errorString;
            }

            // シリアライズ中に呼ばれ、追加したフィールド情報を格納します
            [SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.SerializationFormatter)]
            public override void GetObjectData(SerializationInfo info,
                StreamingContext context)
            {
                base.GetObjectData(info, context);
                info.AddValue("ServerName", this.ServerName);
            }
            #endregion // オーバーライドメソッド

            // base.ToStringメソッドを呼び出すメソッド
            public string ToBaseString()
            {
                return base.ToString();
            }
        }
        #endregion

        #region "4.8 スタックトレースを取得する"

        public static void ObtainingStackTrace()
        {
            string stackTraceInfo = GetStackTraceInfo();
            Console.WriteLine(stackTraceInfo);
        }

        public static string GetStackTraceInfo()
        {
            StringBuilder result = new StringBuilder();
            StackTrace st = new StackTrace(true);
            StackFrame[] frames = st.GetFrames();
            result.AppendLine("----------");
            // 最初のStackFrameはGetStackTraceInfo自身なので除外する
            for (int i = 1; i < frames.Length; i++)
            {
                result.Append(frames[i].ToString());
                result.AppendLine("----------");
            }
            return result.ToString();
        }

        #endregion

        #region "4.9 初回例外発生時にアプリケーションを一時停止する"
        // 本書の該当レシピを参照してください
        #endregion

        #region "4.10 非同期デリゲートからスローされた例外の処理"
        public static void TestPollingAsyncDelegate()
        {
            AsyncInvoke MI = new AsyncInvoke(TestAsyncInvoke.Method1);
            IAsyncResult AR = MI.BeginInvoke(null, null);

            while (!AR.IsCompleted)
            {
                System.Threading.Thread.Sleep(100);
                Console.Write('.');
            }
            Console.WriteLine("待機完了");

            // 非同期デリゲートのEndInvokeメソッドを呼び出す
            try
            {
                int RetVal = MI.EndInvoke(AR);
                Console.WriteLine("返り値：" + RetVal);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
        }


        public delegate int AsyncInvoke();
        public class TestAsyncInvoke
        {
            public static int Method1()
            {
                throw (new Exception("Method1"));
                //Console.WriteLine("Method1の呼び出し");
                //return 1;
            }
        }
        #endregion

        #region "4.11 Exception.Dataを使用して例外オブジェクトに追加情報を設定する"
        public static void TestExceptionData()
        {
            try
            {
                try
                {
                    try
                    {
                        try
                        {
                            ArgumentException irritable =
                                new ArgumentException("ああもう！");
                            irritable.Data["Cause"] = "マシンがクラッシュしました";
                            irritable.Data["Length"] = 10;
                            throw irritable;
                        }
                        catch (Exception e)
                        {
                            // 処理できるかどうかを確認...
                            if (e.Data.Contains("Cause"))
                                e.Data["Cause"] = "マシンは復旧しました";
                            throw;
                        }
                    }
                    catch (Exception e)
                    {
                        e.Data["Comment"] = "常に調子が悪いです";
                        throw;
                    }
                }
                catch (Exception e)
                {
                    e.Data["Reassurance"] = "エラー処理済みです";
                    throw;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("例外の追加情報：");
                foreach (DictionaryEntry de in e.Data)
                {
                    Console.WriteLine("\t{0} : {1}", de.Key, de.Value);
                }
            }
        }
        #endregion

        #region "4.12 Windowsフォームアプリケーションにおける未処理例外への対応"
        // UnhandledThreadExceptionsプロジェクトを参照してください
        #endregion

        #region "4.13 WPFアプリケーションにおける未処理例外への対応"
        // UnhandledWPFExceptionsプロジェクトを参照してください
        #endregion

        #region "4.14 一般的なエラーに対する例外の分析"
        public static void TestGetNestedSetOfExceptions()
        {
            // オブジェクトのインスタンスに
            // オブジェクト参照がない場合をシミュレーション
            NullReferenceException nrex = new NullReferenceException("不正なオブジェクトにアクセスしました");
            nrex.Data.Add("Bad ObjectID", "0x34874573");
            // メッセージのフォーマットに対するエラーが発生
            FormatException fex = new FormatException("Null参照による例外です。", nrex);
            // データ層によるエラーとして例外をマスク
            ApplicationException appex = new ApplicationException("データ層でエラーが発生しました。", fex);
            // 例外情報を整形する際にさらに別の例外が発生
            FormatException fmtEx = new FormatException("エラーメッセージの整形中にエラーが発生しました。", appex);

            // LINQを使用して一般的あるいは既知の例外を検索
            var query = from ex in fmtEx.GetNestedExceptionList()
                        where ex is NullReferenceException ||
                              ex is OutOfMemoryException ||
                              ex is ThreadAbortException
                        select ex;
            // 見つかった例外情報をレポート
            foreach (Exception ex in query)
            {
                Console.WriteLine("一般的な例外({0})が見つかりました。メッセージ：{1}",
                     ex.GetType(), ex.Message);
            }
        }

        public static IEnumerable<Exception> GetNestedExceptionList(this Exception exception)
        {
            Exception current = exception;
            do
            {
                current = current.InnerException;
                if (current != null)
                    yield return current;
            }
            while (current != null);
        }
        #endregion
    }
}
