﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Runtime.Remoting.Messaging;
using System.Runtime.Serialization;
using System.Security.Permissions;

namespace CSharpRecipes
{
    public static class DelegatesEventsLambdaExpressions
    {
        #region "6.0 概要"
        // デリゲートの定義
        public delegate int IncreaseByANumber(int j);
        // 積算用のデリゲートの定義
        public delegate int MultipleIncreaseByANumber(int j, int k, int l);

        // デリゲートに登録するメソッドです
        static public int MultiplyByANumber(int j)
        {
            return j * 42;
        }

        public static void ExecuteCSharp1_0()
        {
            // デリゲートのインスタンスを作成します
            IncreaseByANumber increase =
               new IncreaseByANumber(MultiplyByANumber);

            // 登録したメソッドを呼び出すことにより、計算結果 420 が表示されます
            Console.WriteLine(increase(10));
        }

        public static void ExecuteCSharp2_0()
        {
            // デリゲートのインスタンスを作成します
            IncreaseByANumber increase =
               new IncreaseByANumber(
                delegate(int j)
                {
                    return j * 42;
                });

            // 登録したメソッドを呼び出すことにより、計算結果 420 が表示されます
            Console.WriteLine(increase(10));
        }

        public static void ExecuteCSharp3_0()
        {
            // ラムダ式を宣言します
            IncreaseByANumber increase = j => j * 42;
            // 登録したメソッドを呼び出すことにより、計算結果 420 が表示されます
            Console.WriteLine(increase(10));

            MultipleIncreaseByANumber multiple = (j, k, l) => ((j * 42) / k) % l;
            Console.WriteLine(multiple(10, 11, 12));
        }
        #endregion

        #region "6.1 マルチキャストデリゲートに含まれるデリゲートを呼び出すタイミングや条件を制御する"

        public static void InvokeInReverse()
        {
            Func<int> myDelegateInstance1 = TestInvokeIntReturn.Method1;
            Func<int> myDelegateInstance2 = TestInvokeIntReturn.Method2;
            Func<int> myDelegateInstance3 = TestInvokeIntReturn.Method3;

            Func<int> allInstances =
                    myDelegateInstance1 +
                    myDelegateInstance2 +
                    myDelegateInstance3;

            Console.WriteLine("逆順でデリゲートを呼び出します");
            Delegate[] delegateList = allInstances.GetInvocationList();
            foreach (Func<int> instance in delegateList.Reverse())
            {
                instance();
            }
        }

        public static void InvokeEveryOtherOperation()
        {
            Func<int> myDelegateInstance1 = TestInvokeIntReturn.Method1;
            Func<int> myDelegateInstance2 = TestInvokeIntReturn.Method2;
            Func<int> myDelegateInstance3 = TestInvokeIntReturn.Method3;

            Func<int> allInstances = //myDelegateInstance1;
                    myDelegateInstance1 +
                    myDelegateInstance2 +
                    myDelegateInstance3;

            Delegate[] delegateList = allInstances.GetInvocationList();
            Console.WriteLine("デリゲートを1つおきに呼び出します");
            foreach (Func<int> instance in delegateList.EveryOther())
            {
                // デリゲートを呼び出します
                int retVal = instance();
                Console.WriteLine("デリゲートの返り値 = " + retVal);
            }
        }

        static IEnumerable<T> EveryOther<T>(this IEnumerable<T> enumerable)
        {
            bool retNext = true;
            foreach (T t in enumerable)
            {
                if (retNext) yield return t;
                retNext = !retNext;
            }
        }

        public class TestInvokeIntReturn
        {
            public static int Method1()
            {
                Console.WriteLine("Method1 が呼び出されました");
                return 1;
            }

            public static int Method2()
            {
                Console.WriteLine("Method2 が呼び出されました");
                return 2;
            }

            public static int Method3()
            {
                //throw (new Exception("Method1"));
                //throw (new SecurityException("Method3"));
                Console.WriteLine("Method3 が呼び出されました");
                return 3;
            }
        }

        public static void InvokeWithTest()
        {
            Func<bool> myDelegateInstanceBool1 = TestInvokeBoolReturn.Method1;
            Func<bool> myDelegateInstanceBool2 = TestInvokeBoolReturn.Method2;
            Func<bool> myDelegateInstanceBool3 = TestInvokeBoolReturn.Method3;

            Func<bool> allInstancesBool =
                    myDelegateInstanceBool1 +
                    myDelegateInstanceBool2 +
                    myDelegateInstanceBool3;

            Console.WriteLine(
                "デリゲートを個別に呼び出します(ただし先に呼び出されたメソッドの結果に依存します)：");
            foreach (Func<bool> instance in allInstancesBool.GetInvocationList())
            {
                if (!instance())
                    break;
            }
        }

        public class TestInvokeBoolReturn
        {
            public static bool Method1()
            {
                Console.WriteLine("Method1 が呼び出されました");
                return true;
            }

            public static bool Method2()
            {
                Console.WriteLine("Method2 が呼び出されました");
                return false;
            }

            public static bool Method3()
            {
                Console.WriteLine("Method3 が呼び出されました");
                return true;
            }
        }
        #endregion

        #region "6.2 マルチキャストデリゲート中の各デリゲートの返り値を取得する"
        public static void TestIndividualInvokesReturnValue()
        {
            Func<int> myDelegateInstance1 = TestInvokeIntReturn.Method1;
            Func<int> myDelegateInstance2 = TestInvokeIntReturn.Method2;
            Func<int> myDelegateInstance3 = TestInvokeIntReturn.Method3;

            Func<int> allInstances =
                    myDelegateInstance1 +
                    myDelegateInstance2 +
                    myDelegateInstance3;

            Console.WriteLine("デリゲートを個別に呼び出します(それぞれの返り値を取得します)：");
            foreach (Func<int> instance in allInstances.GetInvocationList())
            {
                int retVal = instance();
                Console.WriteLine("\t結果：" + retVal);
            }
        }
        #endregion

        #region "6.3 マルチキャストデリゲート中の各デリゲートの呼び出し時に発生した例外を個別に処理する"
        [Serializable]
        public class MulticastInvocationException : Exception
        {
            private List<Exception> _invocationExceptions;

            public MulticastInvocationException()
                : base()
            {
            }

            public MulticastInvocationException(IEnumerable<Exception> invocationExceptions)
            {
                _invocationExceptions = new List<Exception>(invocationExceptions);
            }

            public MulticastInvocationException(string message)
                : base(message)
            {
            }

            public MulticastInvocationException(string message, Exception innerException) :
                base(message, innerException)
            {
            }

            protected MulticastInvocationException(SerializationInfo info, StreamingContext context) :
                base(info, context)
            {
                _invocationExceptions =
                    (List<Exception>)info.GetValue("InvocationExceptions",
                        typeof(List<Exception>));
            }

            [SecurityPermissionAttribute(SecurityAction.Demand, SerializationFormatter = true)]
            public override void GetObjectData(
               SerializationInfo info, StreamingContext context)
            {
                info.AddValue("InvocationExceptions", this.InvocationExceptions);
                base.GetObjectData(info, context);
            }

            public ReadOnlyCollection<Exception> InvocationExceptions
            {
                get { return new ReadOnlyCollection<Exception>(_invocationExceptions); }
            }
        }

        public static void TestIndividualInvokesExceptions()
        {
            Func<int> myDelegateInstance1 = TestInvokeIntReturn.Method1;
            Func<int> myDelegateInstance2 = TestInvokeIntReturn.Method2;
            Func<int> myDelegateInstance3 = TestInvokeIntReturn.Method3;

            Func<int> allInstances =
                    myDelegateInstance1 +
                    myDelegateInstance2 +
                    myDelegateInstance3;

            Console.WriteLine("デリゲートを個別に呼び出します(例外処理あり)：");

            // 各デリゲートの呼び出し時に発生したすべての例外を保持するため
            // Exceptionオブジェクトのリストを用意します
            List<Exception> invocationExceptions = new List<Exception>();

            foreach (Func<int> instance in allInstances.GetInvocationList())
            {
                try
                {
                    int retVal = instance();
                    Console.WriteLine("\t結果：" + retVal);
                }
                catch (Exception ex)
                {
                    // 例外を表示してログに記録した後、呼び出しを継続します
                    Console.WriteLine(ex.ToString());
                    EventLog myLog = new EventLog();
                    myLog.Source = "MyApplicationSource";
                    myLog.WriteEntry("メソッド " +
                        instance.Method.Name + " の呼び出しに失敗しました。エラー：" +
                        ex.ToString(),
                        EventLogEntryType.Error);
                    // 例外をリストに追加します
                    invocationExceptions.Add(ex);
                }
            }
            // 1つ以上の例外が発生した場合、発生したすべての例外を含むような
            // 独自の例外をスローします
            if (invocationExceptions.Count > 0)
            {
                throw new MulticastInvocationException(invocationExceptions);
            }
        }
        #endregion

        #region "6.4 デリゲートの同期呼び出しを非同期呼び出しに変更する"

        public delegate void SyncDelegateTypeSimple();
        public delegate int SyncDelegateType(string message);

        public static void TestSimpleSyncDelegate()
        {
            SyncDelegateTypeSimple sdtsInstance = TestSyncDelegateTypeSimple.Method1;
            sdtsInstance();
        }

        public static void TestSimpleAsyncDelegate()
        {
            AsyncCallback callBack = new AsyncCallback(DelegateSimpleCallback);

            SyncDelegateTypeSimple sdtsInstance = TestSyncDelegateTypeSimple.Method1;

            IAsyncResult asyncResult =
                sdtsInstance.BeginInvoke(callBack, null);

            Console.WriteLine("実行中...");
        }

        public static void TestComplexSyncDelegate()
        {
            SyncDelegateType sdtInstance = TestSyncDelegateType.Method1;

            int retVal = sdtInstance("同期呼び出し");

            Console.WriteLine("同期呼び出し：" + retVal);
        }

        public static void TestCallbackAsyncDelegate()
        {
            AsyncCallback callBack =
                new AsyncCallback(DelegateCallback);

            SyncDelegateType sdtInstance = TestSyncDelegateType.Method1;

            IAsyncResult asyncResult =
                sdtInstance.BeginInvoke("非同期呼び出し", callBack, null);

            Console.WriteLine("実行中...");
        }

        // TestSyncDelegateTypeSimple.Method1 が終了した時点で呼び出される
        // コールバックです。
        private static void DelegateSimpleCallback(IAsyncResult iResult)
        {
            AsyncResult result = (AsyncResult)iResult;
            SyncDelegateTypeSimple sdtsInstance =
                (SyncDelegateTypeSimple)result.AsyncDelegate;

            sdtsInstance.EndInvoke(result);
            Console.WriteLine("シンプルコールバックが実行されました");
        }

        // TestSyncDelegateType.Method1 が終了した時点で呼び出される
        // コールバックです。
        private static void DelegateCallback(IAsyncResult iResult)
        {
            AsyncResult result = (AsyncResult)iResult;
            SyncDelegateType sdtInstance =
                (SyncDelegateType)result.AsyncDelegate;

            int retVal = sdtInstance.EndInvoke(result);
            Console.WriteLine("返り値(コールバック中で参照)：" + retVal);
        }

        // SyncDelegateTypeSimpleデリゲート経由で呼び出されるメソッドを含んだクラスです。
        public class TestSyncDelegateTypeSimple
        {
            public static void Method1()
            {
                Console.WriteLine("Method1 が呼び出されました");
            }
        }

        public class TestSyncDelegateType
        {
            public static int Method1(string message)
            {
                Console.WriteLine("Method1 がメッセージを指定して呼び出されました：" + message);
                return 1;
            }
        }
        #endregion

        #region "6.5 インターフェイス検索メカニズムの詳細"
        public static void FindSpecificInterfaces()
        {
            // 検索対象のインターフェイスを設定します
            Type[] interfaces = {
                typeof(System.ICloneable),
                typeof(System.Collections.ICollection),
                typeof(System.IAppDomainSetup) };

            // 検索対象の型を設定します
            Type searchType = typeof(System.Collections.ArrayList);

            var matches = from t in searchType.GetInterfaces()
                          join s in interfaces on t equals s
                          select s;

            Console.WriteLine("条件に一致したインターフェイス：");
            foreach (Type match in matches)
            {
                Console.WriteLine(match.ToString());
            }


            // 特定の名前空間(たとえばSystem.Collections)に定義されたインターフェイスが
            // 実装されているかどうかを判別するためのフィルタを作成します
            var collectionsInterfaces = from type in searchType.GetInterfaces()
                                        where type.Namespace == "System.Collections"
                                        select type;
            foreach (Type t in collectionsInterfaces)
            {
                Console.WriteLine("System.Collections に実装されたインターフェイス：" + t);
            }

            // Int32を返り値とするメソッドAddを持つすべてのインターフェイスを
            // 取得するためのフィルタを作成します
            var addInterfaces = from type in searchType.GetInterfaces()
                                from method in type.GetMethods()
                                where (method.Name == "Add") &&
                                      (method.ReturnType == typeof(int))
                                select type;
            foreach (Type t in addInterfaces)
            {
                Console.WriteLine("int Add()メソッドを持つインターフェイス：" + t);
            }

            // グローバルアセンブリキャッシュ(GAC: Global Assembly Cache)からロードされたすべてのインターフェイスを
            // 取得するためのフィルタを作成します
            var gacInterfaces = from type in searchType.GetInterfaces()
                                where type.Assembly.GlobalAssemblyCache
                                select type;
            foreach (Type t in gacInterfaces)
            {
                Console.WriteLine("GACからロードされたインターフェイス：" + t);
            }

            // バージョンが2.0.0.0のアセンブリに定義されたすべてのインターフェイスを
            // 取得するためのフィルタを作成します
            var versionInterfaces = from type in searchType.GetInterfaces()
                                    where type.Assembly.GetName().Version == new Version(2, 0, 0, 0)
                                    select type;
            foreach (Type t in versionInterfaces)
            {
                Console.WriteLine("Implemented interface from assembly with version 2.0.0.0: " + t);
            }
        }

        #endregion

        #region "6.6 ディクショナリに対する追加や変更の監視"
        public static void TestObserverPattern()
        {
            Dictionary<int, string> dictionary1 = new Dictionary<int, string>();
            Dictionary<int, string> dictionary2 = new Dictionary<int, string>();
            Dictionary<int, string> dictionary3 = new Dictionary<int, string>();

            // 3つのObservableDictionaryのインスタンスを作成します
            var obsDict1 = dictionary1.MakeObservableDictionary();
            var obsDict2 = dictionary2.MakeObservableDictionary();
            var obsDict3 = dictionary3.MakeObservableDictionary();

            // 3つのオブジェクトを監視するためのオブジェクトを作成します
            var observer = new ObservableDictionaryObserver<int, string>();

            // 3つの監視対象を監視用オブジェクトに登録します
            observer.Register(obsDict1);
            observer.Register(obsDict2);
            observer.Register(obsDict3);

            // AddingおよびChangingイベントで承認を行うためのメソッドを登録します
            observer.ApproveAdd +=
                new ObservableDictionaryObserver<int, string>.
                    Approval(SeekApproval);
            observer.ApproveChange +=
                new ObservableDictionaryObserver<int, string>.
                    Approval(SeekApproval);

            // 監視中のオブジェクトを使用します
            obsDict1.Add(1, "one");
            obsDict2.Add(2, "two");
            obsDict3.Add(3, "three");

            // 承認処理が機能していることを確認します
            Debug.Assert(obsDict1.Count == 1);
            Debug.Assert(obsDict2.Count == 1);
            // 3文字以上の値が入力されたため値は登録されません
            Debug.Assert(obsDict3.Count == 0);

            // 監視対象のオブジェクトの登録を解除します
            observer.Unregister(obsDict3);
            observer.Unregister(obsDict2);
            observer.Unregister(obsDict1);

            ///////////////////////////////////////////////////////////////
            // 以下は異なるDictionary型のデータの処理です
            ///////////////////////////////////////////////////////////////
            // 監視対象となる2つのSortedListのインスタンスを作成します
            SortedList<string, bool> sortedList1 = new SortedList<string, bool>();
            SortedList<string, bool> sortedList2 = new SortedList<string, bool>();

            var obsSortedList1 = sortedList1.MakeObservableDictionary();
            var obsSortedList2 = sortedList2.MakeObservableDictionary();

            // 2つのオブジェクトを監視するためのオブジェクトを作成します
            ObservableDictionaryObserver<string, bool> listObserver =
                new ObservableDictionaryObserver<string, bool>();

            // 2つの監視対象を監視用オブジェクトに登録します
            listObserver.Register(obsSortedList1);
            listObserver.Register(obsSortedList2);

            // AddingおよびChangingイベントで承認を行うためのメソッドを登録します
            listObserver.ApproveAdd +=
                new ObservableDictionaryObserver<string, bool>.
                    Approval(ApprovePositive);
            listObserver.ApproveChange +=
                new ObservableDictionaryObserver<string, bool>.
                    Approval(ApprovePositive);

            // 監視中のオブジェクトを使用します
            obsSortedList1.Add("Item 1", true);
            obsSortedList2.Add("Item 2", false);

            // 承認処理が機能していることを確認します
            Debug.Assert(obsSortedList1.Count == 1);
            // trueだけが登録されるため、2番目のリストは空になります
            Debug.Assert(obsSortedList2.Count == 0);

            // 監視対象のオブジェクトの登録を解除します
            listObserver.Unregister(obsSortedList2);
            listObserver.Unregister(obsSortedList1);
        }

        static bool SeekApproval(object sender,
                ObservableDictionaryEventArgs<int, string> args)
        {
            // 3文字以下の値のみDictionaryに登録できるよう制限します
            string value = args.Value.ToString();
            if (value.Length <= 3)
                return true;
            return false;
        }

        static bool ApprovePositive(object sender,
                ObservableDictionaryEventArgs<string, bool> args)
        {
            // 値がtrueの場合のみ登録できるよう制限します
            return args.Value;
        }


        // 登録されたObservableDictionaryオブジェクトを監視するオブジェクトを表します
        public class ObservableDictionaryObserver<TKey, TValue>
        {
            public ObservableDictionaryObserver() { }

            // 監視対象のオブジェクトへの追加や削除を承認するためのデリゲートおよびイベントを設定します
            public delegate bool Approval(object sender,
                    ObservableDictionaryEventArgs<TKey, TValue> e);

            public Approval ApproveAdd { get; set; }
            public Approval ApproveChange { get; set; }

            public void Register(ObservableDictionary<TKey, TValue> dictionary)
            {
                // ObservableDictionaryインスタンスのイベントにハンドラを登録します
                dictionary.AddingEntry +=
                    new EventHandler<ObservableDictionaryEventArgs<TKey, TValue>>(OnAddingListener);
                dictionary.AddedEntry +=
                    new EventHandler<ObservableDictionaryEventArgs<TKey, TValue>>(OnAddedListener);
                dictionary.ChangingEntry +=
                    new EventHandler<ObservableDictionaryEventArgs<TKey, TValue>>(OnChangingListener);
                dictionary.ChangedEntry +=
                    new EventHandler<ObservableDictionaryEventArgs<TKey, TValue>>(OnChangedListener);
            }

            public void Unregister(ObservableDictionary<TKey, TValue> dictionary)
            {
                // ObservableDictionaryインスタンスのイベントへの登録を解除します
                dictionary.AddingEntry -=
                    new EventHandler<ObservableDictionaryEventArgs<TKey, TValue>>(OnAddingListener);
                dictionary.AddedEntry -=
                    new EventHandler<ObservableDictionaryEventArgs<TKey, TValue>>(OnAddedListener);
                dictionary.ChangingEntry -=
                    new EventHandler<ObservableDictionaryEventArgs<TKey, TValue>>(OnChangingListener);
                dictionary.ChangedEntry -=
                    new EventHandler<ObservableDictionaryEventArgs<TKey, TValue>>(OnChangedListener);
            }

            private void CheckApproval(Approval approval,
                    ObservableDictionaryEventArgs<TKey, TValue> args)
            {
                // すべての監視オブジェクトが承認したかどうかを確認します
                foreach (Approval approvalInstance in
                                approval.GetInvocationList())
                {
                    if (!approvalInstance(this, args))
                    {
                        // 1つ以上の監視オブジェクトが却下を選択した場合、
                        // 追加または変更は却下されます。既定では承認されます。
                        args.KeepChanges = false;
                        break;
                    }
                }
            }

            private void OnAddingListener(object sender,
                        ObservableDictionaryEventArgs<TKey, TValue> args)
            {
                // 監視中のオブジェクトの有無を確認します
                if (ApproveAdd != null)
                {
                    CheckApproval(ApproveAdd, args);
                }

                Debug.WriteLine("[NOTIFY] 追加前...: 追加の承認 = " +
                                    args.KeepChanges.ToString());
            }

            private void OnAddedListener(object sender,
                    ObservableDictionaryEventArgs<TKey, TValue> args)
            {
                Debug.WriteLine("[NOTIFY] ...追加後: 追加された要素：" +
                                    args.ToString());
            }

            private void OnChangingListener(object sender,
                    ObservableDictionaryEventArgs<TKey, TValue> args)
            {
                // 監視中のオブジェクトの有無を確認します
                if (ApproveChange != null)
                {
                    CheckApproval(ApproveChange, args);
                }

                Debug.WriteLine("[NOTIFY] 変更前...: 変更の承認 = " +
                                    args.KeepChanges.ToString());
            }

            private void OnChangedListener(object sender,
                    ObservableDictionaryEventArgs<TKey, TValue> args)
            {
                Debug.WriteLine("[NOTIFY] ...変更後: 変更された要素：" +
                                    args.ToString());
            }
        }

        public class ObservableDictionary<TKey, TValue> : IDictionary<TKey, TValue>
        {
            IDictionary<TKey, TValue> _internalDictionary;
            public ObservableDictionary(IDictionary<TKey, TValue> dictionary)
            {
                if (dictionary == null)
                    throw new ArgumentNullException("dictionary");
                _internalDictionary = dictionary;
            }

            #region イベントおよびイベントの初期設定

            public event EventHandler<ObservableDictionaryEventArgs<TKey, TValue>> AddingEntry;
            public event EventHandler<ObservableDictionaryEventArgs<TKey, TValue>> AddedEntry;
            public event EventHandler<ObservableDictionaryEventArgs<TKey, TValue>> ChangingEntry;
            public event EventHandler<ObservableDictionaryEventArgs<TKey, TValue>> ChangedEntry;

            protected virtual bool OnAdding(ObservableDictionaryEventArgs<TKey, TValue> e)
            {
                if (AddingEntry != null)
                {
                    AddingEntry(this, e);
                    return e.KeepChanges;
                }

                return true;
            }

            protected virtual void OnAdded(ObservableDictionaryEventArgs<TKey, TValue> e)
            {
                if (AddedEntry != null)
                {
                    AddedEntry(this, e);
                }
            }

            protected virtual bool OnChanging(ObservableDictionaryEventArgs<TKey, TValue> e)
            {
                if (ChangingEntry != null)
                {
                    ChangingEntry(this, e);
                    return e.KeepChanges;
                }

                return true;
            }

            protected virtual void OnChanged(ObservableDictionaryEventArgs<TKey, TValue> e)
            {
                if (ChangedEntry != null)
                {
                    ChangedEntry(this, e);
                }
            }
            #endregion // イベントおよびイベントの初期設定

            #region インターフェイスの実装
            #region IDictionary<TKey,TValue> Members

            public ICollection<TValue> Values
            {
                get { return _internalDictionary.Values; }
            }

            public ICollection<TKey> Keys
            {
                get { return _internalDictionary.Keys; }
            }

            public TValue this[TKey key]
            {
                get
                {
                    TValue value;
                    if (_internalDictionary.TryGetValue(key, out value))
                        return value;
                    else
                    {
                        return default(TValue);
                    }
                }
                set
                {
                    // 現在のキーが既に登録済みであれば変更中、そうでなければ追加中を表します
                    if (_internalDictionary.ContainsKey(key))
                    {
                        ObservableDictionaryEventArgs<TKey, TValue> args =
                            new ObservableDictionaryEventArgs<TKey, TValue>(key, value);

                        if (OnChanging(args))
                        {
                            _internalDictionary[key] = value;
                        }
                        else
                        {
                            Debug.WriteLine("値の変更は中断されました");
                        }

                        OnChanged(args);
                    }
                    else
                    {
                        Debug.WriteLine("要素が見つからなかったため追加します");
                        _internalDictionary.Add(key, value);
                    }
                }
            }

            public void Add(TKey key, TValue value)
            {
                ObservableDictionaryEventArgs<TKey, TValue> args =
                    new ObservableDictionaryEventArgs<TKey, TValue>(key, value);
                if (OnAdding(args))
                {
                    this._internalDictionary.Add(key, value);
                }
                else
                {
                    Debug.WriteLine("キー値ペアの追加は中断されました");
                }

                OnAdded(args);
            }

            public bool ContainsKey(TKey key)
            {
                return _internalDictionary.ContainsKey(key);
            }

            public bool Remove(TKey key)
            {
                return _internalDictionary.Remove(key);
            }

            public bool TryGetValue(TKey key, out TValue value)
            {
                return _internalDictionary.TryGetValue(key, out value);
            }

            #endregion

            #region ICollection<KeyValuePair<TKey,TValue>> メンバ

            public void Add(KeyValuePair<TKey, TValue> item)
            {
                _internalDictionary.Add(item.Key, item.Value);
            }

            public void Clear()
            {
                _internalDictionary.Clear();
            }

            public bool Contains(KeyValuePair<TKey, TValue> item)
            {
                return _internalDictionary.Contains(item);
            }

            public void CopyTo(KeyValuePair<TKey, TValue>[] array, int arrayIndex)
            {
                _internalDictionary.CopyTo(array, arrayIndex);
            }

            public int Count
            {
                get { return _internalDictionary.Count; }
            }

            public bool IsReadOnly
            {
                get { return _internalDictionary.IsReadOnly; }
            }

            public bool Remove(KeyValuePair<TKey, TValue> item)
            {
                return _internalDictionary.Remove(item);
            }

            #endregion

            #region IEnumerable<KeyValuePair<TKey,TValue>> メンバ

            public IEnumerator<KeyValuePair<TKey, TValue>> GetEnumerator()
            {
                return _internalDictionary.GetEnumerator();
            }

            #endregion

            #region IEnumerable メンバ

            IEnumerator IEnumerable.GetEnumerator()
            {
                return _internalDictionary.GetEnumerator();
            }

            #endregion
            #endregion // インターフェイスの実装
        }

        public static ObservableDictionary<TKey, TValue> MakeObservableDictionary<TKey, TValue>(
            this IDictionary<TKey, TValue> dictionary)
        {
            return new ObservableDictionary<TKey, TValue>(dictionary);
        }

        public class ObservableDictionaryEventArgs<TKey, TValue> : EventArgs
        {
            TKey _key;
            TValue _value;

            public ObservableDictionaryEventArgs(TKey key, TValue value)
            {
                _key = key;
                _value = value;
                this.KeepChanges = true;
            }

            public bool KeepChanges { get; set; }
            public TKey Key { get { return _key; } }
            public TValue Value { get { return _value; } }
        }

        #endregion

        #region "6.7 ラムダ式の使用"
        public static void TestUsingLambdaExpressions()
        {
            OldWay ow = new OldWay();
            ow.WorkItOut();

            LambdaWay iw = new LambdaWay();
            iw.WorkItOut();

            DirectAssignmentWay diw = new DirectAssignmentWay();
            diw.WorkItOut();

            GenericWay gw = new GenericWay();
            gw.WorkItOut();

            GenericEventConsumer gec = new GenericEventConsumer();
            gec.Test();

            OuterVars ov = new OuterVars();
            ov.SeeOuterWork();
        }

        class OuterVars
        {
            public void SeeOuterWork()
            {
                int count = 0;
                int total = 0;
                Func<int> countUp = () => count++;
                for (int i = 0; i < 10; i++)
                {
                    total += countUp();
                }
                Debug.WriteLine("総和 = " + total);
            }
        }

        class OldWay
        {
            // デリゲートを定義します
            delegate int DoWork(string work);

            // デリゲートのインスタンスを作成して呼び出すメソッドを定義します
            public void WorkItOut()
            {
                // インスタンスを宣言します
                DoWork dw = new DoWork(DoWorkMethodImpl);
                // デリゲートを呼び出します
                int i = dw("以前の方法で実行します");
            }

            // デリゲートのシグネチャに一致するメソッドを定義して、
            // デリゲートの呼び出し時に実行されるようデリゲートに登録します
            public int DoWorkMethodImpl(string s)
            {
                Console.WriteLine(s);
                return s.GetHashCode();
            }
        }

        class LambdaWay
        {
            // デリゲートを定義します
            delegate int DoWork(string work);

            // デリゲートのインスタンスを作成して呼び出すメソッドを定義します
            public void WorkItOut()
            {
                // インスタンスを宣言します
                DoWork dw = s =>
                {
                    Console.WriteLine(s);
                    return s.GetHashCode();
                };
                // デリゲートを呼び出します
                int i = dw("同じ処理をインラインで定義できます");
            }
        }

        class DirectAssignmentWay
        {
            // デリゲートを定義します
            delegate int DoWork(string work);

            // デリゲートのインスタンスを作成して呼び出すメソッドを定義します
            public void WorkItOut()
            {
                // インスタンスを宣言してメソッドを割り当てます
                DoWork dw = DoWorkMethodImpl;
                // invoke delegate
                int i = dw("メソッドを直接割り当てます");
            }
            // デリゲートのシグネチャに一致するメソッドを定義して、
            // デリゲートの呼び出し時に実行されるようデリゲートに登録します
            public int DoWorkMethodImpl(string s)
            {
                Console.WriteLine(s);
                return s.GetHashCode();
            }
        }

        class GenericWay
        {
            // 2つのデリゲートのインスタンスを作成してそれぞれを呼び出すメソッドです
            public void WorkItOut()
            {
                Func<string, string> dwString = s =>
                {
                    Console.WriteLine(s);
                    return s;
                };

                // 文字列用のデリゲートを呼び出します
                string retStr = dwString("ジェネリックデリゲートとして実行します");

                Func<int, int> dwInt = i =>
                {
                    Console.WriteLine(i);
                    return i;
                };

                // int用のデリゲートを呼び出します
                int j = dwInt(5);
            }
        }

        public class GenericEventArgs<T> : EventArgs
        {
            public GenericEventArgs(T value)
            {
                this.Value = value;
            }

            public T Value { get; set; }
        }

        public class GenericEvent
        {
            // ジェネリックイベントを定義します
            public event EventHandler<GenericEventArgs<string>> DoingStringWork;
            public event EventHandler<GenericEventArgs<int>> DoingIntWork;

            public void WorkItOut()
            {
                DoingStringWork(this, new GenericEventArgs<string>("文字列処理"));
                DoingIntWork(this, new GenericEventArgs<int>(5));
            }
        }

        public class GenericEventConsumer
        {
            public void Test()
            {
                GenericEvent ge = new GenericEvent();
                ge.DoingStringWork += new EventHandler<GenericEventArgs<string>>(ge_DoingStringWork);
                ge.DoingIntWork += new EventHandler<GenericEventArgs<int>>(ge_DoingIntWork);
            }
            void ge_DoingIntWork(object sender, GenericEventArgs<int> workArgs)
            {
                throw new NotImplementedException();
            }
            void ge_DoingStringWork(object sender, GenericEventArgs<string> workArgs)
            {
                throw new NotImplementedException();
            }
        }

        #endregion

        #region "6.8 イベントハンドラを簡潔に設定する"
        public class IWantToKnowNow
        {
            public static void Test()
            {
                eNewspaper DailyBitFlash = new eNewspaper();
                DailyBitFlash.NewsEvent +=
                           new EventHandler<NewsEventArgs>(BreakingNews);
                DailyBitFlash.SubscriptionExpired +=
                           new EventHandler<ExpiredEventArgs>(Expired);

                // send breaking news
                DailyBitFlash.TransmitBreakingNews("Patriots 勝利！");
                DailyBitFlash.TransmitBreakingNews("W は4年目に突入");
                DailyBitFlash.TransmitBreakingNews("VS2008と.NET 3.5のイベントがLAで開催");

                // offer expired subscriber great deal
                string offer = "購読期間が終了しました！";
                offer += "今なら19.95ドルで98週間購読できます！";
                DailyBitFlash.NotifySubscriptionExpired(offer);

                IWantToKnowThen.TryMe();
            }

            private static void BreakingNews(object src, NewsEventArgs nea)
            {
                Console.WriteLine(nea.LatestNews);
            }

            private static void Expired(object src, ExpiredEventArgs eea)
            {
                Console.WriteLine(eea.NewSubscriptionOffer);
            }
        }

        public class NewsEventArgs : EventArgs
        {
            private string _latestNews;

            public NewsEventArgs(string latestNews)
            {
                _latestNews = latestNews;
            }
            public string LatestNews
            {
                get { return _latestNews; }
            }
        }

        public class ExpiredEventArgs : EventArgs
        {
            private string _offer;

            public ExpiredEventArgs(string newOffer)
            {
                _offer = newOffer;
            }
            public string NewSubscriptionOffer
            {
                get { return _offer; }
            }
        }

        public class eNewspaper
        {
            public event EventHandler<NewsEventArgs> NewsEvent;

            public void TransmitBreakingNews(string news)
            {
                // スレッドセーフのために一時変数へコピーします
                EventHandler<NewsEventArgs> breakingNews = NewsEvent;
                if (breakingNews != null)
                    breakingNews(this, new NewsEventArgs(news));
            }

            // 定期購読の期限切れを通知できるようにします
            public event EventHandler<ExpiredEventArgs> SubscriptionExpired;

            public void NotifySubscriptionExpired(string newOffer)
            {
                EventHandler<ExpiredEventArgs> subscriptionExpired =
                    SubscriptionExpired;
                if (subscriptionExpired != null)
                    subscriptionExpired(this, new ExpiredEventArgs(newOffer));
            }
        }

        public class IWantToKnowThen
        {
            public static void TryMe()
            {
                OldNewspaper DailyPaperFlash = new OldNewspaper();
                DailyPaperFlash.NewsEvent +=
                    new OldNewspaper.NewsEventHandler(StaleNews);
                DailyPaperFlash.SubscriptionExpired +=
                    new OldNewspaper.SubscriptionExpiredEventHandler(Expired);

                // ニュースを送信します
                DailyPaperFlash.TransmitStaleNews("Patriots スーパーボウル3回目の優勝！");
                DailyPaperFlash.TransmitStaleNews("W が再度就任。");
                DailyPaperFlash.TransmitStaleNews("VS2005はまもなく時代遅れに。");

                // 定期購読の期限切れを通知します
                string offer = "購読期間が終了しました！";
                offer += "今なら19.95ドルで24週間購読できます！";
                DailyPaperFlash.NotifySubscriptionExpired(offer);
            }

            private static void StaleNews(object src, NewsEventArgs nea)
            {
                Console.WriteLine(nea.LatestNews);
            }

            private static void Expired(object src, OldExpiredEventArgs eea)
            {
                Console.WriteLine(eea.NewSubscriptionOffer);
            }
        }

        // ニュースのデータを保持するためのEventArgs派生クラスです
        public class OldExpiredEventArgs : EventArgs
        {
            private string _offer;

            public OldExpiredEventArgs(string newOffer)
            {
                _offer = newOffer;
            }
            public string NewSubscriptionOffer
            {
                get { return _offer; }
            }
        }

        public class OldNewspaper
        {
            // 顧客がニュースを購読できるようにします
            public delegate void NewsEventHandler(Object sender, NewsEventArgs e);
            public event NewsEventHandler NewsEvent;

            // 顧客にニュースを伝達するためのラッパメソッドです
            public void TransmitStaleNews(string news)
            {
                // スレッドセーフのために一時変数へコピーします
                NewsEventHandler newsEvent = NewsEvent;
                if (newsEvent != null)
                    newsEvent(this, new NewsEventArgs(news));
            }

            // 定期購読の期限切れを通知できるようにします
            public delegate void SubscriptionExpiredEventHandler(Object sender, OldExpiredEventArgs e);
            public event SubscriptionExpiredEventHandler SubscriptionExpired;

            public void NotifySubscriptionExpired(string newOffer)
            {
                SubscriptionExpiredEventHandler subscriptionExpired =
                    SubscriptionExpired;
                if (subscriptionExpired != null)
                    subscriptionExpired(this, new OldExpiredEventArgs(newOffer));
            }
        }
        #endregion

        #region "6.9 異なるパラメータ修飾子を持つラムダ式を使用する"
        public static void TestParameterModifiers()
        {
            ParameterMods pm = new ParameterMods();
            pm.WorkItOut();

            OldParams op = new OldParams();
            op.WorkItOut();

            OuterVariablesParameterModifiers ovpm = new OuterVariablesParameterModifiers();
            ovpm.TestParams();
        }

        class ParameterMods
        {
            // outデリゲートを定義します
            delegate int DoOutWork(out string work);
            // refデリゲートを定義します
            delegate int DoRefWork(ref string work);
            // paramsデリゲートを定義します
            delegate int DoParamsWork(params object[] workItems);
            // 擬似paramsデリゲートを定義します
            delegate int DoNonParamsWork(object[] workItems);

            //デリゲートのインスタンスを作成して呼び出すためのメソッドを定義します
            public void WorkItOut()
            {
                // インスタンスを定義してメソッドを割り当てます
                DoOutWork dow = (out string s) =>
                {
                    s = "WorkFinished";
                    Console.WriteLine(s);
                    return s.GetHashCode();
                };
                // デリゲートを呼び出します
                string work;
                int i = dow(out work);
                Console.WriteLine(work);

                // インスタンスを定義してメソッドを割り当てます
                DoRefWork drw = (ref string s) =>
                {
                    Console.WriteLine(s);
                    s = "WorkFinished";
                    return s.GetHashCode();
                };
                // デリゲートを呼び出します
                work = "WorkStarted";
                i = drw(ref work);
                Console.WriteLine(work);

                // 以下のように匿名メソッドを定義した場合、コンパイルエラーCS1670「params はこのコンテキストでは有効ではありません。」が発生します
                //DoParamsWork dpw = delegate (params object[] workItems)
                // 以下のようにラムダ式を定義した場合、コンパイルエラーCS1525「'params' は無効です。」が発生します。
                //DoParamsWork dpw = (params object[] workItems) =>
                //{
                //    foreach (object o in workItems)
                //    {
                //        Console.WriteLine(o.ToString());
                //       return o.GetHashCode();
                //    }
                //};

                // paramsキーワードを省略することで解決できます
                DoParamsWork dpw = workItems =>
                {
                    foreach (object o in workItems)
                    {
                        Console.WriteLine(o.ToString());
                    }
                    return workItems.GetHashCode();
                };

                i = dpw("Hello", "42", "bar");

                // 可変長引数を指定する代わりにobject[]を指定したデリゲートを用意する場合、
                // objectの配列がコンパイラによって暗黙的に作成されないことに注意してください。
                // デリゲートを呼び出す際、以下のように明示的にobject[]として引数を指定する必要があります。
                DoNonParamsWork dnpw = (object[] items) =>
                {
                    foreach (object o in items)
                    {
                        Console.WriteLine(o.ToString());
                    }
                    if (items.Length > 0)
                        return items[0].GetHashCode();
                    else
                        return -1;
                };
                // デリゲートを呼び出します。
                i = dnpw(new object[] { "WorkItem1", 5, 65.99, true });
            }
        }

        class OldParams
        {
            // デリゲートの宣言
            delegate int DoWork(params string[] workItems);

            // デリゲートのインスタンスの生成と呼び出しを行うメソッド
            public void WorkItOut()
            {
                // インスタンスを作成します
                DoWork dw = new DoWork(DoWorkMethodImpl);
                string[] items = new string[3];
                items[0] = "item 0";
                items[1] = "item 1";
                items[2] = "item 2";
                // デリゲートを呼び出します
                int i = dw(items);

                items = new string[1];
                items[0] = "item 0";
                // デリゲートを呼び出します
                i = dw(items);

            }

            // デリゲートのシグネチャに一致するメソッドで、デリゲートに登録されることによって
            // デリゲートの呼び出し時に実行されます。
            public int DoWorkMethodImpl(params string[] items)
            {
                foreach (string s in items)
                {
                    Console.WriteLine(s);
                }
                return items.GetHashCode();
            }
        }

        class OuterVariablesParameterModifiers
        {
            // デリゲートの宣言です
            delegate int DoWork(string work);

            public void TestParams(params string[] items)
            {
                // インスタンスを宣言します
                DoWork dw = s =>
                {
                    Console.WriteLine(s);
                    foreach (string item in items)
                    {
                        Console.WriteLine(item);
                    }
                    return s.GetHashCode();
                };
                // デリゲートを呼び出します
                int i = dw("DoWorkMethodImpl1");
            }

            //public void TestOut(out string outStr)
            //{
            //    // インスタンスを宣言します
            //    DoWork dw = s =>
            //    {
            //        Console.WriteLine(s);
            //        // コンパイルエラー CS1628 が発生します: 
            //        // "ref または out パラメータ 'outStr' は、匿名メソッド、ラムダ式、
            //        // またはクエリ式の内部では使用できません
            //        //outStr = s;
            //        return s.GetHashCode();
            //    };
            //    // デリゲートを呼び出します
            //    int i = dw("DoWorkMethodImpl1");
            //}

            //public void TestRef(ref string refStr)
            //{
            //    // インスタンスを宣言します
            //    DoWork dw = s =>
            //    {
            //        Console.WriteLine(s);
            //        // コンパイルエラー CS1628 が発生します: 
            //        // "ref または out パラメータ 'outStr' は、匿名メソッド、ラムダ式、
            //        // またはクエリ式の内部では使用できません
            //        //refStr = s;
            //        return s.GetHashCode();
            //    };
            //    // デリゲートを呼び出します
            //    int i = dw("DoWorkMethodImpl1");
            //}
        }

        #endregion

        #region "6.10 C#でのクロージャの使用"
        /// <summary>
        /// 楽して儲けている人を表します...
        /// </summary>
        class SalesWeasel
        {
            #region コンストラクタ
            public SalesWeasel()
            {
            }

            public SalesWeasel(string name,
                                decimal annualQuota,
                                decimal commissionRate)
            {
                this.Name = name;
                this.AnnualQuota = annualQuota;
                this.CommissionRate = commissionRate;
            }
            #endregion //コンストラクタ

            #region private メンバ
            decimal _commission;
            #endregion private メンバ

            #region プロパティ
            public string Name { get; set; }

            public decimal AnnualQuota { get; set; }

            public decimal CommissionRate { get; set; }

            public decimal Commission
            {
                get { return _commission; }
                set
                {
                    _commission = value;
                    this.TotalCommission += _commission;
                }
            }

            public decimal TotalCommission { get; private set; }
            #endregion // プロパティ
        }

        delegate void CalculateEarnings(SalesWeasel weasel);

        static CalculateEarnings GetEarningsCalculator(decimal quarterlySales,
                                                        decimal bonusRate)
        {
            return salesWeasel =>
            {
                // 四半期における販売スタッフのノルマを計算します
                decimal quarterlyQuota = (salesWeasel.AnnualQuota / 4);
                // ノルマを満たしたかどうか？
                if (quarterlySales < quarterlyQuota)
                {
                    // ノルマを満たせなかったので手数料は0
                    salesWeasel.Commission = 0;
                }
                // 追加手数料が必要になるほど貢献したかどうか(基準はノルマの200%)
                else if (quarterlySales > (quarterlyQuota * 2.0m))
                {
                    decimal baseCommission = quarterlyQuota * salesWeasel.CommissionRate;
                    salesWeasel.Commission = (baseCommission +
                            ((quarterlySales - quarterlyQuota) *
                            (salesWeasel.CommissionRate * (1 + bonusRate))));
                }
                else // 一般的な手数料
                {
                    salesWeasel.Commission = salesWeasel.CommissionRate * quarterlySales;
                }
            };
        }


        public class QuarterlyEarning
        {
            public string Name { get; set; }
            public decimal Earnings { get; set; }
            public decimal Rate { get; set; }
        }

        public static void TestClosure()
        {
            // SalesWeaselsの初期設定を行います...
            SalesWeasel[] weasels = {
                new SalesWeasel { Name="Chas", AnnualQuota=100000m, CommissionRate=0.10m },
                new SalesWeasel { Name="Ray", AnnualQuota=200000m, CommissionRate=0.025m },
                new SalesWeasel { Name="Biff", AnnualQuota=50000m, CommissionRate=0.001m }};


            QuarterlyEarning[] quarterlyEarnings = 
                           { new QuarterlyEarning(){ Name="Q1", Earnings = 65000m, Rate = 0.1m },
                             new QuarterlyEarning(){ Name="Q2", Earnings = 20000m, Rate = 0.1m },
                             new QuarterlyEarning(){ Name="Q3", Earnings = 37000m, Rate = 0.1m },
                             new QuarterlyEarning(){ Name="Q4", Earnings = 110000m, Rate = 0.15m } };

            var calculators = from e in quarterlyEarnings
                              select new
                              {
                                  Calculator =
                                      GetEarningsCalculator(e.Earnings, e.Rate),
                                  QuarterlyEarning = e
                              };

            decimal annualEarnings = 0;
            foreach (var c in calculators)
            {
                WriteQuarterlyReport(c.QuarterlyEarning.Name,
                    c.QuarterlyEarning.Earnings, c.Calculator, weasels);
                annualEarnings += c.QuarterlyEarning.Earnings;
            }

            // 引き続き雇用するかどうかを確認...
            WriteCommissionReport(annualEarnings, weasels);
        }

        static void WriteQuarterlyReport(string quarter,
                                        decimal quarterlySales,
                                        CalculateEarnings eCalc,
                                        SalesWeasel[] weasels)
        {
            Console.WriteLine("四半期 {0} の売り上げ総額 {1}：",
                quarter, quarterlySales.ToString("C"));
            foreach (SalesWeasel weasel in weasels)
            {
                // コストを計算します
                eCalc(weasel);
                // レポートとして表示します
                Console.WriteLine("	 販売スタッフ {0} のコスト：{1}",
                    weasel.Name, weasel.Commission.ToString("C"));
            }
        }

        static void WriteCommissionReport(decimal annualEarnings,
                                        SalesWeasel[] weasels)
        {
            decimal revenueProduced = ((annualEarnings) / weasels.Length);
            Console.WriteLine();
            Console.WriteLine("年間売り上げ額 {0}",
                annualEarnings.ToString("C"));
            Console.WriteLine();
            var whoToCan = from weasel in weasels
                           select new
                           {
                               // コストが売り上げの20%を越えた場合は
                               // 解雇の対象にします
                               CanThem = (revenueProduced * 0.2m) <
                                           weasel.TotalCommission,
                               weasel.Name,
                               weasel.TotalCommission
                           };

            foreach (var weaselInfo in whoToCan)
            {
                Console.WriteLine("    {0} の売り上げ {2} に対するコストは {1}",
                    weaselInfo.Name,
                    weaselInfo.TotalCommission.ToString("C"),
                    revenueProduced.ToString("C"));
                if (weaselInfo.CanThem)
                {
                    Console.WriteLine("        {0} を解雇！", weaselInfo.Name);
                }
            }
        }
        #endregion

        #region "6.11 ファンクタを使用して一連のデータに対して複数の処理を行う"
        public static void TestFunctors()
        {
            // 以下の株式はいずれも実在しないものばかりです...
            // OU81
            // C#4VR
            // PCKD
            // BTML
            // NOVB
            // MGDCD
            // GNRCS
            // FNCTR
            // LMBDA
            // PCLS

            StockPortfolio tech = new StockPortfolio() {
                {"OU81", -10.5},
                {"C#4VR", 2.0},
                {"PCKD", 12.3},
                {"BTML", 0.5},
                {"NOVB", -35.2},
                {"MGDCD", 15.7},
                {"GNRCS", 4.0},
                {"FNCTR", 9.16},
                {"LMBDA", 9.12},
                {"PCLS", 6.11}};

            tech.PrintPortfolio("ポートフォリオの運営を開始します");
            // 下位3位の株式を売却します
            var worstPerformers = tech.GetWorstPerformers(3);
            Console.WriteLine("低パフォーマンスの株式を売却します：");
            worstPerformers.DisplayStocks();
            tech.SellStocks(worstPerformers);
            tech.PrintPortfolio("下位3位の株式を売却しました");

            //出力:
            //ポートフォリオの運営を開始します
            //  (OU81) 10.5% 減少
            //  (C#4VR) 2% 増加
            //  (PCKD) 12.3% 増加
            //  (BTML) 0.5% 増加
            //  (NOVB) 35.2% 減少
            //  (MGDCD) 15.7% 増加
            //  (GNRCS) 4% 増加
            //  (FNCTR) 9.16% 増加
            //  (LMBDA) 9.12% 増加
            //  (PCLS) 6.11% 増加
            //低パフォーマンスの株式を売却します：
            //  (NOVB) 35.2% 減少
            //  (OU81) 10.5% 減少
            //  (BTML) 0.5% 増加
            //下位3位の株式を売却しました
            //  (C#4VR) 2% 増加
            //  (PCKD) 12.3% 増加
            //  (MGDCD) 15.7% 増加
            //  (GNRCS) 4% 増加
            //  (FNCTR) 9.16% 増加
            //  (LMBDA) 9.12% 増加
            //  (PCLS) 6.11% 増加
        }

        public class Stock
        {
            public double GainLoss { get; set; }
            public string Ticker { get; set; }
        }

        public static void DisplayStocks(this IEnumerable<Stock> stocks)
        {
            var gainLoss = from stock in stocks
                           select new
                           {
                               Result = stock.GainLoss < 0 ? "減少" : "増加",
                               stock.Ticker,
                               stock.GainLoss
                           };

            foreach (var s in gainLoss)
            {
                Console.WriteLine("  ({0}) {2}% {1}", s.Ticker, s.Result,
                    System.Math.Abs(s.GainLoss));
            }
        }

        public class StockPortfolio : IEnumerable<Stock>
        {
            List<Stock> _stocks;

            public StockPortfolio()
            {
                _stocks = new List<Stock>();
            }

            public void Add(string ticker, double gainLoss)
            {
                _stocks.Add(new Stock() { Ticker = ticker, GainLoss = gainLoss });
            }

            public IEnumerable<Stock> GetWorstPerformers(int topNumber)
            {
                return _stocks.OrderBy(
                            (Stock stock) => stock.GainLoss).Take(topNumber);
            }

            public void SellStocks(IEnumerable<Stock> stocks)
            {
                foreach (Stock s in stocks)
                    _stocks.Remove(s);
            }

            public void PrintPortfolio(string title)
            {
                Console.WriteLine(title);
                _stocks.DisplayStocks();
            }

            #region IEnumerable<Stock> メンバ

            public IEnumerator<Stock> GetEnumerator()
            {
                return _stocks.GetEnumerator();
            }

            #endregion

            #region IEnumerable メンバ

            IEnumerator IEnumerable.GetEnumerator()
            {
                return this.GetEnumerator();
            }

            #endregion
        }
        #endregion
    }
}
