﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Security.Cryptography;
using System.Text;

// 7.6でオーバーロードした==演算子をテストするため、意図的に同じ変数同士を比較しています
#pragma warning disable 1718 // 同じ変数と比較されました。他の変数と比較しますか?

namespace CSharpRecipes
{
    public class DataStructsAndAlgorithms
    {
        #region "7.1 データの型に合わせたハッシュコードの作成"
        public static void CreateHashCodeDataType()
        {
            SimpleClass simpleClass = new SimpleClass("foo");

            Hashtable hashTable = new Hashtable();
            hashTable.Add(simpleClass, 100);

            Dictionary<SimpleClass, int> dict = new Dictionary<SimpleClass, int>();
            dict.Add(simpleClass, 100);
        }

        public class SimpleClass
        {
            private string text = "";

            public SimpleClass(string inputText)
            {
                text = inputText;
            }

            public override int GetHashCode()
            {
                return ShiftAndAddHash(text);
            }

            public int ShiftAndAddHash(string strValue)
            {
                int hashCode = 0;

                foreach (char c in strValue)
                {
                    hashCode = (hashCode << 5) + (int)c + (hashCode >> 2);
                }

                return hashCode;
            }
        }

        public int SimpleHash(params int[] values)
        {
            int hashCode = 0;
            if (values != null)
            {
                foreach (int val in values)
                {
                    hashCode ^= val;
                }
            }

            return hashCode;
        }

        public int FoldingHash(params long[] values)
        {
            int hashCode = 0;
            if (values != null)
            {
                int tempLowerVal = 0;
                int tempUpperVal = 0;
                foreach (long val in values)
                {
                    tempLowerVal = (int)(val & 0x000000007FFFFFFF);
                    tempUpperVal = (int)((val >> 32) & 0xFFFFFFFF);
                    hashCode ^= tempLowerVal ^ tempUpperVal;
                }
            }

            return hashCode;
        }

        public int ContainedObjHash(params object[] values)
        {
            int hashCode = 0;
            if (values != null)
            {
                foreach (int val in values)
                {
                    hashCode ^= val.GetHashCode();
                }
            }

            return hashCode;
        }

        public int CryptoHash(string strValue)
        {
            int hashCode = 0;
            if (strValue != null)
            {
                byte[] encodedUnHashedString =
                    Encoding.Unicode.GetBytes(strValue);

                byte[] key = new byte[16];
                RandomNumberGenerator.Create().GetBytes(key);

                MACTripleDES hashingObj = new MACTripleDES(key);
                byte[] code =
                    hashingObj.ComputeHash(encodedUnHashedString);

                // BitConverterクラスを使って最初の4バイトを取得し、
                // それを最後の4バイトにたたみ込みます。
                // これらの値をハッシュコードのintとして利用します。
                int hashCodeStart = BitConverter.ToInt32(code, 0);
                int hashCodeEnd = BitConverter.ToInt32(code, 4);
                hashCode = hashCodeStart ^ hashCodeEnd;
            }

            return hashCode;
        }

        public int CryptoHash(long longValue)
        {
            int hashCode = 0;
            byte[] encodedUnHashedString =
                Encoding.Unicode.GetBytes(longValue.ToString());

            SHA256Managed hashingObj = new SHA256Managed();
            byte[] code = hashingObj.ComputeHash(encodedUnHashedString);

            // BitConverterクラスを使って最初の4バイトを取得し、
            // それを最後の4バイトにたたみ込みます。
            // これらの値をハッシュコードのintとして利用します。
            int hashCodeStart = BitConverter.ToInt32(code, 0);
            int hashCodeEnd = BitConverter.ToInt32(code, 4);
            hashCode = hashCodeStart ^ hashCodeEnd;

            return hashCode;
        }

        public int ShiftAndAddHash(string strValue)
        {
            int hashCode = 0;

            foreach (char c in strValue)
            {
                hashCode = (hashCode << 5) + (int)c + (hashCode >> 2);
            }

            return hashCode;
        }

        public int CalcHash(short someShort, int someInt, long someLong,
            float someFloat, object someObject)
        {
            int hashCode = 7;
            hashCode = hashCode * 31 + (int)someShort;
            hashCode = hashCode * 31 + someInt;
            hashCode = hashCode * 31 +
                (int)(someLong ^ (someLong >> 32));
            long someFloatToLong = (long)someFloat;
            hashCode = hashCode * 31 +
                (int)(someFloatToLong ^ (someFloatToLong >> 32));

            if (someObject != null)
            {
                hashCode = hashCode * 31 +
                    someObject.GetHashCode();
            }

            return hashCode;
        }

        public int ConcatStringGetHashCode(int[] someIntArray)
        {
            int hashCode = 0;
            StringBuilder hashString = new StringBuilder();

            if (someIntArray != null)
            {
                foreach (int i in someIntArray)
                {
                    hashString.Append(i.ToString() + "^");
                }
            }
            hashCode = hashString.GetHashCode();

            return hashCode;
        }



        #endregion

        #region "7.2 優先待ち行列の作成"

        public static void CreatePriorityQueue()
        {
            // メッセージを格納するListを作成します。
            List<string> msgs = new List<string>();
            msgs.Add("foo");
            msgs.Add("これは長めのメッセージです。");
            msgs.Add("bar");
            msgs.Add(@"特殊文字を使用するメッセージです。
                   !@#$%^&*( )_+=-0987654321~|}{[]\\;:?/>.<,");
            msgs.Add(@"<

                     >");
            msgs.Add("<text>一</text><text>二</text><text>三</text>" +
                     "<text>四</text>");
            msgs.Add("");
            msgs.Add("1234567890");

            // 適切な比較演算子を伴う優先待ち行列を作成します。
            // 比較演算子は、解説セクションで定義するCompareLen型から作成します。
            CompareLen<string> comparer = new CompareLen<string>();
            PriorityQueue<string> pqueue = new PriorityQueue<string>(comparer);

            // リストのすべてのメッセージを優先待ち行列に加えます。
            foreach (string msg in msgs)
            {
                pqueue.Enqueue(msg);
            }
            // 優先順位に従って待ち行列の中のメッセージを表示します。
            foreach (string msg in pqueue)
            {
                Console.WriteLine("メッセージ: " + msg);
            }
            Console.WriteLine("pqueue.IndexOf('bar') == " + pqueue.IndexOf("bar"));
            Console.WriteLine("pqueue.IndexOf('_bar_') == " + pqueue.IndexOf("_bar_"));

            Console.WriteLine("pqueue.Contains('bar') == " + pqueue.Contains("bar"));
            Console.WriteLine("pqueue.Contains('_bar_') == " +
                    pqueue.Contains("_bar_"));

            Console.WriteLine("pqueue.BinarySearch('bar') == " +
                    pqueue.BinarySearch("bar"));
            Console.WriteLine("pqueue.BinarySearch('_bar_') == " +
                    pqueue.BinarySearch("_bar_"));

            // もっとも小さい方からメッセージを削除します。
            int currCount = pqueue.Count;
            for (int index = 0; index < currCount; index++)
            {
                Console.WriteLine("pqueue.DequeueLargest( ): " +
                pqueue.DequeueLargest().ToString());
            }
        }

        public class PriorityQueue<T> : IEnumerable
        {
            public PriorityQueue() { }
            public PriorityQueue(IComparer<T> icomparer)
            {
                specialComparer = icomparer;
            }


            private List<T> internalQueue = new List<T>();
            protected IComparer<T> specialComparer = null;


            protected List<T> InternalQueue
            {
                get { return internalQueue; }
            }

            public int Count
            {
                get { return internalQueue.Count; }
            }

            public void Clear()
            {
                internalQueue.Clear();
            }

            public object Clone()
            {
                // 新しい優先待ち行列を作成し、同じ比較演算オブジェクトを与えます。
                PriorityQueue<T> newPQ = new PriorityQueue<T>(specialComparer);
                newPQ.CopyTo(internalQueue.ToArray(), 0);
                return newPQ;
            }

            public int IndexOf(T item)
            {
                return internalQueue.IndexOf(item);
            }

            public bool Contains(T item)
            {
                return internalQueue.Contains(item);
            }

            public int BinarySearch(T item)
            {
                return internalQueue.BinarySearch(item, specialComparer);
            }

            public bool Contains(T item, IComparer<T> comparer)
            {
                if (internalQueue.BinarySearch(item, comparer) >= 0)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }

            public void CopyTo(T[] array, int index)
            {
                internalQueue.CopyTo(array, index);
            }

            public T[] ToArray()
            {
                return internalQueue.ToArray();
            }

            public void TrimExcess()
            {
                internalQueue.TrimExcess();
            }

            public void Enqueue(T item)
            {
                internalQueue.Add(item);
                internalQueue.Sort(specialComparer);
            }

            public T DequeueLargest()
            {
                T item = internalQueue[internalQueue.Count - 1];
                internalQueue.RemoveAt(internalQueue.Count - 1);

                return item;
            }

            public T PeekLargest()
            {
                return internalQueue[internalQueue.Count - 1];
            }

            public IEnumerator GetEnumerator()
            {
                return internalQueue.GetEnumerator();
            }
        }

        public class CompareLen<T> : IComparer<T>
            where T : IComparable<T>
        {
            public int Compare(T obj1, T obj2)
            {
                int result = 0;
                if ((obj1 is string) && (obj2 is string))
                {
                    result = CompareStrings(obj1 as string, obj2 as string);
                }
                else
                {
                    // オブジェクト既定の動作による比較を行います
                    result = Compare(obj1, obj2);
                }
                return result;
            }

            private int CompareStrings(string str1, string str2)
            {
                if (str1.Length == str2.Length)
                {
                    return 0;
                }
                else if (str1.Length > str2.Length)
                {
                    return 1;
                }
                else
                {
                    return -1;
                }
            }

            public bool Equals(T item1, T item2)
            {
                return item1.Equals(item2);
            }

            public int GetHashCode(T obj)
            {
                return obj.GetHashCode();
            }
        }

        public class CompareObjs<T> : IComparer<T>
            where T : IComparable<T>
        {
            public int Compare(T obj1, T obj2)
            {
                int result = 0;

                if ((!obj1.Equals(null)) && (!obj2.Equals(null)))
                {
                    result = obj1.CompareTo(obj2);
                }

                return result;
            }

            public bool Equals(T item1, T item2)
            {
                return item1.Equals(item2);
            }

            public int GetHashCode(T item)
            {
                return item.GetHashCode();
            }
        }
        #endregion

        #region	"7.3 一対多マップ(マルチマップ)の作成"
        public static void TestMultiMap()
        {
            string s = "foo";

            // MultiMapオブジェクトを生成し、値を投入します。
            MultiMap<int, string> myMap = new MultiMap<int, string>();
            myMap.Add(0, "零");
            myMap.Add(1, "一");
            myMap.Add(2, "二");
            myMap.Add(3, "三");
            myMap.Add(3, "重複した三");
            myMap.Add(3, "重複した三");
            myMap.Add(4, "null");
            myMap.Add(5, s);
            myMap.Add(6, s);

            // 中身を表示します。
            foreach (KeyValuePair<int, List<string>> entry in myMap)
            {
                Console.Write("キー: " + entry.Key.ToString() + "\t値: ");
                foreach (string str in myMap[entry.Key])
                {
                    Console.Write(str + " : ");
                }
                Console.WriteLine();
            }

            // インデクサを通して値を取得します。
            Console.WriteLine();
            Console.WriteLine("((ArrayList) myMap[3])[0]: " + myMap[3][0]);
            Console.WriteLine("((ArrayList) myMap[3])[1]: " + myMap[3][1]);

            // Listを使用してMultiMapに要素を追加します。
            List<string> testArray = new List<string>();
            testArray.Add("BAR");
            testArray.Add("BAZ");
            myMap[10] = testArray;
            myMap[10] = testArray;

            // MultiMapから要素を削除します。
            myMap.Remove(0);
            myMap.Remove(1);

            // MultiMapを表示します。
            Console.WriteLine();
            Console.WriteLine("myMap.Count: " + myMap.Count);
            foreach (KeyValuePair<int, List<string>> entry in myMap)
            {
                Console.Write("entry.Key: " + entry.Key.ToString() +
                    "\tentry.Value(s): ");
                foreach (string str in myMap[entry.Key])
                {
                    if (str == null)
                    {
                        Console.Write("null : ");
                    }
                    else
                    {
                        Console.Write(str + " : ");
                    }
                }
                Console.WriteLine();
            }

            // マップがキーや値を持っているかどうかを確認します。
            Console.WriteLine();
            Console.WriteLine("myMap.ContainsKey(2): " + myMap.ContainsKey(2));
            Console.WriteLine("myMap.ContainsValue(二): " +
            myMap.ContainsValue("二"));
            Console.WriteLine("Contains Key 2: " + myMap.ContainsKey(2));
            Console.WriteLine("Contains Key 12: " + myMap.ContainsKey(12));
            Console.WriteLine("Contains Value 二: " + myMap.ContainsValue("二"));
            Console.WriteLine("Contains Value BAR: " + myMap.ContainsValue("BAR"));

            // MultiMapのすべての要素をクリアします。
            myMap.Clear();

        }
        /* 出力結果
        
        キー: 0 値: 零 :
        キー: 1 値: 一 :
        キー: 2 値: 二 :
        キー: 3 値: 三 : 重複した三 : 重複した三 :
        キー: 4 値: null :
        キー: 5 値: foo :
        キー: 6 値: foo :

        ((ArrayList) myMap[3])[0]: 三
        ((ArrayList) myMap[3])[1]: 重複した三

        myMap.Count: 6
        entry.Key: 2    entry.Value(s): 二 :
        entry.Key: 3    entry.Value(s): 三 : 重複した三 : 重複した三 :
        entry.Key: 4    entry.Value(s): null :
        entry.Key: 5    entry.Value(s): foo :
        entry.Key: 6    entry.Value(s): foo :
        entry.Key: 10   entry.Value(s): BAR : BAZ :

        myMap.ContainsKey(2): True
        myMap.ContainsValue(二): True
        Contains Key 2: True
        Contains Key 12: False
        Contains Value 二: True
        Contains Value BAR: True
        */

        public class MultiMap<TKey, UValue> : IDictionary<TKey, List<UValue>>
        {
            private Dictionary<TKey, List<UValue>> map =
                new Dictionary<TKey, List<UValue>>();

            public List<UValue> this[TKey key]
            {
                get { return map[key]; }
                set { map[key] = value; }
            }

            public void Add(TKey key, UValue item)
            {
                AddSingleMap(key, item);
            }

            public void Add(TKey key, List<UValue> items)
            {
                foreach (UValue val in items)
                    AddSingleMap(key, val);
            }

            public void Add(KeyValuePair<TKey, List<UValue>> keyValuePair)
            {
                foreach (UValue val in keyValuePair.Value)
                    AddSingleMap(keyValuePair.Key, val);
            }

            public void Clear()
            {
                map.Clear();
            }

            public int Count
            {
                get { return map.Count; }
            }

            public bool ContainsKey(TKey key)
            {
                return map.ContainsKey(key);
            }

            public bool ContainsValue(UValue item)
            {
                if (item == null)
                {
                    foreach (KeyValuePair<TKey, List<UValue>> kvp in map)
                    {
                        if (((List<UValue>)kvp.Value).Count == 0)
                        {
                            return (true);
                        }
                    }

                    return (false);
                }
                else
                {
                    foreach (KeyValuePair<TKey, List<UValue>> kvp in map)
                    {
                        if (((List<UValue>)kvp.Value).Contains(item))
                        {
                            return (true);
                        }
                    }

                    return (false);
                }
            }

            public IEnumerator<KeyValuePair<TKey, List<UValue>>> GetEnumerator()
            {
                return map.GetEnumerator();
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return map.GetEnumerator();
            }

            public bool Remove(TKey key)
            {
                return RemoveSingleMap(key);
            }

            public bool Remove(KeyValuePair<TKey, List<UValue>> keyValuePair)
            {
                return Remove(keyValuePair.Key);
            }

            public ICollection<TKey> Keys
            {
                get { return map.Keys; }
            }

            public bool TryGetValue(TKey key, out List<UValue> value)
            {
                throw new NotImplementedException();
            }

            public ICollection<List<UValue>> Values
            {
                get { return map.Values; }
            }

            public bool Contains(KeyValuePair<TKey, List<UValue>> item)
            {
                return map.ContainsKey(item.Key) && map.ContainsValue(item.Value);
            }

            public void CopyTo(KeyValuePair<TKey, List<UValue>>[] array, int arrayIndex)
            {
                int index = 0;
                foreach (KeyValuePair<TKey, List<UValue>> item in map)
                {
                    if (index < arrayIndex)
                        continue;

                    array[index++] = item;
                }
            }

            public bool IsReadOnly
            {
                get { return false; }
            }

            protected void AddSingleMap(TKey key, UValue item)
            {
                // Dictionary型のmapの中でkeyを検索します。
                if (map.ContainsKey(key))
                {
                    // mapのリストに値を追加します。
                    List<UValue> values = (List<UValue>)map[key];

                    // すでにキーが存在していれば、そのキーに対応づけて値を追加します。
                    values.Add(item);
                }
                else
                {
                    if (item == null)
                    {
                        // 新しいキーを作成し、空のリストに対応付けます。
                        map.Add(key, new List<UValue>());
                    }
                    else
                    {
                        List<UValue> values = new List<UValue>();
                        values.Add(item);
                        // 新しいキーを作成し、値に対応付けます。
                        map.Add(key, values);
                    }
                }
            }

            protected bool RemoveSingleMap(TKey key)
            {
                if (this.ContainsKey(key))
                {
                    // キーを削除します。
                    return (map.Remove(key));
                }
                else
                {
                    throw (new ArgumentOutOfRangeException("key", key,
                    "マップにこのキーは存在しません。"));
                }
            }
        }

        #endregion

        #region "7.4 二分木の作成"
        public static void TestBinaryTree()
        {
            BinaryTree<string> tree = new BinaryTree<string>("d");
            tree.AddNode("a");
            tree.AddNode("b");
            tree.AddNode("f");
            tree.AddNode("e");
            tree.AddNode("c");
            tree.AddNode("g");

            tree.Print();
            tree.Print();

            Console.WriteLine("tree.TreeSize: " + tree.TreeSize);
            Console.WriteLine("tree.Root.DepthFirstSearch(a).NumOfChildren: " +
                tree.Root.DepthFirstSearch("b").Children);
            Console.WriteLine("tree.Root.DepthFirstSearch(a).NumOfChildren: " +
                tree.Root.DepthFirstSearch("a").Children);
            Console.WriteLine("tree.Root.DepthFirstSearch(g).NumOfChildren: " +
                tree.Root.DepthFirstSearch("g").Children);

            Console.WriteLine("tree.SearchDepthFirst(a): " +
                tree.SearchDepthFirst("a").Value.ToString());
            Console.WriteLine("tree.SearchDepthFirst(b): " +
                tree.SearchDepthFirst("b").Value.ToString());
            Console.WriteLine("tree.SearchDepthFirst(c): " +
                tree.SearchDepthFirst("c").Value.ToString());
            Console.WriteLine("tree.SearchDepthFirst(d): " +
                tree.SearchDepthFirst("d").Value.ToString());
            Console.WriteLine("tree.SearchDepthFirst(e): " +
                tree.SearchDepthFirst("e").Value.ToString());
            Console.WriteLine("tree.SearchDepthFirst(f): " +
                tree.SearchDepthFirst("f").Value.ToString());

            tree.Root.RemoveLeftNode();
            tree.Print();

            tree.Root.RemoveRightNode();
            tree.Print();
        }
        /*  ORIGINAL DATA
        a
                Contains Left:  NULL
                Contains Right: b
        b
                Contains Left:  NULL
                Contains Right: c
        c
                Contains Left:  NULL
                Contains Right: NULL
        d
                Contains Left: a
                Contains Right: f
        e
                Contains Left:  NULL
                Contains Right: NULL
        f
                Contains Left: e
                Contains Right: g
        g
                Contains Left:  NULL
                Contains Right: NULL
        a
                Contains Left:  NULL
                Contains Right: b
        b
                Contains Left:  NULL
                Contains Right: c
        c
                Contains Left:  NULL
                Contains Right: NULL
        d
                Contains Left: a
                Contains Right: f
        e
                Contains Left:  NULL
                Contains Right: NULL
        f
                Contains Left: e
                Contains Right: g
        g
                Contains Left:  NULL
                Contains Right: NULL
        tree.TreeSize: 7
        tree.Root.DepthFirstSearch(a).NumOfChildren: 1
        tree.Root.DepthFirstSearch(a).NumOfChildren: 2
        tree.Root.DepthFirstSearch(g).NumOfChildren: 0
        tree.SearchDepthFirst(a): a
        tree.SearchDepthFirst(b): b
        tree.SearchDepthFirst(c): c
        tree.SearchDepthFirst(d): d
        tree.SearchDepthFirst(e): e
        tree.SearchDepthFirst(f): f
        d
                Contains Left:  NULL
                Contains Right: f
        e
                Contains Left:  NULL
                Contains Right: NULL
        f
                Contains Left: e
                Contains Right: g
        g
                Contains Left:  NULL
                Contains Right: NULL
        d
                Contains Left:  NULL
                Contains Right: NULL
        */

        public static void TestManagedTreeWithNoBinaryTreeClass()
        {
            // ルートノードを作成します。
            BinaryTreeNode<string> topLevel = new BinaryTreeNode<string>("d");

            // ツリーに追加されるすべてのノードを作成します。
            BinaryTreeNode<string> one = new BinaryTreeNode<string>("b");
            BinaryTreeNode<string> two = new BinaryTreeNode<string>("c");
            BinaryTreeNode<string> three = new BinaryTreeNode<string>("a");
            BinaryTreeNode<string> four = new BinaryTreeNode<string>("e");
            BinaryTreeNode<string> five = new BinaryTreeNode<string>("f");
            BinaryTreeNode<string> six = new BinaryTreeNode<string>("g");

            // ルートからツリーを追加します。
            topLevel.AddNode(three);
            topLevel.AddNode(one);
            topLevel.AddNode(five);
            topLevel.AddNode(four);
            topLevel.AddNode(two);
            topLevel.AddNode(six);

            // ルートノードを開始点とするツリーを表示します。
            topLevel.PrintDepthFirst();

            // ノード「three」を開始点とするツリーを表示します。
            three.PrintDepthFirst();

            // 様々なノードの子ノード数を表示します。
            Console.WriteLine("topLevel.Children: " + topLevel.Children);
            Console.WriteLine("one.Children: " + one.Children);
            Console.WriteLine("three.Children: " + three.Children);
            Console.WriteLine("six.Children: " + six.Children);

            // 深さ優先探索のメソッドをしようして検索します。
            Console.WriteLine("topLevel.DepthFirstSearch(a): " +
                                topLevel.DepthFirstSearch("a").Value.ToString());
            Console.WriteLine("topLevel.DepthFirstSearch(b): " +
                                topLevel.DepthFirstSearch("b").Value.ToString());
            Console.WriteLine("topLevel.DepthFirstSearch(c): " +
                                topLevel.DepthFirstSearch("c").Value.ToString());
            Console.WriteLine("topLevel.DepthFirstSearch(d): " +
                                topLevel.DepthFirstSearch("d").Value.ToString());
            Console.WriteLine("topLevel.DepthFirstSearch(e): " +
                                topLevel.DepthFirstSearch("e").Value.ToString());
            Console.WriteLine("topLevel.DepthFirstSearch(f): " +
                                topLevel.DepthFirstSearch("f").Value.ToString());

            // ルートノードの左の子ノードを取り除き、ツリー全体を表示します。
            topLevel.RemoveLeftNode();
            topLevel.PrintDepthFirst();

            // ルートノード以外のすべてのノードをツリーから取り除き、ツリーを表示します。
            topLevel.RemoveRightNode();
            topLevel.PrintDepthFirst();
        }
        /*  ORIGINAL DATA
        a
                Contains Left:  NULL
                Contains Right: b
        b
                Contains Left:  NULL
                Contains Right: c
        c
                Contains Left:  NULL
                Contains Right: NULL
        d
                Contains Left: a
                Contains Right: f
        e
                Contains Left:  NULL
                Contains Right: NULL
        f
                Contains Left: e
                Contains Right: g
        g
                Contains Left:  NULL
                Contains Right: NULL
        a
                Contains Left:  NULL
                Contains Right: b
        b
                Contains Left:  NULL
                Contains Right: c
        c
                Contains Left:  NULL
                Contains Right: NULL
        topLevel.Children: 6
        one.Children: 1
        three.Children: 2
        six.Children: 0
        topLevel.DepthFirstSearch(a): a
        topLevel.DepthFirstSearch(b): b
        topLevel.DepthFirstSearch(c): c
        topLevel.DepthFirstSearch(d): d
        topLevel.DepthFirstSearch(e): e
        topLevel.DepthFirstSearch(f): f
        d
                Contains Left:  NULL
                Contains Right: f
        e
                Contains Left:  NULL
                Contains Right: NULL
        f
                Contains Left: e
                Contains Right: g
        g
                Contains Left:  NULL
                Contains Right: NULL
        d
                Contains Left:  NULL
                Contains Right: NULL
        */


        public class BinaryTree<T>
            where T : IComparable<T>
        {
            public BinaryTree() { }

            public BinaryTree(T value)
            {
                BinaryTreeNode<T> node = new BinaryTreeNode<T>(value);
                root = node;
                counter = 1;
            }

            private int counter = 0;               // ツリーのノード数
            private BinaryTreeNode<T> root = null; // ツリーのルートへのポインタ

            public void AddNode(T value)
            {
                BinaryTreeNode<T> node = new BinaryTreeNode<T>(value);
                ++counter;

                if (root == null)
                {
                    root = node;
                }
                else
                {
                    root.AddNode(node);
                }
            }

            public void AddNode(T value, int index)
            {
                BinaryTreeNode<T> node = new BinaryTreeNode<T>(value, index);
                ++counter;

                if (root == null)
                {
                    root = node;
                }
                else
                {
                    root.AddNode(node);
                }
            }

            public BinaryTreeNode<T> SearchDepthFirst(T value)
            {
                return (root.DepthFirstSearch(value));
            }

            public void Print()
            {
                root.PrintDepthFirst();
            }

            public BinaryTreeNode<T> Root
            {
                get { return (root); }
            }

            public int TreeSize
            {
                get { return (counter); }
            }
        }

        public class BinaryTreeNode<T>
            where T : IComparable<T>
        {
            public BinaryTreeNode() { }

            public BinaryTreeNode(T value)
            {
                nodeValue = value;
            }

            public BinaryTreeNode(int index)
            {
                nodeIndex = index;
            }

            public BinaryTreeNode(T value, int index)
            {
                nodeValue = value;
                nodeIndex = index;
            }

            private int nodeIndex = 0;
            private T nodeValue = default(T);
            private BinaryTreeNode<T> leftNode = null; // leftNode.nodeValue < Value
            private BinaryTreeNode<T> rightNode = null; // rightNode.nodeValue >= Value

            public int Children
            {
                get
                {
                    int currCount = 0;
                    if (leftNode != null)
                    {
                        ++currCount;
                        currCount += leftNode.Children;
                    }

                    if (rightNode != null)
                    {
                        ++currCount;
                        currCount += rightNode.Children;
                    }

                    return (currCount);
                }
            }

            public BinaryTreeNode<T> Left
            {
                get { return (leftNode); }
            }

            public BinaryTreeNode<T> Right
            {
                get { return (rightNode); }
            }

            public T Value
            {
                get { return (nodeValue); }
            }

            public void AddNode(BinaryTreeNode<T> node)
            {
                if (node.nodeValue.CompareTo(nodeValue) < 0)
                {
                    if (leftNode == null)
                    {
                        leftNode = node;
                    }
                    else
                    {
                        leftNode.AddNode(node);
                    }
                }
                else if (node.nodeValue.CompareTo(nodeValue) >= 0)
                {
                    if (rightNode == null)
                    {
                        rightNode = node;
                    }
                    else
                    {
                        rightNode.AddNode(node);
                    }
                }
            }

            public bool AddUniqueNode(BinaryTreeNode<T> node)
            {
                bool isUnique = true;
                if (node.nodeValue.CompareTo(nodeValue) < 0)
                {
                    if (leftNode == null)
                    {
                        leftNode = node;
                    }
                    else
                    {
                        leftNode.AddNode(node);
                    }
                }
                else if (node.nodeValue.CompareTo(nodeValue) > 0)
                {
                    if (rightNode == null)
                    {
                        rightNode = node;
                    }
                    else
                    {
                        rightNode.AddNode(node);
                    }
                }
                else //node.nodeValue.CompareTo(nodeValue) = 0
                {
                    isUnique = false;
                    // ここで例外を投げることもできます。
                }
                return (isUnique);
            }

            public BinaryTreeNode<T> DepthFirstSearch(T targetObj)
            {
                // 注意: foo.CompareTo(bar) == -1 --> (foo < bar)
                BinaryTreeNode<T> retObj = null;
                int comparisonResult = targetObj.CompareTo(nodeValue);
                if (comparisonResult == 0)
                {
                    retObj = this;
                }
                else if (comparisonResult > 0)
                {
                    if (rightNode != null)
                    {
                        retObj = rightNode.DepthFirstSearch(targetObj);
                    }
                }
                else if (comparisonResult < 0)
                {
                    if (leftNode != null)
                    {
                        retObj = leftNode.DepthFirstSearch(targetObj);
                    }
                }
                return (retObj);
            }

            public void PrintDepthFirst()
            {
                if (leftNode != null)
                {
                    leftNode.PrintDepthFirst();
                }

                Console.WriteLine(this.nodeValue.ToString());

                if (leftNode != null)
                {
                    Console.WriteLine("\t左の子ノード: " +
                        leftNode.nodeValue.ToString());
                }
                else
                {
                    Console.WriteLine("\t左の子ノード: NULL");
                }
                if (rightNode != null)
                {
                    Console.WriteLine("\t右の子ノード: " +
                        rightNode.nodeValue.ToString());
                }
                else
                {
                    Console.WriteLine("\t右の子ノード: NULL");
                }

                if (rightNode != null)
                {
                    rightNode.PrintDepthFirst();
                }
            }

            public List<T> CopyToList()
            {
                List<T> tempList = new List<T>();
                if (leftNode != null)
                {
                    tempList.AddRange(leftNode.CopyToList());
                    tempList.Add(leftNode.nodeValue);
                }
                if (rightNode != null)
                {
                    tempList.Add(rightNode.nodeValue);
                    tempList.AddRange(rightNode.CopyToList());
                }

                return (tempList);
            }

            public void RemoveLeftNode()
            {
                leftNode = null;
            }

            public void RemoveRightNode()
            {
                rightNode = null;
            }

            public int Index
            {
                get { return nodeIndex; }
            }
        }
        #endregion

        #region "7.5 n分木の作成"
        public static void TestNTree()
        {
            NTree<string> topLevel = new NTree<string>(3);
            NTreeNodeFactory<string> nodeFactory = new NTreeNodeFactory<string>(topLevel);

            NTreeNodeFactory<string>.NTreeNode<string> one = nodeFactory.CreateNode("One");
            NTreeNodeFactory<string>.NTreeNode<string> two = nodeFactory.CreateNode("Two");
            NTreeNodeFactory<string>.NTreeNode<string> three = nodeFactory.CreateNode("Three");
            NTreeNodeFactory<string>.NTreeNode<string> four = nodeFactory.CreateNode("Four");
            NTreeNodeFactory<string>.NTreeNode<string> five = nodeFactory.CreateNode("Five");
            NTreeNodeFactory<string>.NTreeNode<string> six = nodeFactory.CreateNode("Six");
            NTreeNodeFactory<string>.NTreeNode<string> seven = nodeFactory.CreateNode("Seven");
            NTreeNodeFactory<string>.NTreeNode<string> eight = nodeFactory.CreateNode("Eight");
            NTreeNodeFactory<string>.NTreeNode<string> nine = nodeFactory.CreateNode("Nine");

            topLevel.AddRoot(one);
            Console.WriteLine("topLevel.GetRoot().CountChildren: " +
                topLevel.GetRoot().CountChildren());

            topLevel.GetRoot().AddNode(two);
            topLevel.GetRoot().AddNode(three);
            topLevel.GetRoot().AddNode(four);

            topLevel.GetRoot().Children[0].AddNode(five);
            topLevel.GetRoot().Children[0].AddNode(eight);
            topLevel.GetRoot().Children[0].AddNode(nine);
            topLevel.GetRoot().Children[1].AddNode(six);
            topLevel.GetRoot().Children[1].Children[0].AddNode(seven);

            Console.WriteLine("ツリー全体の表示:");
            topLevel.GetRoot().PrintDepthFirst();

            Console.WriteLine("ノード[two]から下のツリー表示:");
            topLevel.GetRoot().Children[0].PrintDepthFirst();

            Console.WriteLine("深さ優先探索:");
            Console.WriteLine("topLevel.DepthFirstSearch(One): " +
                topLevel.GetRoot().DepthFirstSearch("One").Value().ToString());
            Console.WriteLine("topLevel.DepthFirstSearch(Two): " +
                topLevel.GetRoot().DepthFirstSearch("Two").Value().ToString());
            Console.WriteLine("topLevel.DepthFirstSearch(Three): " +
                topLevel.GetRoot().DepthFirstSearch("Three").Value().ToString());
            Console.WriteLine("topLevel.DepthFirstSearch(Four): " +
                topLevel.GetRoot().DepthFirstSearch("Four").Value().ToString());
            Console.WriteLine("topLevel.DepthFirstSearch(Five): " +
                topLevel.GetRoot().DepthFirstSearch("Five").Value().ToString());

            Console.WriteLine("\r\n\r\n幅優先探索:");
            Console.WriteLine("topLevel.BreadthFirstSearch(One): " +
                topLevel.GetRoot().BreadthFirstSearch("One").Value().ToString());
            Console.WriteLine("topLevel.BreadthFirstSearch(Two): " +
                topLevel.GetRoot().BreadthFirstSearch("Two").Value().ToString());
            Console.WriteLine("topLevel.BreadthFirstSearch(Three): " +
                topLevel.GetRoot().BreadthFirstSearch("Three").Value().ToString());
            Console.WriteLine("topLevel.BreadthFirstSearch(Four): " +
                topLevel.GetRoot().BreadthFirstSearch("Four").Value().ToString());
        }
        /*
        topLevel.GetRoot().CountChildren: 0
        ツリー全体の表示:
        this: One
                childNodes[0]:  Two
                childNodes[1]:  Three
                childNodes[2]:  Four
        this: Two
                childNodes[0]:  Five
                childNodes[1]:  Eight
                childNodes[2]:  Nine
        this: Five
                childNodes[0]:  NULL
                childNodes[1]:  NULL
                childNodes[2]:  NULL
        this: Eight
                childNodes[0]:  NULL
                childNodes[1]:  NULL
                childNodes[2]:  NULL
        this: Nine
                childNodes[0]:  NULL
                childNodes[1]:  NULL
                childNodes[2]:  NULL
        this: Three
                childNodes[0]:  Six
                childNodes[1]:  NULL
                childNodes[2]:  NULL
        this: Six
                childNodes[0]:  Seven
                childNodes[1]:  NULL
                childNodes[2]:  NULL
        this: Seven
                childNodes[0]:  NULL
                childNodes[1]:  NULL
                childNodes[2]:  NULL
        this: Four
                childNodes[0]:  NULL
                childNodes[1]:  NULL
                childNodes[2]:  NULL
        ノード[two]から下のツリー表示:
        this: Two
                childNodes[0]:  Five
                childNodes[1]:  Eight
                childNodes[2]:  Nine
        this: Five
                childNodes[0]:  NULL
                childNodes[1]:  NULL
                childNodes[2]:  NULL
        this: Eight
                childNodes[0]:  NULL
                childNodes[1]:  NULL
                childNodes[2]:  NULL
        this: Nine
                childNodes[0]:  NULL
                childNodes[1]:  NULL
                childNodes[2]:  NULL
        深さ優先探索:
        topLevel.DepthFirstSearch(One): One
        topLevel.DepthFirstSearch(Two): Two
        topLevel.DepthFirstSearch(Three): Three
        topLevel.DepthFirstSearch(Four): Four
        topLevel.DepthFirstSearch(Five): Five


        幅優先探索:
        topLevel.BreadthFirstSearch(One): One
        topLevel.BreadthFirstSearch(Two): Two
        topLevel.BreadthFirstSearch(Three): Three
        topLevel.BreadthFirstSearch(Four): Four
        */

        public class NTree<T>
            where T : IComparable<T>
        {
            public NTree()
            {
                maxChildren = int.MaxValue;
            }

            public NTree(int maxNumChildren)
            {
                maxChildren = maxNumChildren;
            }

            // ツリーのルートノード
            private NTreeNodeFactory<T>.NTreeNode<T> root = null;
            // 親ノードが保持できる最大子ノード数
            private int maxChildren = 0;


            public void AddRoot(NTreeNodeFactory<T>.NTreeNode<T> node)
            {
                root = node;
            }

            public NTreeNodeFactory<T>.NTreeNode<T> GetRoot()
            {
                return root;
            }

            public int MaxChildren
            {
                get { return maxChildren; }
            }
        }

        public class NTreeNodeFactory<T>
            where T : IComparable<T>
        {
            public NTreeNodeFactory(NTree<T> root)
            {
                maxChildren = root.MaxChildren;
            }


            private int maxChildren = 0;


            public int MaxChildren
            {
                get { return maxChildren; }
            }

            public NTreeNode<T> CreateNode(T value)
            {
                return new NTreeNode<T>(value, maxChildren);
            }


            // ネストされたNodeクラス
            public class NTreeNode<U>
                where U : IComparable<U>
            {
                public NTreeNode(U value, int maxChildren)
                {
                    if (!value.Equals(null))
                    {
                        nodeValue = value;
                    }

                    childNodes = new NTreeNode<U>[maxChildren];
                }


                protected U nodeValue = default(U);
                protected NTreeNode<U>[] childNodes = null;


                public int NumOfChildren
                {
                    get { return CountChildren(); }
                }

                public int CountChildren()
                {
                    int currCount = 0;

                    for (int index = 0; index <= childNodes.GetUpperBound(0); index++)
                    {
                        if (childNodes[index] != null)
                        {
                            ++currCount;
                            currCount += childNodes[index].CountChildren();
                        }
                    }

                    return currCount;
                }

                public int CountImmediateChildren()
                {
                    int currCount = 0;

                    for (int index = 0; index <= childNodes.GetUpperBound(0); index++)
                    {
                        if (childNodes[index] != null)
                        {
                            ++currCount;
                        }
                    }

                    return currCount;
                }


                public NTreeNode<U>[] Children
                {
                    get { return childNodes; }
                }

                public NTreeNode<U> GetChild(int index)
                {
                    return childNodes[index];
                }

                public U Value()
                {
                    return nodeValue;
                }

                public void AddNode(NTreeNode<U> node)
                {
                    int numOfNonNullNodes = CountImmediateChildren();

                    if (numOfNonNullNodes < childNodes.Length)
                    {
                        childNodes[numOfNonNullNodes] = node;
                    }
                    else
                    {
                        throw (new Exception("現在のノードにはこれ以上子ノードを追加できません。"));
                    }
                }

                public NTreeNode<U> DepthFirstSearch(U targetObj)
                {
                    NTreeNode<U> retObj = default(NTreeNode<U>);

                    if (targetObj.CompareTo(nodeValue) == 0)
                    {
                        retObj = this;
                    }
                    else
                    {
                        for (int index = 0; index <= childNodes.GetUpperBound(0); index++)
                        {
                            if (childNodes[index] != null)
                            {
                                retObj = childNodes[index].DepthFirstSearch(targetObj);
                                if (retObj != null)
                                {
                                    break;
                                }
                            }
                        }
                    }

                    return retObj;
                }

                public NTreeNode<U> BreadthFirstSearch(U targetObj)
                {
                    Queue<NTreeNode<U>> row = new Queue<NTreeNode<U>>();
                    row.Enqueue(this);

                    while (row.Count > 0)
                    {
                        // 待ち行列のなかの次のノードを取得
                        NTreeNode<U> currentNode = row.Dequeue();

                        // このノードは探しているノードか？
                        if (targetObj.CompareTo(currentNode.nodeValue) == 0)
                        {
                            return currentNode;
                        }

                        for (int index = 0;
                            index < currentNode.CountImmediateChildren();
                            index++)
                        {
                            if (currentNode.Children[index] != null)
                            {
                                row.Enqueue(currentNode.Children[index]);
                            }
                        }
                    }

                    return null;
                }

                public void PrintDepthFirst()
                {
                    Console.WriteLine("this: " + nodeValue.ToString());

                    for (int index = 0; index < childNodes.Length; index++)
                    {
                        if (childNodes[index] != null)
                        {
                            Console.WriteLine("\tchildNodes[" + index + "]:  " +
                                childNodes[index].nodeValue.ToString());
                        }
                        else
                        {
                            Console.WriteLine("\tchildNodes[" + index + "]:  NULL");
                        }
                    }

                    for (int index = 0; index < childNodes.Length; index++)
                    {
                        if (childNodes[index] != null)
                        {
                            childNodes[index].PrintDepthFirst();
                        }
                    }
                }

                public void RemoveNode(int index)
                {
                    // 配列からノードを削除し、配列を圧縮します。
                    if (index < childNodes.GetLowerBound(0) ||
                        index > childNodes.GetUpperBound(0))
                    {
                        throw (new ArgumentOutOfRangeException("index", index,
                            "インデックスが配列の境界外です。"));
                    }
                    else if (index < childNodes.GetUpperBound(0))
                    {
                        Array.Copy(childNodes, index + 1, childNodes, index,
                            childNodes.Length - index - 1);
                    }

                    childNodes.SetValue(null, childNodes.GetUpperBound(0));
                }
            }
        }
        #endregion

        #region "7.6 HashSetオブジェクトの利用"
        public static void TestSet()
        {
            HashSet<int> set1 = new HashSet<int>();
            HashSet<int> set2 = new HashSet<int>();
            HashSet<int> set3 = new HashSet<int>();

            set1.Add(1);
            set1.Add(2);
            set1.Add(3);
            set1.Add(4);
            set1.Add(5);
            set1.Add(6);

            set2.Add(-10);
            set2.Add(2);
            set2.Add(40);

            set3.Add(3);
            set3.Add(6);

            foreach (int o in set2)
            {
                Console.WriteLine(o.ToString());
            }

            Console.WriteLine("set1.Contains(2): " + set1.Contains(2));
            Console.WriteLine("set1.Contains(0): " + set1.Contains(0));

            Console.WriteLine("\r\nset1.Count: " + set1.Count);
            Console.WriteLine();
            Console.WriteLine("set1.DisplaySet: " + DisplaySet(set1));
            Console.WriteLine("set2.DisplaySet: " + DisplaySet(set2));
            Console.WriteLine("set3.DisplaySet: " + DisplaySet(set3));
            Console.WriteLine();
            set1.UnionWith(set2);
            Console.WriteLine("set1.UnionWith(set2): " +
                DisplaySet(set1));
            set1.IntersectWith(set2);
            Console.WriteLine("set1.IntersectWith(set2): " +
                DisplaySet(set1));
            set1.SymmetricExceptWith(set2);
            Console.WriteLine("set1.SymmetricExceptWith(set2): " +
                DisplaySet(set1));
            Console.WriteLine("set1.Equals(set2): " + set1.Equals(set2));
            Console.WriteLine("set1 == set2: " + (set1 == set2));
            Console.WriteLine("set1 != set2: " + (set1 != set2));
            Console.WriteLine("set1.IsSubsetOf(set2): " + set1.IsSubsetOf(set2));
            Console.WriteLine("set1.IsSupersetOf(set2): " + set1.IsSupersetOf(set2));
            Console.WriteLine();
            set2.UnionWith(set1);
            Console.WriteLine("set2.UnionWith(set1): " +
                DisplaySet(set2));
            set2.IntersectWith(set1);
            Console.WriteLine("set2.IntersectWith(set1): " +
                DisplaySet(set2));
            set2.ExceptWith(set1);
            Console.WriteLine("set2.ExceptWith(set1): " +
                DisplaySet(set2));
            Console.WriteLine("set2.Equals(set1): " + set2.Equals(set1));
            Console.WriteLine("set2 == set1): " + (set2 == set1));
            Console.WriteLine("set2 != set1): " + (set2 != set1));
            Console.WriteLine("set2.IsSubsetOf(set1): " + set2.IsSubsetOf(set1));
            Console.WriteLine("set2.IsSupersetOf(set1): " + set2.IsSupersetOf(set1));
            Console.WriteLine();
            set3.UnionWith(set1);
            Console.WriteLine("set3.UnionWith(set1): " +
                DisplaySet(set3));
            set3.IntersectWith(set1);
            Console.WriteLine("set3.IntersectWith(set1): " +
                DisplaySet(set3));
            set3.ExceptWith(set1);
            Console.WriteLine("set3.ExceptWith(set1): " +
                DisplaySet(set3));
            Console.WriteLine("set3.Equals(set1): " + set3.Equals(set1));
            Console.WriteLine("set3 == set1: " + (set3 == set1));
            Console.WriteLine("set3 != set1: " + (set3 != set1));
            Console.WriteLine("set3.IsSubsetOf(set1): " + set3.IsSubsetOf(set1));
            Console.WriteLine("set3.IsSupersetOf(set1): " + set3.IsSupersetOf(set1));
            Console.WriteLine("set1.IsSubsetOf(set3): " + set1.IsSubsetOf(set3));
            Console.WriteLine("set1.IsSupersetOf(set3): " + set1.IsSupersetOf(set3));
            Console.WriteLine();
            set3.UnionWith(set2);
            Console.WriteLine("set3.UnionWith(set2): " +
                DisplaySet(set3));
            set3.IntersectWith(set2);
            Console.WriteLine("set3.IntersectWith(set2): " +
                DisplaySet(set3));
            set3.ExceptWith(set2);
            Console.WriteLine("set3.ExceptWith(set2): " +
                DisplaySet(set3));
            Console.WriteLine("set3.Equals(set2): " + set3.Equals(set2));
            Console.WriteLine("set3 == set2: " + (set3 == set2));
            Console.WriteLine("set3 != set2: " + (set3 != set2));
            Console.WriteLine("set3.IsSubsetOf(set2): " + set3.IsSubsetOf(set2));
            Console.WriteLine("set3.IsSupersetOf(set2): " + set3.IsSupersetOf(set2));
            Console.WriteLine();
            Console.WriteLine("set3.Equals(set3): " + set3.Equals(set3));
            Console.WriteLine("set3 == set3: " + (set3 == set3));
            Console.WriteLine("set3 != set3: " + (set3 != set3));
            Console.WriteLine("set3.IsSubsetOf(set3): " + set3.IsSubsetOf(set3));
            Console.WriteLine("set3.IsSupersetOf(set3): " + set3.IsSupersetOf(set3));
        }

        public static string DisplaySet(HashSet<int> set)
        {
            if (set.Count == 0)
            {
                return "{}";
            }
            else
            {
                StringBuilder displayStr = new StringBuilder("{ ");

                foreach (int i in set)
                {
                    displayStr.Append(i);
                    displayStr.Append(", ");
                }

                displayStr.Remove(displayStr.Length - 2, 2);
                displayStr.Append(" }");

                return displayStr.ToString();
            }
        }

        #endregion
    }
}
