﻿#define DOTRACE

using System;
using System.Configuration;
using System.IO;
using System.IO.IsolatedStorage;
using System.Net;
using System.Runtime.InteropServices;
using System.Runtime.Serialization;
using System.Security;
using System.Security.AccessControl;
using System.Security.Cryptography;
using System.Security.Permissions;
using System.Security.Principal;
using System.Text;
using System.Text.RegularExpressions;
using System.Web.Configuration;
using System.Xml;
using Microsoft.Win32;

namespace CSharpRecipes
{
    public class Security
    {
        #region "10.1 ローカルアセンブリの型に対するアクセスを制御する"
        public static void ControlAccess()
        {
            // セキュリティプロキシを作成します
            CompanyDataSecProxy companyDataSecProxy = new CompanyDataSecProxy();

            // データを読み取ります
            Console.WriteLine("CEOPhoneNumExt: " +
                companyDataSecProxy.CEOPhoneNumExt);

            // データを書き込みます
            companyDataSecProxy.AdminPwd = "asdf";
            companyDataSecProxy.AdminUserName = "asdf";

            // データを保存して更新します
            companyDataSecProxy.SaveNewData();
            companyDataSecProxy.RefreshData();


            // プロキシを経由せずにCompanyDataオブジェクトを作成します
            CompanyData companyData = new CompanyData();

            // データを読み取ります
            Console.WriteLine("CEOPhoneNumExt: " + companyData.CEOPhoneNumExt);

            // データを書き込みます
            companyData.AdminPwd = "asdf";
            companyData.AdminUserName = "asdf";

            // データを保存して更新します
            companyData.SaveNewData();
            companyData.RefreshData();
        }

        internal interface ICompanyData
        {
            string AdminUserName
            {
                get;
                set;
            }

            string AdminPwd
            {
                get;
                set;
            }

            string CEOPhoneNumExt
            {
                get;
                set;
            }

            void RefreshData();
            void SaveNewData();
        }

        internal class CompanyData : ICompanyData
        {
            public CompanyData()
            {
                Console.WriteLine("[CONCRETE] CompanyDataを作成しました");
                // この位置で機密性の高い初期化処理を行います
                this.AdminUserName = "admin";
                this.AdminPwd = "password";
                this.CEOPhoneNumExt = "0000";
            }

            public string AdminUserName
            {
                get;
                set;
            }

            public string AdminPwd
            {
                get;
                set;
            }

            public string CEOPhoneNumExt
            {
                get;
                set;
            }

            public void RefreshData()
            {
                Console.WriteLine("[CONCRETE] データが更新されました");
            }

            public void SaveNewData()
            {
                Console.WriteLine("[CONCRETE] データが保存されました");
            }
        }

        public class CompanyDataSecProxy : ICompanyData
        {
            public CompanyDataSecProxy()
            {
                Console.WriteLine("[SECPROXY] 作成されました");

                // 最初にプリンシパルポリシーを設定する必要があります
                AppDomain.CurrentDomain.SetPrincipalPolicy(PrincipalPolicy.WindowsPrincipal);
            }


            private ICompanyData coData = null;
            private PrincipalPermission admPerm =
                new PrincipalPermission(null, @"BUILTIN\Administrators", true);
            private PrincipalPermission guestPerm =
                new PrincipalPermission(null, @"BUILTIN\Guest", true);
            private PrincipalPermission powerPerm =
                new PrincipalPermission(null, @"BUILTIN\PowerUser", true);
            private PrincipalPermission userPerm =
                new PrincipalPermission(null, @"BUILTIN\User ", true);

            public string AdminUserName
            {
                get
                {
                    string userName = "";
                    try
                    {
                        admPerm.Demand();
                        Startup();
                        userName = coData.AdminUserName;
                    }
                    catch (SecurityException e)
                    {
                        Console.WriteLine("AdminUserName_get が失敗しました！{0}", e.ToString());
                    }
                    return userName;
                }
                set
                {
                    try
                    {
                        admPerm.Demand();
                        Startup();
                        coData.AdminUserName = value;
                    }
                    catch (SecurityException e)
                    {
                        Console.WriteLine("AdminUserName_set が失敗しました！{0}", e.ToString());
                    }
                }
            }

            public string AdminPwd
            {
                get
                {
                    string pwd = "";
                    try
                    {
                        admPerm.Demand();
                        Startup();
                        pwd = coData.AdminPwd;
                    }
                    catch (SecurityException e)
                    {
                        Console.WriteLine("AdminPwd_get が失敗しました！{0}", e.ToString());
                    }

                    return pwd;
                }
                set
                {
                    try
                    {
                        admPerm.Demand();
                        Startup();
                        coData.AdminPwd = value;
                    }
                    catch (SecurityException e)
                    {
                        Console.WriteLine("AdminPwd_set が失敗しました！{0}", e.ToString());
                    }
                }
            }



            public string CEOPhoneNumExt
            {
                get
                {
                    string ceoPhoneNum = "";
                    try
                    {
                        admPerm.Union(powerPerm).Demand();
                        Startup();
                        ceoPhoneNum = coData.CEOPhoneNumExt;
                    }
                    catch (SecurityException e)
                    {
                        Console.WriteLine("CEOPhoneNum_set が失敗しました！{0}", e.ToString());
                    }
                    return ceoPhoneNum;
                }
                set
                {
                    try
                    {
                        admPerm.Demand();
                        Startup();
                        coData.CEOPhoneNumExt = value;
                    }
                    catch (SecurityException e)
                    {
                        Console.WriteLine("CEOPhoneNum_set が失敗しました！{0}", e.ToString());
                    }
                }
            }

            public void RefreshData()
            {
                try
                {
                    admPerm.Union(powerPerm.Union(userPerm)).Demand();
                    Startup();
                    Console.WriteLine("[SECPROXY] データが更新されました");
                    coData.RefreshData();
                }
                catch (SecurityException e)
                {
                    Console.WriteLine("RefreshData が失敗しました！{0}", e.ToString());
                }
            }

            public void SaveNewData()
            {
                try
                {
                    admPerm.Union(powerPerm).Demand();
                    Startup();
                    Console.WriteLine("[SECPROXY] データが保存されました");
                    coData.SaveNewData();
                }
                catch (SecurityException e)
                {
                    Console.WriteLine("SaveNewData が失敗しました！{0}", e.ToString());
                }
            }

            // プロキシによる追跡を有効にする場合 [#define DOTRACE] を忘れないようにしてください。
            private void Startup()
            {
                if (coData == null)
                {
#if (DOTRACE)
                    coData = new CompanyDataTraceProxy();
#else
                    coData = new CompanyData();
#endif
                    Console.WriteLine("[SECPROXY] データが更新されました");
                    coData.RefreshData();
                }
            }

            public class CompanyDataTraceProxy : ICompanyData
            {
                public CompanyDataTraceProxy()
                {
                    Console.WriteLine("[TRACEPROXY] 作成されました");
                    string path = Path.GetTempPath() + @"\CompanyAccessTraceFile.txt";
                    fileStream = new FileStream(path, FileMode.Append,
                        FileAccess.Write, FileShare.None);
                    traceWriter = new StreamWriter(fileStream);
                    coData = new CompanyData();
                }

                private ICompanyData coData = null;
                private FileStream fileStream = null;
                private StreamWriter traceWriter = null;


                public string AdminPwd
                {
                    get
                    {
                        traceWriter.WriteLine("AdminPwd がユーザにより読み取られました。");
                        traceWriter.Flush();
                        return coData.AdminPwd;
                    }
                    set
                    {
                        traceWriter.WriteLine("AdminPwd がユーザにより書き込まれました。");
                        traceWriter.Flush();
                        coData.AdminPwd = value;
                    }
                }

                public string AdminUserName
                {
                    get
                    {
                        traceWriter.WriteLine("AdminUserName がユーザにより読み取られました。");
                        traceWriter.Flush();
                        return coData.AdminUserName;
                    }
                    set
                    {
                        traceWriter.WriteLine("AdminUserName がユーザにより書き込まれました。");
                        traceWriter.Flush();
                        coData.AdminUserName = value;
                    }
                }

                public string CEOPhoneNumExt
                {
                    get
                    {
                        traceWriter.WriteLine("CEOPhoneNumExt がユーザにより読み取られました。");
                        traceWriter.Flush();
                        return coData.CEOPhoneNumExt;
                    }
                    set
                    {
                        traceWriter.WriteLine("CEOPhoneNumExt がユーザにより書き込まれました。");
                        traceWriter.Flush();
                        coData.CEOPhoneNumExt = value;
                    }
                }

                public void RefreshData()
                {
                    Console.WriteLine("[TRACEPROXY] データが更新されました");
                    coData.RefreshData();
                }

                public void SaveNewData()
                {
                    Console.WriteLine("[TRACEPROXY] データが保存されました");
                    coData.SaveNewData();
                }
            }

        }

        #endregion

        #region "10.2 文字列の暗号化および複合化"
        public static void EncDecString()
        {
            string encryptedString = CryptoString.Encrypt("パスワード", Encoding.UTF8);
            Console.WriteLine("暗号化された文字列：" + encryptedString);
            // 秘密鍵およびお初期ベクタを保持しておき、復号化の際に使用できるようにします
            byte[] key = CryptoString.Key;
            byte[] IV = CryptoString.IV;

            CryptoString.Key = key;
            CryptoString.IV = IV;
            string decryptedString = CryptoString.Decrypt(encryptedString, Encoding.UTF8);
            Console.WriteLine("複合化された文字列：" + decryptedString);
        }

        public sealed class CryptoString
        {
            private CryptoString() { }

            private static byte[] savedKey = null;
            private static byte[] savedIV = null;

            public static byte[] Key
            {
                get { return savedKey; }
                set { savedKey = value; }
            }

            public static byte[] IV
            {
                get { return savedIV; }
                set { savedIV = value; }
            }

            private static void RdGenerateSecretKey(RijndaelManaged rdProvider)
            {
                if (savedKey == null)
                {
                    rdProvider.KeySize = 256;
                    rdProvider.GenerateKey();
                    savedKey = rdProvider.Key;
                }
            }

            private static void RdGenerateSecretInitVector(RijndaelManaged rdProvider)
            {
                if (savedIV == null)
                {
                    rdProvider.GenerateIV();
                    savedIV = rdProvider.IV;
                }
            }

            public static string Encrypt(string originalStr, Encoding stringEncoding)
            {
                // データ文字列をエンコードしてメモリ上に保存します
                byte[] originalStrAsBytes = stringEncoding.GetBytes(originalStr);
                byte[] originalBytes = { };

                // 出力を保持するためのMemoryStreamを作成します
                MemoryStream memStream = new MemoryStream(originalStrAsBytes.Length);

                RijndaelManaged rijndael = new RijndaelManaged();

                // 秘密鍵および初期ベクタを生成します
                RdGenerateSecretKey(rijndael);
                RdGenerateSecretInitVector(rijndael);

                if (savedKey == null || savedIV == null)
                {
                    throw new NullReferenceException(
                        "savedKey および savedIV はnullにできません。");
                }

                // 暗号化オブジェクトおよび暗号ストリームを作成します
                ICryptoTransform rdTransform =
                    rijndael.CreateEncryptor((byte[])savedKey.Clone(),
                                            (byte[])savedIV.Clone());
                CryptoStream cryptoStream = new CryptoStream(memStream, rdTransform,
                    CryptoStreamMode.Write);

                // MemoryStreamに暗号化されたデータを書き込みます
                cryptoStream.Write(originalStrAsBytes, 0, originalStrAsBytes.Length);
                cryptoStream.FlushFinalBlock();
                originalBytes = memStream.ToArray();

                // すべてのリソースを解放します
                memStream.Close();
                cryptoStream.Close();
                rdTransform.Dispose();
                rijndael.Clear();

                // 暗号化された文字列を変換します
                string encryptedStr = Convert.ToBase64String(originalBytes);
                return encryptedStr;
            }

            public static string Decrypt(string encryptedStr, Encoding stringEncoding)
            {
                // 暗号化された文字列をバイト列に逆変換します
                byte[] encryptedStrAsBytes = Convert.FromBase64String(encryptedStr);
                byte[] initialText = new Byte[encryptedStrAsBytes.Length];

                RijndaelManaged rijndael = new RijndaelManaged();
                MemoryStream memStream = new MemoryStream(encryptedStrAsBytes);

                if (savedKey == null || savedIV == null)
                {
                    throw new NullReferenceException(
                        "savedKey および savedIV はnullにできません。");
                }

                // 復号化オブジェクト、および復号ストリームを作成します
                ICryptoTransform rdTransform =
                    rijndael.CreateDecryptor((byte[])savedKey.Clone(),
                                            (byte[])savedIV.Clone());
                CryptoStream cryptoStream = new CryptoStream(memStream, rdTransform,
                    CryptoStreamMode.Read);

                // 復号化された文字列をbyte[]として読み取ります
                cryptoStream.Read(initialText, 0, initialText.Length);

                // すべてのリソースを解放します
                memStream.Close();
                cryptoStream.Close();
                rdTransform.Dispose();
                rijndael.Clear();

                // byte[]を文字列に変換します
                string decryptedStr = stringEncoding.GetString(initialText);
                return decryptedStr;
            }
        }

        #endregion

        #region "10.3 ファイルの暗号化および復号化"
        public static void EncDecFile()
        {
            // TripleDES暗号を使用します
            using (TripleDESCryptoServiceProvider tdes =
                new TripleDESCryptoServiceProvider())
            {
                SecretFile secretTDESFile = new SecretFile(tdes, "tdestext.secret");

                string encrypt = "テスト。\r\nTDES用の秘密データです！";

                Console.WriteLine("秘密データを書き込みます：{0}", encrypt);
                secretTDESFile.SaveSensitiveData(encrypt, Encoding.UTF8);
                // ファイルを読み取るために秘密鍵と初期ベクタを保存します
                byte[] key = secretTDESFile.Key;
                byte[] IV = secretTDESFile.IV;

                string decrypt = secretTDESFile.ReadSensitiveData(Encoding.UTF8);
                Console.WriteLine("読み取った秘密データ：{0}", decrypt);
            }

            // Rijndael暗号を使用します
            using (RijndaelManaged rdProvider = new RijndaelManaged())
            {
                SecretFile secretRDFile = new SecretFile(rdProvider, "rdtext.secret");

                string encrypt = "テスト。\r\nRijndael用の秘密データです！";

                Console.WriteLine("秘密データを書き込みます：{0}", encrypt);
                secretRDFile.SaveSensitiveData(encrypt, Encoding.UTF8);
                // ファイルを読み取るために秘密鍵と初期ベクタを保存します
                byte[] key = secretRDFile.Key;
                byte[] IV = secretRDFile.IV;

                string decrypt = secretRDFile.ReadSensitiveData(Encoding.UTF8);
                Console.WriteLine("Read secret data: {0}", decrypt);
            }
        }

        public class SecretFile
        {
            private byte[] savedKey = null;
            private byte[] savedIV = null;
            private SymmetricAlgorithm symmetricAlgorithm;
            string path;

            public byte[] Key
            {
                get { return savedKey; }
                set { savedKey = value; }
            }

            public byte[] IV
            {
                get { return savedIV; }
                set { savedIV = value; }
            }

            public SecretFile(SymmetricAlgorithm algorithm, string fileName)
            {
                symmetricAlgorithm = algorithm;
                path = fileName;
            }

            public void SaveSensitiveData(string sensitiveData, Encoding dataEncoding)
            {
                // 暗号化されたファイルに書き込む文字列をエンコードします
                byte[] encodedData = dataEncoding.GetBytes(sensitiveData);

                // 暗号化ファイル用のFileStreamを作成します。
                FileStream fileStream = new FileStream(path,
                    FileMode.Create,
                    FileAccess.Write);

                // 秘密鍵と初期ベクタを生成します
                GenerateSecretKey();
                GenerateSecretInitVector();

                // 暗号変換オブジェクトと暗号ストリームオブジェクトを作成します
                ICryptoTransform transform = symmetricAlgorithm.CreateEncryptor(savedKey,
                    savedIV);
                CryptoStream cryptoStream =
                    new CryptoStream(fileStream, transform, CryptoStreamMode.Write);

                // ファイルに暗号化データを書き込みます
                cryptoStream.Write(encodedData, 0, encodedData.Length);

                // すべてのリソースを解放します
                cryptoStream.Close();
                transform.Dispose();
                fileStream.Close();
            }

            public string ReadSensitiveData(Encoding dataEncoding)
            {
                // 暗号化されたファイルを表すFileStreamを作成します
                FileStream fileStream = new FileStream(path,
                    FileMode.Open,
                    FileAccess.Read);

                // 暗号化されたファイルの内容を表示します
                BinaryReader binReader = new BinaryReader(fileStream);
                Console.WriteLine("---------- 暗号化されたデータ ---------");
                int count = (Convert.ToInt32(binReader.BaseStream.Length));
                byte[] bytes = binReader.ReadBytes(count);
                char[] array = dataEncoding.GetChars(bytes);
                string encdata = new string(array);
                Console.WriteLine(encdata);
                Console.WriteLine("---------- 暗号化されたデータ ---------\r\n");

                // ストリームをリセットします
                fileStream.Seek(0, SeekOrigin.Begin);

                // 復号オブジェクトを作成します
                ICryptoTransform transform = symmetricAlgorithm.CreateDecryptor(savedKey,
                    savedIV);
                CryptoStream cryptoStream = new CryptoStream(fileStream,
                    transform,
                    CryptoStreamMode.Read);

                // 復号化されたファイルの内容を表示します
                StreamReader srDecrypted = new StreamReader(cryptoStream, dataEncoding);
                Console.WriteLine("---------- 復号化されたデータ ---------");
                string decrypted = srDecrypted.ReadToEnd();
                Console.WriteLine(decrypted);
                Console.WriteLine("---------- 復号化されたデータ ---------");

                // すべてのリソースを解放します
                binReader.Close();
                srDecrypted.Close();
                cryptoStream.Close();
                transform.Dispose();
                fileStream.Close();
                return decrypted;
            }

            private void GenerateSecretKey()
            {
                if (null != (symmetricAlgorithm as TripleDESCryptoServiceProvider))
                {
                    TripleDESCryptoServiceProvider tdes;
                    tdes = symmetricAlgorithm as TripleDESCryptoServiceProvider;
                    tdes.KeySize = 192; //  秘密鍵の最大サイズ
                    tdes.GenerateKey();
                    savedKey = tdes.Key;
                }
                else if (null != (symmetricAlgorithm as RijndaelManaged))
                {
                    RijndaelManaged rdProvider;
                    rdProvider = symmetricAlgorithm as RijndaelManaged;
                    rdProvider.KeySize = 256; // 秘密鍵の最大サイズ
                    rdProvider.GenerateKey();
                    savedKey = rdProvider.Key;
                }
            }

            private void GenerateSecretInitVector()
            {
                if (null != (symmetricAlgorithm as TripleDESCryptoServiceProvider))
                {
                    TripleDESCryptoServiceProvider tdes;
                    tdes = symmetricAlgorithm as TripleDESCryptoServiceProvider;
                    tdes.GenerateIV();
                    savedIV = tdes.IV;
                }
                else if (null != (symmetricAlgorithm as RijndaelManaged))
                {
                    RijndaelManaged rdProvider;
                    rdProvider = symmetricAlgorithm as RijndaelManaged;
                    rdProvider.GenerateIV();
                    savedIV = rdProvider.IV;
                }
            }
        }

        #endregion

        #region "10.4 暗号化に関する情報を消去する"
        public static void CleanUpCrypto()
        {
            string originalStr = "極秘情報";
            // データ文字列をエンコードしてメモリ上に保存します
            byte[] originalStrAsBytes = Encoding.UTF8.GetBytes(originalStr);
            byte[] originalBytes = { };

            // データを出力するためのMemoryStreamを作成します
            MemoryStream memStream = new MemoryStream(originalStrAsBytes.Length);

            RijndaelManaged rijndael = new RijndaelManaged();

            // 秘密鍵と初期ベクタを生成します
            rijndael.KeySize = 256;
            rijndael.GenerateKey();
            rijndael.GenerateIV();

            // 秘密鍵と初期ベクタを復号化用に保存します
            byte[] key = rijndael.Key;
            byte[] IV = rijndael.IV;

            // 暗号化オブジェクトおよび暗号ストリームを作成します
            ICryptoTransform transform = rijndael.CreateEncryptor(rijndael.Key,
                rijndael.IV);
            CryptoStream cryptoStream = new CryptoStream(memStream, transform,
                CryptoStreamMode.Write);

            // 暗号化データをMemoryStreamに書き込みます
            cryptoStream.Write(originalStrAsBytes, 0, originalStrAsBytes.Length);
            cryptoStream.FlushFinalBlock();

            // 暗号化処理の直後にすべてのリソースを解放して
            // メモリ上にデータが残されないようにします
            memStream.Close();
            memStream = null;
            cryptoStream.Close();
            cryptoStream = null;
            transform.Dispose();
            transform = null;
            // 以下の呼び出しにより秘密鍵および初期ベクタが消去されるため、
            // 暗号化時に使用したこれらのデータはメモリからも消去されます
            rijndael.Clear();
            // GCによって即座にオブジェクトが回収されるようにします
            rijndael = null;
        }

        public static void CleanUpCryptoWithUsing()
        {
            string originalStr = "極秘情報";
            // データ文字列をエンコードしてメモリ上に保存します
            byte[] originalStrAsBytes = Encoding.UTF8.GetBytes(originalStr);
            byte[] originalBytes = { };

            // データを出力するためのMemoryStreamを作成します
            using (MemoryStream memStream =
                new MemoryStream(originalStrAsBytes.Length))
            using (RijndaelManaged rijndael = new RijndaelManaged())
            {
                // 秘密鍵と初期ベクタを生成します
                rijndael.KeySize = 256;
                rijndael.GenerateKey();
                rijndael.GenerateIV();

                // 秘密鍵と初期ベクタを復号化用に保存します
                byte[] key = rijndael.Key;
                byte[] IV = rijndael.IV;

                // 暗号化オブジェクトおよび暗号ストリームを作成します
                using (ICryptoTransform transform =
                    rijndael.CreateEncryptor(rijndael.Key, rijndael.IV))
                using (CryptoStream cryptoStream =
                    new CryptoStream(memStream, transform, CryptoStreamMode.Write))
                {

                    // 暗号化データをMemoryStreamに書き込みます
                    cryptoStream.Write(originalStrAsBytes, 0,
                        originalStrAsBytes.Length);
                    cryptoStream.FlushFinalBlock();
                }
            }
        }

        #endregion

        #region "10.5 データ転送時に文字列が改ざんされていないかどうかを検証する"
        public static void VerifyNonStringCorruption()
        {
            string testString = "これはテスト用の文字列です。";
            string unhashedString;
            string hashedString = HashOps.CreateStringHash(testString, Encoding.UTF8);

            bool result = HashOps.IsStringCorrupted(hashedString, Encoding.UTF8,
                out unhashedString);
            Console.WriteLine(result);
            if (!result)
                Console.WriteLine("送信された文字列：" + unhashedString);
            else
                Console.WriteLine("文字列は改ざんされています。");
        }

        public class HashOps
        {
            // 44はハッシュ値をbase64エンコードした際の文字列長で、
            // この長さの分だけ文字列に追加されます
            private const int HASH_LENGTH = 44;

            public static string CreateStringHash(string unHashedString, Encoding stringEncoding)
            {
                byte[] encodedUnHashedString = stringEncoding.GetBytes(unHashedString);

                SHA256Managed hashingObj = new SHA256Managed();
                byte[] hashCode = hashingObj.ComputeHash(encodedUnHashedString);

                string hashBase64 = Convert.ToBase64String(hashCode);
                string stringWithHash = unHashedString + hashBase64;

                hashingObj.Clear();

                return stringWithHash;
            }

            public static bool IsStringCorrupted(string stringWithHash, Encoding stringEncoding,
                out string originalStr)
            {
                // 次のコードを有効にすることで文字列の改ざんをテストできます
                stringWithHash = stringWithHash.ToUpper();

                if (stringWithHash.Length <= HASH_LENGTH)
                {
                    originalStr = null;
                    return true;
                }

                string hashCodeString =
                    stringWithHash.Substring(stringWithHash.Length - HASH_LENGTH);
                string unHashedString =
                    stringWithHash.Substring(0, stringWithHash.Length - HASH_LENGTH);

                byte[] hashCode = Convert.FromBase64String(hashCodeString);

                byte[] encodedUnHashedString = stringEncoding.GetBytes(unHashedString);

                SHA256Managed hashingObj = new SHA256Managed();
                byte[] receivedHashCode = hashingObj.ComputeHash(encodedUnHashedString);

                bool hasBeenTamperedWith = false;
                for (int counter = 0; counter < receivedHashCode.Length; counter++)
                {
                    if (receivedHashCode[counter] != hashCode[counter])
                    {
                        hasBeenTamperedWith = true;
                        break;
                    }
                }

                if (!hasBeenTamperedWith)
                {
                    originalStr = unHashedString;
                }
                else
                {
                    originalStr = null;
                }

                hashingObj.Clear();

                return hasBeenTamperedWith;
            }
        }

        #endregion

        #region "10.6 データを安全に保持する"
        public static void SecurelyStoringData()
        {
            UserSettings settings = new UserSettings("89734kdkjs8734kjhd");
            if (settings.IsPasswordValid("89734kdkjs8734kjhd"))
            {
                Console.WriteLine("ようこそ");
            }
        }

        // ユーザ設定を保持するクラスです
        public class UserSettings
        {
            IsolatedStorageFile isoStorageFile = null;
            IsolatedStorageFileStream isoFileStream = null;
            XmlDocument settingsDoc = null;
            XmlTextWriter writer = null;
            const string storageName = "SettingsStorage.xml";

            // コンストラクタ
            public UserSettings(string password)
            {
                // 分離ストレージを取得します
                isoStorageFile = IsolatedStorageFile.GetUserStoreForDomain();
                // 設定用の内部DOMを作成します
                settingsDoc = new XmlDocument();
                // 設定が無い場合は既定値で設定を作成します
                if (isoStorageFile.GetFileNames(storageName).Length == 0)
                {
                    isoFileStream =
                        new IsolatedStorageFileStream(storageName,
                        FileMode.Create,
                        isoStorageFile);

                    writer = new XmlTextWriter(isoFileStream, Encoding.UTF8);
                    writer.WriteStartDocument();
                    writer.WriteStartElement("Settings");
                    writer.WriteStartElement("User");
                    // アプリケーションを操作している現在のユーザを取得します
                    WindowsIdentity user = WindowsIdentity.GetCurrent();
                    writer.WriteString(user.Name);
                    writer.WriteEndElement();
                    writer.WriteStartElement("Password");
                    // salt値としてnullを指定してパスワードを作成します
                    string hashedPassword = CreateHashedPassword(password, null);
                    writer.WriteString(hashedPassword);
                    writer.WriteEndElement();
                    writer.WriteEndElement();
                    writer.WriteEndDocument();
                    writer.Flush();
                    writer.Close();
                    Console.WriteLine("ユーザ {0} の設定を保存しました。", user.Name);
                }

                // 設定を保持する分離ストレージファイルのストリームを作成します
                isoFileStream =
                    new IsolatedStorageFileStream(storageName,
                    FileMode.Open,
                    isoStorageFile);

                // 分離ストレージファイルのストリームから設定を読み取ります
                settingsDoc.Load(isoFileStream);
                Console.WriteLine("ユーザ {0} の設定を読み込みました。", User);
            }

            // User プロパティ
            public string User
            {
                get
                {
                    XmlNode userNode = settingsDoc.SelectSingleNode("Settings/User");
                    if (userNode != null)
                    {
                        return userNode.InnerText;
                    }
                    return "";
                }
            }

            // Password プロパティ
            public string Password
            {
                get
                {
                    XmlNode pwdNode =
                        settingsDoc.SelectSingleNode("Settings/Password");
                    if (pwdNode != null)
                    {
                        return pwdNode.InnerText;
                    }
                    return "";
                }
                set
                {
                    XmlNode pwdNode =
                        settingsDoc.SelectSingleNode("Settings/Password");

                    string hashedPassword = CreateHashedPassword(value, null);
                    if (pwdNode != null)
                    {
                        pwdNode.InnerText = hashedPassword;
                    }
                    else
                    {
                        XmlNode settingsNode =
                            settingsDoc.SelectSingleNode("Settings");
                        XmlElement pwdElem =
                            settingsDoc.CreateElement("Password");
                        pwdElem.InnerText = hashedPassword;
                        settingsNode.AppendChild(pwdElem);
                    }
                }
            }

            // ハッシュ化されたパスワードを作成します。
            private string CreateHashedPassword(string password,
                                                byte[] existingSalt)
            {
                byte[] salt = null;
                if (existingSalt == null)
                {
                    // ランダムな大きさのsalt値を作成します
                    Random random = new Random();
                    int size = random.Next(16, 64);

                    // salt値の配列を作成します
                    salt = new byte[size];

                    // salt値をより厳密な乱数として設定します
                    RNGCryptoServiceProvider rng =
                            new RNGCryptoServiceProvider();
                    rng.GetNonZeroBytes(salt);
                }
                else
                    salt = existingSalt;

                // 文字列をbyte配列に変換します
                byte[] pwd = Encoding.UTF8.GetBytes(password);

                // パスワードとsalt値を保持する変数を作成します
                byte[] saltedPwd = new byte[pwd.Length + salt.Length];

                // 平文パスワードのbyte配列を最初に追加します
                pwd.CopyTo(saltedPwd, 0);
                // 次にsalt値を追加します
                salt.CopyTo(saltedPwd, pwd.Length);

                // SHA512アルゴリズムを使用してハッシュ値を計算します
                SHA512Managed sha512 = new SHA512Managed();

                // salt値付きの平文パスワードのハッシュ値を計算します
                byte[] hash = sha512.ComputeHash(saltedPwd);

                // ハッシュ値とsalt値を保持する変数を作成します
                byte[] hashWithSalt = new byte[hash.Length + salt.Length];

                // ハッシュ値とsalt値をそれぞれ追加します
                hash.CopyTo(hashWithSalt, 0);
                salt.CopyTo(hashWithSalt, hash.Length);

                // ハッシュ値とsalt値をbase64エンコードした文字列を返します
                return Convert.ToBase64String(hashWithSalt);
            }

            // 保存されているパスワードとの差異を比較します
            public bool IsPasswordValid(string password)
            {
                // 保存されているパスワードをbyte配列として取得します。
                // この配列にはsalt値とハッシュ化されたパスワードが含まれます。
                byte[] hashWithSalt = Convert.FromBase64String(Password);

                // ここではハッシュのサイズは512ビットです(SHA512)
                int hashSizeInBytes = 512 / 8;

                // 保存されているパスワードに対するsalt値を保持する変数を用意します
                int saltSize = hashWithSalt.Length - hashSizeInBytes;
                byte[] salt = new byte[saltSize];

                // salt値をコピーします
                Array.Copy(hashWithSalt, hashSizeInBytes, salt, 0, saltSize);

                // 入力されたパスワードに対するハッシュ値を計算します
                string passwordHash = CreateHashedPassword(password, salt);

                // 計算したハッシュ値が保存されているパスワードのハッシュと一致する場合、
                // 平文パスワードは保存されているパスワードと一致します。
                return (Password == passwordHash);
            }
        }

        #endregion

        #region "10.7 セキュリティ権限を安全な方法で設定する"
        public static void SafeAssert()
        {
            CallSecureFunctionSafelyAndEfficiently();
        }

        public static void CallSecureFunctionSafelyAndEfficiently()
        {

            // リフレクション経由時に非公開メンバへアクセスできるよう
            // 権限を設定します
            ReflectionPermission perm =
                new ReflectionPermission(ReflectionPermissionFlag.MemberAccess);

            // Assertメソッドを実行する前にDemandメソッドを実行することにより、
            // ユーザの持つアクセス権限を確認します。
            // Assertメソッドを実行するとスタックウォークによる権限の確認が行われなくなるため
            // Assertを実行する前にDemandを実行するべきです。
            // それによって安全性を向上させることができる他、
            // パフォーマンスの向上を図ることも可能です。
            perm.Demand();

            // メソッドを呼び出す前にAssertを実行することによって、呼び出すメソッド内において
            // Demandが実行された場合でもスタックウォークが保留されるようになります。
            // Assertメソッドを実行することによって
            // SecureFunctionの動作を保証できるようになります。
            perm.Assert();

            // 以下ではセキュリティの保証された処理を100回実行しています。
            // ただしスタック全体を100回検証するのではなく、呼び出し先と現在のメソッドの
            // スタックだけを100回確認しています。
            for (int i = 0; i < 100; i++)
            {
                SecureFunction();
            }
        }

        public static void SecureFunction()
        {
            // リフレクション経由時に非公開メンバへアクセスできるよう
            // 権限を設定します
            ReflectionPermission perm =
                new ReflectionPermission(ReflectionPermissionFlag.MemberAccess);

            // Demandを実行して、適切な権限があるかどうかスタックをさかのぼって確認します
            perm.Demand();

            // この位置で必要な処理を行います
        }

        #endregion

        #region "10.8 アセンブリが特定の権限を持つことを検証する"
        public static void VerifyAssemblyPerms()
        {
            // 以下では文字列「www.oreilly.com」を含むすべてのサイトにアクセスを許可するための
            // WebPermissionを設定するために、O'ReillyのウェブサイトのURLについての正規表現を
            // 作成しています。その後SecurityManagerクラスによって権限が正しく取得できたか
            // どうかを確認しています。
            Regex regex = new Regex(@"http://www\.oreilly\.com/.*");
            WebPermission webConnectPerm =
                new WebPermission(NetworkAccess.Connect, regex);
            if (SecurityManager.IsGranted(webConnectPerm))
            {
                // O'Reillyのサイトへ接続します
            }
        }
        #endregion

        #region "10.9 アセンブリの攻撃対象となり得る領域を最小限にする"
        public static void MinimizeAttackSurface()
        {
            Console.WriteLine("「方法 : RequestRefuse フラグを使用することにより、アクセス許可を拒否する」を参照してください。");
        }
        #endregion

        #region "10.10 セキュリティおよび監査情報を取得する"
        // ms-help://MS.VSCC.v80/MS.MSDNQTR.v80.en/MS.MSDN.v80/MS.NETDEVFX.v20.en/cpref/html/M_System_Security_AccessControl_ObjectSecurity_GetOwner_1_a6998a77.htm


        public static void TestViewFileRegRights()
        {
            ViewFileRights();
            ViewRegKeyRights();
        }

        public static void ViewFileRights()
        {
            // ファイルに対するセキュリティ情報を取得します
            string file = @"c:\WINDOWS\bootstat.dat";
            FileSecurity fileSec = File.GetAccessControl(file);
            DisplayFileSecurityInfo(fileSec);
        }

        public static void ViewRegKeyRights()
        {
            // レジストリに対するセキュリティ情報を取得します
            using (RegistryKey regKey =
                Registry.LocalMachine.OpenSubKey(@"SOFTWARE\MyCompany\MyApp"))
            {
                RegistrySecurity regSecurity = regKey.GetAccessControl();
                DisplayRegKeySecurityInfo(regSecurity);
            }
        }

        public static void DisplayRegKeySecurityInfo(RegistrySecurity regSec)
        {
            Console.WriteLine("GetSecurityDescriptorSddlForm:  " +
                regSec.GetSecurityDescriptorSddlForm(AccessControlSections.All));

            foreach (RegistryAccessRule ace in
                regSec.GetAccessRules(true, true, typeof(NTAccount)))
            {
                Console.WriteLine("\tIdentityReference.Value: " +
                    ace.IdentityReference.Value);
                Console.WriteLine("\tAccessControlType: " + ace.AccessControlType);
                Console.WriteLine("\tRegistryRights: " + ace.RegistryRights.ToString());
                Console.WriteLine("\tInheritanceFlags: " + ace.InheritanceFlags);
                Console.WriteLine("\tIsInherited: " + ace.IsInherited);
                Console.WriteLine("\tPropagationFlags: " + ace.PropagationFlags);

                Console.WriteLine("-----------------\r\n\r\n");
            }

            foreach (RegistryAuditRule ace in
                regSec.GetAuditRules(true, true, typeof(NTAccount)))
            {
                Console.WriteLine("\tIdentityReference.Value: " +
                    ace.IdentityReference.Value);
                Console.WriteLine("\tAuditFlags: " + ace.AuditFlags);
                Console.WriteLine("\tRegistryRights: " + ace.RegistryRights.ToString());
                Console.WriteLine("\tInheritanceFlags: " + ace.InheritanceFlags);
                Console.WriteLine("\tIsInherited: " + ace.IsInherited);
                Console.WriteLine("\tPropagationFlags: " + ace.PropagationFlags);

                Console.WriteLine("-----------------\r\n\r\n");
            }

            // NTAccountをSecurityIdentifierに変更することによってSIDを表示できます
            Console.WriteLine("GetGroup(typeof(NTAccount)).Value: " +
                regSec.GetGroup(typeof(NTAccount)).Value);
            Console.WriteLine("GetOwner(typeof(NTAccount)).Value: " +
                regSec.GetOwner(typeof(NTAccount)).Value);

            Console.WriteLine("---------------------------------------\r\n\r\n\r\n");
        }

        public static void DisplayFileSecurityInfo(FileSecurity fileSec)
        {
            Console.WriteLine("GetSecurityDescriptorSddlForm: {0}",
                fileSec.GetSecurityDescriptorSddlForm(AccessControlSections.All));

            foreach (FileSystemAccessRule ace in
                fileSec.GetAccessRules(true, true, typeof(NTAccount)))
            {
                Console.WriteLine("\tIdentityReference.Value: " +
                    ace.IdentityReference.Value);
                Console.WriteLine("\tAccessControlType: " + ace.AccessControlType);
                Console.WriteLine("\tFileSystemRights: " + ace.FileSystemRights);
                Console.WriteLine("\tInheritanceFlags: " + ace.InheritanceFlags);
                Console.WriteLine("\tIsInherited: " + ace.IsInherited);
                Console.WriteLine("\tPropagationFlags: " + ace.PropagationFlags);

                Console.WriteLine("-----------------\r\n\r\n");
            }

            foreach (FileSystemAuditRule ace in
                fileSec.GetAuditRules(true, true, typeof(NTAccount)))
            {
                Console.WriteLine("\tIdentityReference.Value: " +
                    ace.IdentityReference.Value);
                Console.WriteLine("\tAuditFlags: " + ace.AuditFlags);
                Console.WriteLine("\tFileSystemRights: " + ace.FileSystemRights);
                Console.WriteLine("\tInheritanceFlags: " + ace.InheritanceFlags);
                Console.WriteLine("\tIsInherited: " + ace.IsInherited);
                Console.WriteLine("\tPropagationFlags: " + ace.PropagationFlags);

                Console.WriteLine("-----------------\r\n\r\n");
            }

            Console.WriteLine("GetGroup(typeof(NTAccount)).Value: " +
                fileSec.GetGroup(typeof(NTAccount)).Value);
            Console.WriteLine("GetOwner(typeof(NTAccount)).Value: " +
                fileSec.GetOwner(typeof(NTAccount)).Value);

            Console.WriteLine("---------------------------------------\r\n\r\n\r\n");
        }
        #endregion

        #region "10.11 ファイルまたはレジストリキーのアクセス権限を設定または解除する"
        // ms-help://MS.VSCC.v80/MS.MSDNQTR.v80.en/MS.MSDN.v80/MS.NETDEVFX.v20.en/cpref/html/M_System_Security_AccessControl_FileSystemSecurity_AddAccessRule_1_609c1893.htm
        // ms-help://MS.VSCC.v80/MS.MSDNQTR.v80.en/MS.MSDN.v80/MS.NETDEVFX.v20.en/cpref/html/O_T_System_IO_File_GetAccessControl.htm
        // ms-help://MS.VSCC.v80/MS.MSDNQTR.v80.en/MS.MSDN.v80/MS.NETDEVFX.v20.en/cpref/html/M_System_IO_File_SetAccessControl_1_8d292e58.htm

        //  http://pluralsight.com/wiki/default.aspx/Keith.GuideBook/HowToProgramWithSIDs.html


        public static void TestGrantRevokeFileRights()
        {
            //GrantRevokeRegKeyRights();
            //GrantRevokeFileRights();
        }

        public static void GrantRevokeRegKeyRights()
        {
            NTAccount user = new NTAccount(@"WRKSTN\ST");

            using (RegistryKey regKey = Registry.LocalMachine.OpenSubKey(
                @"SOFTWARE\MyCompany\MyApp"))
            {
                GrantRegKeyRights(regKey, user, RegistryRights.WriteKey,
                    InheritanceFlags.None, PropagationFlags.None, AccessControlType.Allow);
                RevokeRegKeyRights(regKey, user, RegistryRights.WriteKey,
                    InheritanceFlags.None, PropagationFlags.None,
                    AccessControlType.Allow);
            }
        }

        public static void GrantRegKeyRights(RegistryKey regKey,
            NTAccount user,
            RegistryRights rightsFlags,
            InheritanceFlags inherFlags,
            PropagationFlags propFlags,
            AccessControlType actFlags)
        {
            RegistrySecurity regSecurity = regKey.GetAccessControl();

            DisplayRegKeySecurityInfo(regSecurity);

            RegistryAccessRule rule = new RegistryAccessRule(user, rightsFlags, inherFlags, propFlags, actFlags);
            regSecurity.AddAccessRule(rule);
            regKey.SetAccessControl(regSecurity);

            DisplayRegKeySecurityInfo(regSecurity);
        }

        public static void RevokeRegKeyRights(RegistryKey regKey,
            NTAccount user,
            RegistryRights rightsFlags,
            InheritanceFlags inherFlags,
            PropagationFlags propFlags,
            AccessControlType actFlags)
        {
            RegistrySecurity regSecurity = regKey.GetAccessControl();

            DisplayRegKeySecurityInfo(regSecurity);

            RegistryAccessRule rule = new RegistryAccessRule(user, rightsFlags, inherFlags, propFlags, actFlags);
            regSecurity.AddAccessRule(rule);

            //System.Security.Principal.IdentityNotMappedException was unhandled
            //      Message="Some or all identity references could not be translated."
            //--Only when account is DISABLED

            regKey.SetAccessControl(regSecurity);

            DisplayRegKeySecurityInfo(regSecurity);
        }

        public static void GrantRevokeFileRights()
        {
            NTAccount user = new NTAccount(@"WRKSTN\ST");

            string file = @"C:\FOO.TXT";
            GrantFileRights(file, user, FileSystemRights.Delete, InheritanceFlags.None,
                PropagationFlags.None, AccessControlType.Allow);
            RevokeFileRights(file, user, FileSystemRights.Delete, InheritanceFlags.None,
                PropagationFlags.None, AccessControlType.Allow);
        }

        public static void GrantFileRights(string file,
            NTAccount user, FileSystemRights rightsFlags, InheritanceFlags inherFlags,
            PropagationFlags propFlags, AccessControlType actFlags)
        {
            FileSecurity fileSecurity = File.GetAccessControl(file);

            DisplayFileSecurityInfo(fileSecurity);

            FileSystemAccessRule rule = new FileSystemAccessRule(user, rightsFlags,
                inherFlags, propFlags, actFlags);
            fileSecurity.AddAccessRule(rule);
            File.SetAccessControl(file, fileSecurity);

            DisplayFileSecurityInfo(fileSecurity);
        }

        public static void RevokeFileRights(string file,
            NTAccount user, FileSystemRights rightsFlags, InheritanceFlags inherFlags,
            PropagationFlags propFlags, AccessControlType actFlags)
        {
            FileSecurity fileSecurity = File.GetAccessControl(file);

            DisplayFileSecurityInfo(fileSecurity);

            FileSystemAccessRule rule = new FileSystemAccessRule(user, rightsFlags,
                inherFlags, propFlags, actFlags);
            fileSecurity.RemoveAccessRuleSpecific(rule);
            File.SetAccessControl(file, fileSecurity);

            DisplayFileSecurityInfo(fileSecurity);
        }
        #endregion

        #region "10.12 文字列をSecureStringによって保護する"
        public static void TestSecureString()
        {
            StreamReader sr = new StreamReader("Categories.xml");
            SecureString secretStr = CreateSecureString(sr);

            // この文字列は変更できません:  secretStr.AppendChar('x');

            Console.WriteLine("secretStr = " + secretStr.ToString());
            ReadSecureString(secretStr);
        }

        public static SecureString CreateSecureString(StreamReader secretStream)
        {
            SecureString secretStr = new SecureString();
            char[] buf = new char[1];

            while (secretStream.Peek() >= 0)
            {
                int nextByte = secretStream.Read(buf, 0, 1);
                secretStr.AppendChar(buf[0]);
            }

            // secretStrオブジェクトを読み取り専用にします
            secretStr.MakeReadOnly();

            return secretStr;
        }

        public static void ReadSecureString(SecureString secretStr)
        {
            // SecureStringから元の文字列を取得するためには
            // 特別なメソッドを使用する必要があります
            IntPtr secretStrPtr = Marshal.SecureStringToBSTR(secretStr);
            string nonSecureStr = Marshal.PtrToStringBSTR(secretStrPtr);

            // 保護されていない文字列を使用します
            Console.WriteLine("nonSecureStr = " + nonSecureStr);

            Marshal.ZeroFreeBSTR(secretStrPtr);

            if (!secretStr.IsReadOnly())
            {
                secretStr.Clear();
            }
        }

        #endregion

        #region "10.13 Streamデータを保護する"
        // ms-help://MS.VSCC.v80/MS.MSDNQTR.v80.en/MS.MSDN.v80/MS.NETDEVFX.v20.en/cpref/html/T_System_Net_Security_SslStream.htm

        //http://www.leastprivilege.com/CategoryView.aspx?category=Samples
        //http://www.devnewsgroups.net/group/microsoft.public.dotnet.framework/topic11749.aspx
        //      http://www.devnewsgroups.net/link.aspx?url=http://blogs.msdn.com/jhoward/archive/2005/02/02/365323.aspx
        //      http://msdn.microsoft.com/library/default.asp?url=/library/en-us/cptools/html/cpgrfcertificatecreationtoolmakecertexe.asp
        //      http://www.inventec.ch/chdh/notes/14.htm
        #endregion

        #region "10.14 web.configファイルの情報を暗号化する"
        public static void TestEncryptDecryptWebConfigSection()
        {
            DisplayWebConfigInfo();

            EncryptWebConfigSection("/WebApplication1", "appSettings", "DataProtectionConfigurationProvider");

            DisplayWebConfigInfo();

            DecryptWebConfigSection("/WebApplication1", "appSettings");

            DisplayWebConfigInfo();
        }

        //http://www.ondotnet.com/pub/a/dotnet/2005/02/15/encryptingconnstring.html
        //http://www.devx.com/dotnet/Article/27562/0/page/2
        //http://beta.asp.net/QUICKSTARTV20/aspnet/doc/management/mgmtapi.aspx
        //http://www.developer.com/net/vb/article.php/10926_3500906_1

        //ms-help://MS.VSCC.v80/MS.MSDNQTR.v80.en/MS.MSDN.v80/MS.VisualStudio.v80.en/dv_aspnetcon/html/e1652f90-eac5-4f51-bff1-cf1acc2e1180.htm



        public static void EncryptWebConfigSection(string appPath,
            string protectedSection,
            string dataProtectionProvider)
        {
            System.Configuration.Configuration webConfig =
                WebConfigurationManager.OpenWebConfiguration(appPath);
            ConfigurationSection webConfigSection =
                webConfig.GetSection(protectedSection);

            if (!webConfigSection.SectionInformation.IsProtected)
            {
                webConfigSection.SectionInformation.ProtectSection(
                    dataProtectionProvider);
                webConfig.Save();
            }
        }

        public static void DecryptWebConfigSection(string appPath,
            string protectedSection)
        {
            System.Configuration.Configuration webConfig =
                WebConfigurationManager.OpenWebConfiguration(appPath);
            ConfigurationSection webConfigSection =
                webConfig.GetSection(protectedSection);

            if (webConfigSection.SectionInformation.IsProtected)
            {
                webConfigSection.SectionInformation.UnprotectSection();
                webConfig.Save();
            }
        }

        public static void DisplayWebConfigInfo()
        {
            //Configuration config = ConfigurationManager.OpenExeConfiguration(ConfigurationUserLevel.None, @"");
            System.Configuration.Configuration webConfig = WebConfigurationManager.OpenWebConfiguration("/WebApplication1");  //"http://localhost/WebApplication1/web.config");   //WebForm1.aspx

            Console.WriteLine("\r\n\r\n------------------------------------");
            Console.WriteLine("FilePath: " + webConfig.FilePath);
            Console.WriteLine("APP_SETTINGS");
            foreach (KeyValueConfigurationElement o in webConfig.AppSettings.Settings)
            {
                Console.WriteLine("\tKEY: " + o.Key);
                Console.WriteLine("\tVALUE: " + o.Value);
            }
            Console.WriteLine("APP_CONN_STRS");
            foreach (ConnectionStringSettings css in webConfig.ConnectionStrings.ConnectionStrings)
            {
                Console.WriteLine("\tNAME: " + css.Name);
                Console.WriteLine("\tCONN_STR: " + css.ConnectionString);
            }
            Console.WriteLine("APP_SECTION_GROUP");
            foreach (ConfigurationSectionGroup sg in webConfig.SectionGroups)
            {
                Console.WriteLine("\tNAME: " + sg.Name);
                //Console.WriteLine("\tGROUP_NAME: " + sg.SectionGroupName);
                //Console.WriteLine("\tTYPE: " + sg.Type);
            }
            Console.WriteLine("SECTIONS");
            foreach (ConfigurationSection cs in webConfig.Sections)
            {
                Console.WriteLine("\tNAME: " + cs.SectionInformation.Name);
                //Console.WriteLine("\tSECTION_NAME: " + cs.SectionInformation.SectionName);
                //Console.WriteLine("\tSOURCE: " + cs.ElementInformation.Source);
            }
        }
        #endregion

        #region "10.15 SecurityExceptionが発生する原因を把握する"
        // SecurityExceptionクラスの各プロパティに関するドキュメントを参照してください
        // ms-help://MS.VSCC.v80/MS.MSDNQTR.v80.en/MS.MSDN.v80/MS.NETDEVFX.v20.en/cpref/html/T_System_Security_SecurityException.htm

        //http://blogs.msdn.com/shawnfa/archive/2004/07/30/202468.aspx
        //http://blog.hundhausen.com/ASPNET+20+SecurityException.aspx


        public static void TestSecurityException()
        {
            RunCode();
            Console.WriteLine("ENTERキーを押して終了します。");
            Console.Read();
        }

        public static void RunCode()
        {
            try
            {
                // 権限を拒否します。
                RegistryPermission regPerm = new RegistryPermission(RegistryPermissionAccess.Read, "HKEY_LOCAL_MACHINE\\Software\\MyApp");
                regPerm.Deny();

                // 拒否された権限に対してDemandを実行することにより例外が発生するため、
                // 各プロパティの値を表示します。
                Display("拒否された権限に対してDemandを実行します。\n\n");
                DemandDeniedPermission();
                Display("************************************************\n");
                CodeAccessPermission.RevertDeny();

                // アセンブリレベルで禁止された属性により拒否された権限に対して
                // Demandを実行します。
                Display("禁止された権限に対してDemandを実行します。\n\n");
                DemandRefusedPermission();
                Display("************************************************\n");

                // PermitOnlyによって暗黙的に禁止された権限に対してDemandを実行します。
                // PermitOnlyメソッドが呼び出されることにより、アクセス可能なリソースが制限されるため、
                // スタックウォークが失敗します。また、以下では失敗したことを表示するために必要な権限を設定しています。
                PermissionSet permitOnly = new PermissionSet(PermissionState.None);
                permitOnly.AddPermission(new KeyContainerPermission(KeyContainerPermissionFlags.Import));
                permitOnly.AddPermission(new SecurityPermission(SecurityPermissionFlag.ControlEvidence |
                    SecurityPermissionFlag.ControlPolicy |
                    SecurityPermissionFlag.SerializationFormatter));
                permitOnly.PermitOnly();
                Display("暗黙的に禁止された権限に対してDemandを実行します。\n\n");
                DemandPermitOnly();
            }
            catch (Exception sE)
            {
                Display("************************************************\n");
                Display("例外情報をToStringメソッドにより表示します：");
                Display(sE.ToString());
            }
        }

        public static void DemandDeniedPermission()
        {
            try
            {
                RegistryPermission regPerm = new RegistryPermission(RegistryPermissionAccess.Read, "HKEY_LOCAL_MACHINE\\Software\\MyApp");
                regPerm.Demand();
            }
            catch (SecurityException sE)
            {
                Display("拒否された権限：" + ((PermissionSet)sE.DenySetInstance).ToString());
                Display("要求された権限：" + sE.Demanded.ToString());
                Display("SecurityAction：" + sE.Action.ToString());
                Display("メソッド：" + sE.Method);
                Display("例外発生時のアクセス許可の状態：" + sE.PermissionState);
                Display("失敗した権限：" + (IPermission)sE.FirstPermissionThatFailed);
                Display("失敗した権限の型：" + sE.PermissionType.ToString());
                Display("GetObjectDataの使用例：");
                SerializationInfo si = new SerializationInfo(typeof(Security), new FormatterConverter());
                sE.GetObjectData(si, new StreamingContext(StreamingContextStates.All));
                Display("GetObjectDataによって取得したFirstPermissionThatFailedプロパティの値：");
                Display(si.GetString("FirstPermissionThatFailed"));
            }
        }

        public static void DemandRefusedPermission()
        {
            try
            {
                RegistryPermission regPerm = new RegistryPermission(RegistryPermissionAccess.Read, "HKEY_LOCAL_MACHINE\\Software\\MyApp");
                regPerm.Demand();
            }
            catch (SecurityException sE)
            {
                Display("禁止された権限セット：" + (sE.RefusedSet).ToString());
                Display("例外メッセージ：" + sE.Message);
                Display("失敗したアセンブリ：" + sE.FailedAssemblyInfo.EscapedCodeBase);
                Display("アセンブリに与えられている権限：\n" + sE.GrantedSet);
                Display("失敗した権限：" + sE.FirstPermissionThatFailed);
                Display("失敗した権限の型：" + sE.PermissionType.ToString());
                Display("ソース：" + sE.Source);
            }
        }

        public static void DemandPermitOnly()
        {
            try
            {
                RegistryPermission regPerm = new RegistryPermission(RegistryPermissionAccess.Read, "HKEY_LOCAL_MACHINE\\Software\\MyApp");
                regPerm.Demand();
            }
            catch (SecurityException sE)
            {
                Display("許可された権限：" + ((PermissionSet)sE.PermitOnlySetInstance).ToString());
                Display("要求された権限：" + sE.Demanded.ToString());
                Display("SecurityAction：" + sE.Action.ToString());
                Display("メソッド：" + sE.Method.ToString());
                Display("例外発生時のアクセス許可の状態：" + sE.PermissionState);
                Display("失敗した権限：" + (IPermission)sE.FirstPermissionThatFailed);
                Display("失敗した権限の型：" + sE.PermissionType.ToString());

                // 以下ではSecurityExceptionのインスタンスを作成する例として
                // 例外を再スローしています
                Display("PermitOnlyセキュリティアクションの結果として例外を再スローします。");
                throw new SecurityException(sE.Message, sE.DenySetInstance, sE.PermitOnlySetInstance,
                                            sE.Method, sE.Demanded, (IPermission)sE.FirstPermissionThatFailed);
            }
        }
        public static void Display(string line)
        {
            Console.WriteLine(line);
        }
        #endregion

        #region "10.16 Unicodeエンコーディングを安全に処理する"
        public static void TestUnicodeEncodingWithSecurity()
        {
            byte[] SourceArray2 = {128, 0, 83, 0, 111, 0, 117, 0, 114, 0, 99, 0, 
			                       101, 0, 32, 0, 83, 0, 116, 0, 114, 0, 105, 0, 
                                   110, 0, 103, 0, 128, 0};

            Console.WriteLine("\r\nFromUnicodeByteArrayを実行します...");
            Console.WriteLine(FromUnicodeByteArray(SourceArray2));

            Console.WriteLine("\r\nToUnicodeByteArrayを実行します...");
            string SourceStr = "元データ文字列";
            foreach (byte b in ToUnicodeByteArray(SourceStr))
                Console.WriteLine(b);

            Console.WriteLine();
        }


        public static string FromUnicodeByteArray(byte[] characters)
        {
            UnicodeEncoding encoding = new UnicodeEncoding(false, true, true);
            string constructedString = encoding.GetString(characters);

            return constructedString;
        }

        public static byte[] ToUnicodeByteArray(string characters)
        {
            UnicodeEncoding encoding = new UnicodeEncoding(false, true, true);
            int numberOfChars = encoding.GetByteCount(characters);
            byte[] retArray = new byte[numberOfChars];

            retArray = encoding.GetBytes(characters);

            return retArray;
        }
        #endregion

        #region "10.17 ファイルハンドルをより安全に操作する"
        //ms-help://MS.VSCC.v80/MS.MSDNQTR.v80.en/MS.MSDN.v80/MS.NETDEVFX.v20.en/cpref/html/T_System_Runtime_InteropServices_SafeHandle.htm
        #endregion
    }
}
