﻿using System;
using System.Diagnostics;

namespace CSharpRecipes
{
    public static class NumbersEnums
    {
        #region "12.1 度とラジアンの変換"
        public static double ConvertDegreesToRadians(double degrees)
        {
            double radians = (Math.PI / 180) * degrees;
            return radians;
        }

        public static double ConvertRadiansToDegrees(double radians)
        {
            double degrees = (180 / Math.PI) * radians;
            return degrees;
        }
        #endregion

        #region "12.2 さまざまなデータ型におけるビットごとの補数演算"
        public static void TestBitwiseOperators()
        {
            uint x = 0x00000001;
            Console.WriteLine("~x = " + ~x);

            sbyte B1 = sbyte.MinValue;
            sbyte B2 = sbyte.MaxValue;
            Console.WriteLine("B1|B2 = " + (((byte)B1 | (byte)B2)));

            ushort x2 = 0x00000001;		// Problem
            Console.WriteLine("~x2 = " + ~x2);

            byte y = 1;		// Problem
            //byte B = ~y;
            Console.WriteLine("~y = " + ~y);

            char x3 = (char)1;		// Problem
            Console.WriteLine("~x3 = " + ~x3);

            sbyte x5 = 1;
            Console.WriteLine("~x5 = " + ~x5);

            uint IntResult = (uint)~x;
            Console.WriteLine("IntResult = " + IntResult);

            byte ByteResult = (byte)~y;
            Console.WriteLine("ByteResult = " + ByteResult);
        }
        #endregion

        #region "12.3 n進数と10進数との変換"
        public static void TestBase10()
        {
            string base2 = "11";
            string base8 = "17";
            string base10 = "110";
            string base16 = "11FF";

            Console.WriteLine("Convert.ToInt32(base2, 2) = " +
                Convert.ToInt32(base2, 2));

            Console.WriteLine("Convert.ToInt32(base8, 8) = " +
                Convert.ToInt32(base8, 8));

            Console.WriteLine("Convert.ToInt32(base10, 10) = " +
                Convert.ToInt32(base10, 10));

            Console.WriteLine("Convert.ToInt32(base16, 16) = " +
                Convert.ToInt32(base16, 16));
        }
        #endregion

        #region "12.4 数値を表す文字列の有効性の確認"
        public static void TestDetermineIfStringIsNumber()
        {
            string IsNotNumber = "111west";
            string IsNum = "  +111  ";
            string IsFloat = "  23.11  ";
            string IsExp = "  +23 e+11  ";

            int i = 0;
            float f = 0;

            Console.WriteLine(int.TryParse(IsNum, out i));      // 111  // .Net Framework 1.1 では機能しません
            Console.WriteLine(float.TryParse(IsNum, out f));    // 111
            Console.WriteLine(float.TryParse(IsFloat, out f));  // 23.11
            //Console.WriteLine(float.Parse(IsExp));            // 例外発生


            Console.WriteLine(IsInt(IsNum));        // True
            Console.WriteLine(IsInt(IsNotNumber));  // False
            Console.WriteLine(IsInt(IsFloat));      // False
            Console.WriteLine(IsInt(IsExp));        // False
            Console.WriteLine();


            Console.WriteLine(IsDoubleFromTryParse(IsNum));         // True
            Console.WriteLine(IsDoubleFromTryParse(IsNotNumber));   // False
            Console.WriteLine(IsDoubleFromTryParse(IsFloat));       // True
            Console.WriteLine(IsDoubleFromTryParse(IsExp));         // False
            Console.WriteLine();
        }


        public static bool IsInt(string value)
        {
            try
            {
                value = value.Trim();
                int foo = int.Parse(value);
                return true;
            }
            catch (FormatException e)
            {
                Console.WriteLine("整数ではありません：{0}", e.ToString());
                return false;
            }
        }
        public static bool IsDoubleFromTryParse(string value)
        {
            double result = 0;
            return double.TryParse(value, System.Globalization.NumberStyles.Float,
                System.Globalization.NumberFormatInfo.CurrentInfo, out result);
        }
        #endregion

        #region "12.5 浮動小数点数値の丸め込み"
        public static void TestRound()
        {
            int X = (int)Math.Round(2.5555);
            Console.WriteLine(X);
            Console.WriteLine(Math.Round(2.5555, 2));
            Console.WriteLine(Math.Round(2.444444, 3));
            Console.WriteLine(Math.Round(2.555555555555555555555555555550001));
            Console.WriteLine(Math.Round(.5));
            Console.WriteLine(Math.Round(1.5));
            Console.WriteLine(Math.Round(2.5));
            Console.WriteLine(Math.Round(3.5));
            Console.WriteLine();

            Console.WriteLine(Math.Floor(.5));
            Console.WriteLine(Math.Floor(1.5));
            Console.WriteLine(Math.Floor(2.5));
            Console.WriteLine(Math.Floor(3.5));
            Console.WriteLine();

            Console.WriteLine(Math.Ceiling(.5));
            Console.WriteLine(Math.Ceiling(1.5));
            Console.WriteLine(Math.Ceiling(2.5));
            Console.WriteLine(Math.Ceiling(3.5));
            Console.WriteLine();

            Console.WriteLine(RoundUp(.4));
            Console.WriteLine(RoundUp(.5));
            Console.WriteLine(RoundUp(.6));
            Console.WriteLine(RoundUp(1.4));
            Console.WriteLine(RoundUp(1.5));
            Console.WriteLine(RoundUp(1.6));
            Console.WriteLine(RoundUp(2.4));
            Console.WriteLine(RoundUp(2.5));
            Console.WriteLine(RoundUp(2.6));
            Console.WriteLine(RoundUp(3.4));
            Console.WriteLine(RoundUp(3.5));
            Console.WriteLine(RoundUp(3.6));
            Console.WriteLine();

            Console.WriteLine(RoundDown(.4));
            Console.WriteLine(RoundDown(.5));
            Console.WriteLine(RoundDown(.6));
            Console.WriteLine(RoundDown(1.4));
            Console.WriteLine(RoundDown(1.5));
            Console.WriteLine(RoundDown(1.6));
            Console.WriteLine(RoundDown(2.4));
            Console.WriteLine(RoundDown(2.5));
            Console.WriteLine(RoundDown(2.6));
            Console.WriteLine(RoundDown(3.4));
            Console.WriteLine(RoundDown(3.5));
            Console.WriteLine(RoundDown(3.6));
        }
        #endregion

        #region "12.6 丸め込みアルゴリズムの選択"
        public static double RoundUp(double valueToRound)
        {
            return Math.Floor(valueToRound + 0.5);
        }

        public static double RoundDown(double valueToRound)
        {
            double floorValue = Math.Floor(valueToRound);
            if ((valueToRound - floorValue) > .5)
            {
                return floorValue + 1;
            }
            else
            {
                return floorValue;
            }
        }
        #endregion

        #region "12.7 温度単位の変換"
        public static double CelsiusToFahrenheit(double celsius)
        {
            return (1.8 * celsius) + 32;
        }

        public static double FahrenheitToCelsius(double fahrenheit)
        {
            return 1.8 * (fahrenheit - 32);
        }

        public static double CelsiusToKelvin(double celsius)
        {
            return celsius + 273;
        }

        public static double KelvinToCelsius(double kelvin)
        {
            return kelvin - 273;
        }

        public static double FahrenheitToKelvin(double fahrenheit)
        {
            return CelsiusToKelvin(FahrenheitToCelsius(fahrenheit));
        }

        public static double KelvinToFahrenheit(double kelvin)
        {
            return CelsiusToFahrenheit(KelvinToCelsius(kelvin));
        }
        #endregion

        #region "12.8 安全な数値の縮小変換"
        public static void TestNarrowing()
        {
            // 2つの変数の宣言および初期化
            int sourceValue = 34000;
            short destinationValue = 0;

            // sourceValueがshortにキャストされる際に情報の欠落が起きるかどうかを確認します
            if (sourceValue <= short.MaxValue && sourceValue >= short.MinValue)
            {
                destinationValue = (short)sourceValue;
            }
            else
            {
                // アプリケーションに、情報の欠落が起きることを知らせます
            }

            long lhs = 34000;
            long rhs = long.MaxValue;
            lhs.AddChecked(rhs);
        }

        public static void AddChecked(this long lhs, long rhs)
        {
            int result = checked((int)(lhs + rhs));
        }
        #endregion

        #region "12.9 列挙体メンバの文字列での表示"
        public static void TestDisplayEnumValue()
        {
            Console.WriteLine(Shapes.Circle.ToString());
            Console.WriteLine(Shapes.Circle.ToString("G"));
            Console.WriteLine(Shapes.Circle.ToString("D"));
            Console.WriteLine(Shapes.Circle.ToString("F"));
            Console.WriteLine(Shapes.Circle.ToString("X"));

            Console.WriteLine();

            Shapes shapeStyle = Shapes.Cylinder;
            Console.WriteLine(shapeStyle.ToString());
            Console.WriteLine(shapeStyle.ToString("G"));
            Console.WriteLine(shapeStyle.ToString("D"));
            Console.WriteLine(shapeStyle.ToString("F"));
            Console.WriteLine(shapeStyle.ToString("X"));

            Console.WriteLine();

            shapeStyle = Shapes.Circle | Shapes.Cylinder;
            Console.WriteLine(shapeStyle.ToString());
            Console.WriteLine(shapeStyle.ToString("G"));
            Console.WriteLine(shapeStyle.ToString("D"));
            Console.WriteLine(shapeStyle.ToString("F"));
            Console.WriteLine(shapeStyle.ToString("X"));


            IceCreamToppings toppings =
                IceCreamToppings.HotFudge | IceCreamToppings.WhippedCream;

            Console.WriteLine(toppings.ToString());
            Console.WriteLine(toppings.ToString("G"));
            Console.WriteLine(toppings.ToString("D"));
            Console.WriteLine(toppings.ToString("F"));
            Console.WriteLine(toppings.ToString("X"));

        }
        #endregion

        #region "12.10 標準文字列の相当する列挙型への変換"
        public static void TestConvertingEnums()
        {
            try
            {
                Language proj1Language = (Language)Enum.Parse(typeof(Language), "VBNET");
                //Language proj2Language = (Language)Enum.Parse(typeof(Language), "UnDefined");

                proj1Language = (Language)Enum.Parse(typeof(Language), "1");
                proj1Language = (Language)Enum.Parse(typeof(Language), "CSharp, VBNET");
            }
            catch (ArgumentException e)
            {
                // 適切でない文字列値(「未定義」文字列など)はここで処理します
                Console.WriteLine(e);
            }
        }
        #endregion

        #region "12.11 列挙型の有効性の確認"
        public static void HandleEnum(Language language)
        {
            if (CheckLanguageEnumValue(language))
            {
                // ここでlanguageを使用します
                Console.WriteLine("OK");
            }
            else
            {
                // languageが不正な値だった場合の処理を行います
                Console.WriteLine("NOT OK");
            }
        }

        public static bool CheckLanguageEnumValue(Language language)
        {
            switch (language)
            {
                // ここで適正な列挙型のメンバをすべて列挙します。
                // これは、列挙体に含まれるすべてのメンバではなく、
                // ここで特定した値のみが有効と判断されることを意味します
                case Language.CSharp:
                case Language.Other:
                case Language.VB6:
                case Language.VBNET:
                    break;
                default:
                    Debug.Assert(false, language + " は引数として適正な列挙体メンバではありません");
                    return false;
            }
            return true;
        }
        #endregion

        #region "12.12 Flags属性を付与された列挙型の値の有効性の確認"
        public static bool ValidateFlagsEnum(IceCreamToppings topping)
        {
            return ((topping > 0) && ((topping & IceCreamToppings.All) == topping));
        }

        public static bool ValidateFlagsEnum(int topping)
        {
            return ((topping > 0) && ((topping & (int)IceCreamToppings.All) == topping));
        }
        #endregion

        #region "12.13 列挙型メンバのビットマスクとしての使用"
        // 本書のレシピ12.13を参照してください
        #endregion

        #region "12.14 1つ以上の列挙型フラグの状態確認"
        // 本書のレシピ12.14を参照してください

        [Flags]
        public enum LanguageFlags
        {
            CSharp = 0x0001, VBNET = 0x0002, VB6 = 0x0004, Cpp = 0x0008,
            AllLanguagesExceptCSharp = VBNET | VB6 | Cpp
        }

        public static void TestEnumFlags()
        {
            LanguageFlags lang = LanguageFlags.CSharp | LanguageFlags.VBNET;

            if ((lang & LanguageFlags.CSharp) == LanguageFlags.CSharp)
            {
                Console.WriteLine("lang には少なくとも Language.CSharp が含まれています");
            }

            if (lang == LanguageFlags.CSharp)
            {
                Console.WriteLine("lang には少なくとも Language.CSharp が含まれています");
            }

            if ((lang > 0) && ((lang & (LanguageFlags.CSharp | LanguageFlags.VBNET)) ==
               (LanguageFlags.CSharp | LanguageFlags.VBNET)))
            {
                Console.WriteLine("lang には少なくとも Language.CSharp および Language.VBNET が含まれています");
            }

            if ((lang > 0) && ((lang | (LanguageFlags.CSharp | LanguageFlags.VBNET)) ==
               (LanguageFlags.CSharp | LanguageFlags.VBNET)))
            {
                Console.WriteLine("lang には Language.CSharp および Language.VBNET だけが含まれています");
            }

            lang = LanguageFlags.CSharp;
            if ((lang & LanguageFlags.CSharp) == LanguageFlags.CSharp)
            {
                //Language_1_21.CSharp      0001
                //lang                 0001
                //ANDed bit values     0001

                Console.WriteLine("AND演算によってCSharpが見つかりました");
            }

            lang = LanguageFlags.CSharp;
            if ((lang > 0) && (LanguageFlags.CSharp == (lang | LanguageFlags.CSharp)))
            {
                // OR演算によってCSharpが見つかりました
            }

            lang = LanguageFlags.CSharp | LanguageFlags.VB6 | LanguageFlags.Cpp;
            if ((lang > 0) && (LanguageFlags.CSharp == (lang | LanguageFlags.CSharp)))
            {
                // OR演算によってCSharpが見つかりました
            }

            lang = LanguageFlags.VBNET | LanguageFlags.VB6 | LanguageFlags.Cpp;
            if ((lang > 0) && (LanguageFlags.CSharp == (lang | LanguageFlags.CSharp)))
            {
                // OR演算によってCSharpが見つかりました
            }

            lang = LanguageFlags.VBNET;
            if ((lang & LanguageFlags.CSharp) == LanguageFlags.CSharp)
            {
                //Language_1_21.CSharp      0001
                //lang                 0010
                //ANDed bit values     0000
                Console.WriteLine("CSharp found in AND comparison (VBNET value)");
            }

            lang = LanguageFlags.CSharp;

            if (lang == LanguageFlags.CSharp)
            {
                //Language_1_21.CSharp      0001
                //lang                 0001
                //ORed bit values      0001

            }

            if ((lang > 0) && (LanguageFlags.CSharp == (lang | LanguageFlags.CSharp)))
            {
            }

            lang = LanguageFlags.CSharp | LanguageFlags.Cpp | LanguageFlags.VB6;

            if (lang == LanguageFlags.CSharp)
            {
                //Language_1_21.CSharp      0001
                //lang                 1101
                //ORed bit values      1101
            }

            lang = LanguageFlags.CSharp | LanguageFlags.VBNET;
            if ((lang > 0) && ((lang & (LanguageFlags.CSharp | LanguageFlags.VBNET)) ==
               (LanguageFlags.CSharp | LanguageFlags.VBNET)))
            {
                // 複数のビットを組み合わせて、それぞれがすべてオンであり、かつその他がオフであることを確認できます
                Console.WriteLine("CSharp および VBNET が見つかりました");
            }

            // Cppの有無を確認します
            lang = LanguageFlags.CSharp | LanguageFlags.VBNET | LanguageFlags.Cpp;
            if ((lang > 0) && ((lang & (LanguageFlags.CSharp | LanguageFlags.VBNET)) ==
               (LanguageFlags.CSharp | LanguageFlags.VBNET)))
            {
                // 複数のビットを組み合わせて、それぞれがすべてオンであり、その他のフラグの状態を確認しないこともできます
                Console.WriteLine("少なくとも CSharp と VBNET が見つかりました");
            }

            lang = LanguageFlags.CSharp | LanguageFlags.VBNET;
            if ((lang > 0) && ((lang | (LanguageFlags.CSharp | LanguageFlags.VBNET)) ==
               (LanguageFlags.CSharp | LanguageFlags.VBNET)))
            {
                // 特定のフラグのいずれかがオンであることを確認できます
                Console.WriteLine("CSharp または VBNET のいずれかが見つかりました");
            }

            lang = LanguageFlags.CSharp | LanguageFlags.VBNET | LanguageFlags.Cpp;
            if ((lang > 0) && ((lang | (LanguageFlags.CSharp | LanguageFlags.VBNET)) ==
               (LanguageFlags.CSharp | LanguageFlags.VBNET)))
            {
                // 特定のフラグのいずれかがオンであることを確認できます
                Console.WriteLine("CSharp またｈ VBNET のいずれかが見つかりました");
            }

            if ((lang > 0) && (lang | LanguageFlags.AllLanguagesExceptCSharp) ==
                LanguageFlags.AllLanguagesExceptCSharp)
            {
                Console.WriteLine("CSharp は指定されていません");
            }

        }
        #endregion

        #region "12.15 DecimalやDoubleの整数部分の確認"
        public static void TestTruncate()
        {
            decimal pi = (decimal)System.Math.PI;
            decimal decRet = System.Math.Truncate(pi);
            double trouble = 5.555;
            double dblRet = System.Math.Truncate(trouble);
        }
        #endregion
    }

    #region "列挙体の定義"
    [Flags]
    public enum Shapes
    {
        Square = 0, Circle = 1, Cylinder = 2, Octagon = 4
    }

    [Flags]
    public enum IceCreamToppings
    {
        Sprinkles = 0,
        HotFudge = 1,
        Cherry = 2,
        WhippedCream = 4,
        All = Sprinkles | HotFudge | Cherry | WhippedCream
    }

    public enum Language
    {
        Other = 0, CSharp = 1, VBNET = 2, VB6 = 3,
        All = (Other | CSharp | VBNET | VB6)
    }

    [Flags]
    public enum RecycleItems
    {
        Glass = 0x01,
        AluminumCans = 0x02,
        MixedPaper = 0x04,
        Newspaper = 0x08,
        TinCans = 0x10,
        Cardboard = 0x20,
        ClearPlastic = 0x40,
        All = (Glass | AluminumCans | MixedPaper | Newspaper | TinCans | Cardboard | ClearPlastic)
    }
    #endregion
}
