﻿using System;
using System.Net;
using System.Net.Security;
using System.Net.Sockets;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading;

namespace ConsoleTCPServer_SSL
{
    class TCPServer_SSL
    {
        private TcpListener _listener = null;
        private IPAddress _address = IPAddress.Parse("127.0.0.1");
        private int _port = 55555;

        #region コンストラクタ
        public TCPServer_SSL()
        {
        }

        public TCPServer_SSL(string address, string port)
        {
            _port = Convert.ToInt32(port);
            _address = IPAddress.Parse(address);
        }
        #endregion // コンストラクタ

        #region プロパティ
        public IPAddress Address
        {
            get { return _address; }
            set { _address = value; }
        }

        public int Port
        {
            get { return _port; }
            set { _port = value; }
        }
        #endregion

        public void Listen()
        {
            try
            {
                _listener = new TcpListener(_address, _port);
                // サーバを起動します
                _listener.Start();

                // リッスン用のループへ入ります
                while (true)
                {
                    Console.Write("通信相手を探しています...");

                    // 接続を待機します
                    TcpClient newClient = _listener.AcceptTcpClient();
                    Console.WriteLine("新しいクライアントが接続しました");

                    // クライアント用のスレッドを開始します
                    ThreadPool.QueueUserWorkItem(new WaitCallback(ProcessClient),
                        newClient);
                }
            }
            catch (SocketException e)
            {
                Console.WriteLine("SocketException: {0}", e);
            }
            finally
            {
                // サーバを停止します
                _listener.Stop();
            }

            Console.WriteLine("\n続行するには何かキーを押してください . . .");
            Console.Read();
        }

        private void ProcessClient(object client)
        {
            using (TcpClient newClient = (TcpClient)client)
            {
                // データを読み込むためのバッファを作成します
                byte[] bytes = new byte[1024];
                string clientData = null;


                using (SslStream sslStream = new SslStream(newClient.GetStream()))
                {
                    sslStream.AuthenticateAsServer(GetServerCert("MyTestCert2"), false,
                        SslProtocols.Default, true);

                    // クライアントからのすべてのデータを受信するためのループです
                    int bytesRead = 0;
                    while ((bytesRead = sslStream.Read(bytes, 0, bytes.Length)) != 0)
                    {
                        // データをASCII文字列に変換します
                        clientData = Encoding.ASCII.GetString(bytes, 0, bytesRead);
                        Console.WriteLine("クライアントの発言：{0}", clientData);

                        // 入力に対する応答メッセージです
                        bytes = Encoding.ASCII.GetBytes("またの呼び出しをお待ちしております！");

                        // 応答を返します
                        sslStream.Write(bytes, 0, bytes.Length);
                    }
                }
            }
        }

        private static X509Certificate GetServerCert(string subjectName)
        {
            X509Store store = new X509Store(StoreName.My, StoreLocation.LocalMachine);
            store.Open(OpenFlags.ReadOnly);
            X509CertificateCollection certificate =
                store.Certificates.Find(X509FindType.FindBySubjectName,
                subjectName, true);

            if (certificate.Count > 0)
                return (certificate[0]);
            else
                return (null);
        }
    }
}
