﻿using System;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading;

namespace MutexFun
{
    [StructLayout(LayoutKind.Sequential)]
    [Serializable()]
    public struct Contact
    {
        public string _name;
        public int _age;
    }

    class Program
    {
        static void Main(string[] args)
        {
            if (args.Length == 0)
            {
                // 処理を開始するにあたって、最初に共有メモリマネージャを作成します。
                using (SharedMemoryManager<Contact> sm = new SharedMemoryManager<Contact>("Contacts", 8092))
                {
                    // これは送信プロセスです。

                    // 続行のための2つめのプロセスを開始します。
                    string processName = Process.GetCurrentProcess().MainModule.FileName;
                    int index = processName.IndexOf("vshost");
                    if (index != -1)
                    {

                        string first = processName.Substring(0, index);
                        int numChars = processName.Length - (index + 7);
                        string second = processName.Substring(index + 7, numChars);

                        processName = first + second;
                    }
                    Process receiver = Process.Start(new ProcessStartInfo(processName, "Receiver"));

                    // 処理の開始まで5秒間待機します。
                    Thread.Sleep(5000);

                    // Contactオブジェクトを作成します。
                    Contact man;
                    man._age = 18;
                    man._name = "コンドルのジョー";

                    // 共有メモリ経由で、他プロセスにデータを送信します。
                    sm.SendObject(man);
                }
            }
            else
            {
                // this is the receiver process

                // 処理を開始するにあたって、最初に共有メモリマネージャを作成します。
                using (SharedMemoryManager<Contact> sm =
                    new SharedMemoryManager<Contact>("Contacts", 8092))
                {

                    // 送信されたContactオブジェクトを受け取ります。
                    Contact c = (Contact)sm.ReceiveObject();

                    // コンソール(あるいはデータベースなど)に書き出します。
                    Console.WriteLine("Contact: {0}は{1}歳です。", c._name, c._age);

                    // メッセージを5秒間表示します。
                    Thread.Sleep(5000);
                }
            }

            // コンソールウィンドウを表示する場合は以下をアンコメントしてください。
            //Console.ReadLine();
        }
    }
}
