﻿using System;
using System.Diagnostics;
using System.Windows;

namespace UnhandledWPFExceptions
{
    /// <summary>
    /// App.xaml の相互作用ロジック
    /// </summary>
    public partial class App : Application
    {
        private void Application_DispatcherUnhandledException(object sender, System.Windows.Threading.DispatcherUnhandledExceptionEventArgs e)
        {
            // 例外が処理されたことを指示
            e.Handled = true;
            ReportUnhandledException(e.Exception);
        }

        private void Application_Startup(object sender, StartupEventArgs e)
        {
            // WPF UI例外
            this.DispatcherUnhandledException +=
                new System.Windows.Threading.DispatcherUnhandledExceptionEventHandler(
                    Application_DispatcherUnhandledException);

            // その他のスレッド例外
            AppDomain.CurrentDomain.UnhandledException +=
                new UnhandledExceptionEventHandler(CurrentDomain_UnhandledException);
        }

        private void CurrentDomain_UnhandledException(object sender, UnhandledExceptionEventArgs e)
        {
            ReportUnhandledException(e.ExceptionObject as Exception);
        }

        private void ReportUnhandledException(Exception ex)
        {
            // イベントログに例外情報を出力
            EventLog.WriteEntry("UnhandledWPFException Application",
                ex.ToString(), EventLogEntryType.Error);
            // ユーザに例外が発生したことを通知
            MessageBox.Show("Unhandled Exception: " + ex.ToString());
            // shut down the application
            this.Shutdown();
        }
    }
}
